#include <string.h>
#include "smex.h"
#include "moods.h"
#include "hhpj.h"
#include "memmng.h"
#include "hashtab.h"

byte StartSearch ( FID *fid, int bg, SCANDESC **sDesc, USERDESC **oDesc,
                   int offs, int count)
{
   BOOL  eof;

   sm_OpenScanWithGroup( fid, PHYSICAL_SCAN, bg, sDesc, NULL);

   sm_ScanNextObject(*sDesc, offs, count, oDesc, &eof );

   return (!eof);
}

int GetNext ( FID *fid, int bg, SCANDESC **sDesc, USERDESC **oDesc,
              int offs, int count)
{
   BOOL eof;

   sm_ScanNextObject(*sDesc, offs, count, oDesc, &eof );

   return (!eof);
}

void CloseSearch(SCANDESC *sDesc)
{
   sm_CloseScan(sDesc);
}

//-------------------------------------------------------------------
// ## Hybrid Hash Partition Join ##
//
// R and S contains info about the files to be joined.
// R->numObjs * oSize should be less than that of S, if the join
// is value based. If it is pointer based, that does not matter.
// buffer must be allocated outside, by the caller. bufSize is the
// size of this buffer in bytes.
// 
// The output is written into an Exodus file
//
// Lines starting with '//!' should be checked when a modification
// is needed to be done.
//-------------------------------------------------------------------

int HHPJ ( fInfo * R, fInfo * S, void * buffer, int bufSize )
{

  int        i, j, e, n, part, NSeg, NRes, goAhead;
  uint       R_oSize, S_oSize;
  FILE       * Result;
  void       * R_oInfo, * S_oInfo;
  MemMng     Buf;
  HashTab    HT;
  uint       * R_oCounts, * S_oCounts;
  byte       * o;
  int        tmpBG;             // Temporary buffer group
  Bucket     * bp;

  FOUR       SegSizeR, SegSizeS, BufGSize;
  FID        Rfid, Sfid;
  OID        *SegmentsR, *SegmentsS, *jList, ResOID;
  SCANDESC   *R_sDesc, *S_sDesc;
  USERDESC   *R_oDesc, *S_oDesc, *Res_oDesc;

// Eval the sizes of entries for R and S
  R_oSize = OIDSIZE + R->attrSize;
  S_oSize = OIDSIZE + S->attrSize;

// Eval # of segments
//!  NSeg = Ceiling(2.5 * 1.2 * R->oSize * R->numObjs / bufSize) ;
  NSeg = Ceiling(2 * 1.2 * R->oSize * R->numObjs / bufSize) ;

// Eval sizes of segments in relations R and S (in 4K pages)
//!  SegSizeR = Ceiling ( bufSize / (2.5 * 4096) );
//!  SegSizeS = Ceiling ( (S->oSize * S->numObjs * 1.2) / (NSeg * 4096) );
  SegSizeR = Ceiling ( bufSize / (2.0 * 4096) );
  SegSizeS = Ceiling ( (S->oSize * S->numObjs * 1.2) / (NSeg * 4096) );

// Allocate space for the OIDs of the temporary disk pages
  SegmentsR = (OID *) malloc ( NSeg * sizeof(OID) );
  if (SegmentsR == NULL) {
    MoodsError(ERR_MEMOUT); return ERROR;
  }

  SegmentsS = (OID *) malloc ( NSeg * sizeof(OID) );
  if (SegmentsS == NULL) {
    MoodsError(ERR_MEMOUT); free(SegmentsR); return ERROR;
  }

// Open the buffer group
//!  BufGSize = max(SegSizeR, SegSizeR);        // Buffer group for one segment
  BufGSize = 20;      // Fixed buffer size of 20 4K-pages
  e = sm_OpenBufferGroup( BufGSize, BF_MRU, &tmpBG, NOFLAGS );
  if (e!=esmNOERROR)
     return(ERROR);

// Create the temporary file for segments of R
//!  e = sm_CreateFile(tmpBG, volid, &Rfid, NOFLAGS);
  e = sm_CreateFile(tmpBG, volid, &Rfid);
  if (e!=esmNOERROR)
     return(ERROR);

// Create the temporary file for segments of S
//!  e = sm_CreateFile(tmpBG, volid, &Sfid, NOFLAGS);
  e = sm_CreateFile(tmpBG, volid, &Sfid);
  if (e!=esmNOERROR)
     return(ERROR);

// Create one temporary object for each segment of the file R & S
  for (i=0; i<NSeg; i++) {

    e = sm_CreateObject ( tmpBG, &Rfid, NEAR_LAST_PHYSICAL, NULL,
			  NULL, 0, NULL, &SegmentsR[i] );
    if (e!=esmNOERROR)
        return(ERROR);

    e = sm_CreateObject ( tmpBG, &Sfid, NEAR_LAST_PHYSICAL, NULL,
			  NULL, 0, NULL, &SegmentsS[i] );
    if (e!=esmNOERROR)
       return(ERROR);

  }

// Allocate space for the memory pointers
  if ( Buf.Setup(buffer, bufSize, NSeg, SegmentsR, tmpBG) == ERROR ) {
    MoodsError(ERR_MEMOUT);
    free(SegmentsR); free(SegmentsS);
    return ERROR;
  }

// Scan thru R, hash each object on the join attribute, and distribute
// onto the partitions.

  R_oInfo = (void *) malloc(R_oSize);      // A bit tricky...
  if (R_oInfo == NULL) {
    MoodsError(ERR_MEMOUT);
    Buf.Destroy();
    free(SegmentsR); free(SegmentsS);
    return ERROR;
  }

  goAhead = StartSearch ( &R->fid, tmpBG, &R_sDesc, &R_oDesc, 0,READ_ALL );
 i = 0;
  while (goAhead) {

// Calculate the partition number for this object's entry
   part = Hash((byte *)(R_oDesc->basePtr+R->offs), R->attrSize, NSeg);

// Prepare the object info package
   memcpy(R_oInfo, &R_oDesc->oid,12);
   memcpy(R_oInfo+12, R_oDesc->basePtr+R->offs, R->attrSize);

// Put it into the buffer. If the bucket gets full, it will be flushed
    if (Buf.Write(part, (byte *)R_oInfo, R_oSize) == ERROR) {
      printf("Error writing to buffer. Possibly non-uniform hashing.\n");
      free(SegmentsR); free(SegmentsS);
      return ERROR;
    }

// Read the next object from the file sequentially
    goAhead = GetNext ( &R->fid, tmpBG, &R_sDesc, &R_oDesc,0,READ_ALL );
    i++;
  }

// Objects of R have been scanned and partitioned. Close the search
  CloseSearch(R_sDesc);

// Allocate space for the object counters of R
  R_oCounts = (uint *) malloc(NSeg * sizeof(uint));
  if (R_oCounts == NULL) {
    MoodsError(ERR_MEMOUT);
    free(SegmentsR); free(SegmentsS);
    return ERROR;
  }

  if (Buf.FlushAll() == ERROR)
     return ERROR;
  for (i=0; i<NSeg; i++)
    R_oCounts[i] = Buf.oCount(i, R_oSize);

// Flush all object entries in the buffer and clean it
  Buf.Destroy();

// Reorganize the buffer for the S objects
  if ( Buf.Setup(buffer, bufSize, NSeg, SegmentsS, tmpBG) == ERROR ) {
    MoodsError(ERR_MEMOUT);
    free(SegmentsR); free(SegmentsS);
    return ERROR;
  }

// Scan thru S, hash each object on the join attribute, and distribute
// onto the partitions.

   S_oInfo = (void *) malloc(S_oSize);      // A bit tricky...

   goAhead = StartSearch ( &S->fid, tmpBG, &S_sDesc, &S_oDesc,0,READ_ALL);

   while (goAhead) {

// Calculate the partition number for this object's entry
     part = Hash((byte *)S_oDesc->basePtr+S->offs, S->attrSize, NSeg);

// Prepare the object info package
     memcpy(S_oInfo, &S_oDesc->oid,12);
     memcpy(S_oInfo+12, S_oDesc->basePtr+S->offs, S->attrSize);

// Put it into the buffer. If the bucket gets full, it will be flushed
     Buf.Write(part, (byte *)S_oInfo, S_oSize);

// Read the next object from the file sequentially
     goAhead = GetNext ( &S->fid, tmpBG, &S_sDesc, &S_oDesc,0,READ_ALL);

   }

// Objects of S have been scanned and partitioned. Close the search
   CloseSearch(S_sDesc);

// Allocate space for the object counters of S
  S_oCounts = (uint *) malloc(NSeg * sizeof(uint));
  if (S_oCounts == NULL) {
    MoodsError(ERR_MEMOUT);
    free(SegmentsR); free(SegmentsS);
    return ERROR;
  }

  if (Buf.FlushAll() == ERROR)
      return ERROR;
  for (i=0; i<NSeg; i++)
    S_oCounts[i] = Buf.oCount(i, S_oSize);

// Flush all object entries in the buffer and clean it
   Buf.Destroy();

// Now the buffer will be used to construct a hash table
//!   e = HT.Setup(buffer, (int)(0.4*bufSize), buffer + (int)(0.4*bufSize),
//!            (int)(0.1*bufSize), HASHSIZE, R_oSize, OIDSIZE, R->attrSize);
   e = HT.Setup(buffer, (int)(0.5*bufSize), buffer + (int)(0.5*bufSize),
                (int)(0.4*bufSize), HASHSIZE, R_oSize, OIDSIZE, R->attrSize);
//                     ^-- This is to spare at least 1/10 of the buffer
//                         for joining the records

   if (e == ERROR) {               // Just to trace. Should be reconsidered.
     printf("Could not initialize the hash table.\n");
     return ERROR;
   }

// Create a file to keep the results
   Result = fopen("hhpj.res", "w");
   if (Result==NULL) {
     printf("Could not create Result file.\n");
     return ERROR;
   }

// For all segments; read the segment from R; build a hash table
// in the first half of the buffer; read the corresponding partition of S;
// hash the join attribute; lookup from the hash table; join the corresponding
// records with the same hash value; output to a file.

   for (i=0; i<NSeg; i++) {

// Create the object where the join results will be hold
    e = sm_CreateObject ( tmpBG, &Rfid, NEAR_LAST_PHYSICAL, NULL,
			  NULL, 0, NULL, &ResOID );
    if (e!=esmNOERROR)
       return(ERROR);

// Read partition i from file R
     e = sm_ReadObject(tmpBG, &SegmentsR[i], 0, READ_ALL, &R_oDesc);
     if (e!=esmNOERROR)
        return(ERROR);

// For all object entries in this segment
     for (j=0, R_oInfo=(void *)R_oDesc->basePtr; j<R_oCounts[i];
          R_oInfo+=R_oSize, j++)
       if (HT.Put((byte *)R_oInfo) == ERROR) {
         printf("No overflow space or insufficient memory.");
         return ERROR;
       }

   sm_ReleaseObject(R_oDesc);

// Setup a memory manager in the rest of the available space
     if ( Buf.Setup(HT.OvfArea(), HT.OvfAreaSize(), 1, &ResOID, tmpBG) == ERROR ) {
       MoodsError(ERR_MEMOUT);
       free(SegmentsR); free(SegmentsS);
       return ERROR;
     }

// Read partition i from file S
     e = sm_ReadObject(tmpBG, &SegmentsS[i], 0, READ_ALL, &S_oDesc);
     if (e!=esmNOERROR)
         return(ERROR);

     NRes=0;
// For all object entries in this segment
     for (j=0, S_oInfo=(void *)S_oDesc->basePtr; j<S_oCounts[i]; S_oInfo+=S_oSize, j++) {
       R_oInfo = (void *) HT.Search((byte *)S_oInfo, &bp, &n);

       while (R_oInfo != NULL) {
         NRes++;
         Buf.Write(0, (byte *)R_oInfo, sizeof(OID));
         Buf.Write(0, (byte *)S_oInfo, sizeof(OID));
         R_oInfo = (void *) HT.SearchNext((byte *)S_oInfo, &bp, &n);
       }
     }

// Read result object
   sm_ReleaseObject(S_oDesc); 
   Buf.FlushAll();
   e = sm_ReadObject(tmpBG, &ResOID, 0, READ_ALL, &Res_oDesc);
   if (e!=esmNOERROR)
     return(ERROR);

   fwrite(Res_oDesc->basePtr, sizeof(OID), 2*NRes, Result);
   sm_ReleaseObject(Res_oDesc);

   e = sm_DestroyObject(tmpBG, &ResOID);
   if (e!=esmNOERROR)
     return(ERROR);

  }

  if ((e = sm_DestroyFile(tmpBG,&Rfid)) != esmNOERROR)
     return ERROR;
  if ((e = sm_DestroyFile(tmpBG,&Sfid)) != esmNOERROR)
     return ERROR;
  if((e = sm_CloseBufferGroup(tmpBG)) != esmNOERROR)
    return ERROR;
  free(R_oInfo);
  free(S_oInfo);
  free(SegmentsR); 
  free(SegmentsS);
  fclose(Result);
  return OK;
}


