#include <stdio.h>
#include <memory.h>
#include <setjmp.h>
#include "sm_client.h"
#include "sm_SortFile.h"

#ifdef __cplusplus
# include <libc.h>
#else
  extern char* getenv(char*);
#endif
 
/* Global Decls. */
OID OidList[SEGMENTSIZE];
void *KeyList[SEGMENTSIZE],*TempKey;
int LengthofKey;   /* for use of SwapLists() */
int NumofSegments;
struct SegmentInfo Segment[SEGMENTSIZE]; 

/* Function Prototypes  */
void InitGreaterFunc(SMDATATYPE);
void ArrangeHeapForward(int,int);
void ArrangeHeapBackward(int);
void SwapLists(int,int);
int FindItem(BOOL *);
BOOL GreaterInt(int *,int *);
BOOL GreaterFloat(float *,float *);
BOOL GreaterChar(char *,char *);
BOOL GreaterString(char *,char *);
BOOL (*Greater)(void *,void *);

int PrepareKeyOidFile(int Volume,int groupIndex,FID GivenFid,FID *NewFid,
                      int KeyStart,int KeyLength, int *TotObjects)
{
  SCANDESC *ScanDesc;
  USERDESC *UserDesc;
  int LBgi;
  OID Oid,NewOid;
  void *Object;
  BOOL Eof;

  if (sm_OpenBufferGroup(MAXNUMOFPAGES,BF_LRU,&LBgi,TRANS_GROUP))
    return esmFAILURE;

  if (sm_OpenScanWithGroup(&GivenFid,LOGICAL_SCAN,LBgi,&ScanDesc,NULL))
    return esmFAILURE;
  
  if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
    return esmFAILURE;

  (*TotObjects)=0; 

  if (!Eof) {
    if (sm_CreateFile(groupIndex,Volume,NewFid))
      return esmFAILURE;
  }
  else {
    printf("No object in the given file!...");
    return esmFAILURE;
  }
  
  while (!Eof) { 
    Object=UserDesc->basePtr;
    Oid=UserDesc->oid;
    if (sm_CreateObject(groupIndex,NewFid,NEAR_LAST_PHYSICAL,NULL,NULL,
                        SizeofOID,&Oid,&NewOid))
      return esmFAILURE;

    if (sm_AppendToObject(groupIndex,&NewOid,KeyLength,Object+KeyStart))
      return esmFAILURE;

    (*TotObjects)++;

    if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
      return esmFAILURE;
  } /* while */

  if (sm_CloseScan(ScanDesc))
    return esmFAILURE;
 
  return esmNOERROR;
} //PrepareKeyOidFile 

int sm_SortFile(int Volume,int groupIndex,FID GivenFid,FID *SortFid,
                 SMDATATYPE KeyType,int KeyStart,int KeyLength)
{
  int ObjectSize,TotObjects,LBgi,HSize,i,Count,j;
  FID SortFID,TempFID,CurrFID;
  SCANDESC *ScanDesc;
  USERDESC *UserDesc,*LUDesc;
  BOOL Eof,AllExhausted;
  OID TempOid;

  /* Initialization of Comparetor function "Greater" */ 
  InitGreaterFunc(KeyType); 

  for(i=0;i<SEGMENTSIZE;i++)
    KeyList[i]=malloc(KeyLength);
  TempKey=malloc(KeyLength);
  LengthofKey=KeyLength;
  

  if (PrepareKeyOidFile(Volume,groupIndex,GivenFid,&TempFID,KeyStart,KeyLength,&TotObjects)) 
    return esmFAILURE;

  if (sm_OpenBufferGroup(LIMITEDPAGES,BF_LRU,&LBgi,TRANS_GROUP))
    return esmFAILURE;

  if (sm_OpenScanWithGroup(&TempFID,LOGICAL_SCAN,LBgi,&ScanDesc,NULL))
    return esmFAILURE;
  
  if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
    return esmFAILURE;
 
  /* This while loop forms the initial sorted segments */
  NumofSegments=0;
  while (!Eof) {
    HSize=0;
    while (!Eof && (HSize < SEGMENTSIZE)) {
      OidList[HSize]=*((OID *)UserDesc->basePtr);
      memcpy(KeyList[HSize],UserDesc->basePtr+12,KeyLength);
      ArrangeHeapBackward(HSize);
      if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
        return esmFAILURE;
      HSize++;
    } /* while */
    
    /* Create file to output the sorted segments */
    if (sm_CreateFile(groupIndex,Volume,&CurrFID))
      return esmFAILURE;
    /* Record the segment information to the global structure */
    Segment[NumofSegments].HeapSize=HSize;
    Segment[NumofSegments].SegFid=CurrFID;
    Segment[NumofSegments].EofReached=FALSE;
    Segment[NumofSegments].CurrItem=malloc(KeyLength);
    NumofSegments++;

    /* Output the sorted segments */
    for (i=HSize-1;i>=0;i--) {
      /* Write to the file */
      if (sm_CreateObject(groupIndex,&CurrFID,NEAR_LAST_PHYSICAL,NULL,NULL,
                          SizeofOID,&OidList[0],&TempOid))
        return esmFAILURE;
      if (sm_AppendToObject(groupIndex,&TempOid,KeyLength,KeyList[0]))
        return esmFAILURE;
      OidList[0]=OidList[i];
      memcpy(KeyList[0],KeyList[i],KeyLength);
      ArrangeHeapForward(0,i);
    } /* for */
  } /* while */
  
  if (sm_CloseScan(ScanDesc))
    return esmFAILURE;
  
  /* Now, merge... */
  for(i=0;i<NumofSegments;i++) {
    if (sm_OpenBufferGroup(LIMITEDPAGES,BF_LRU,&(Segment[i].SegBgi),TRANS_GROUP))
      return esmFAILURE;

    if (sm_OpenScanWithGroup(&(Segment[i].SegFid),LOGICAL_SCAN,
                               Segment[i].SegBgi,&(Segment[i].SegSDesc),NULL))
      return esmFAILURE;
  
    if (sm_ScanNextObject(Segment[i].SegSDesc,0,READ_ALL,&UserDesc,
                        &(Segment[i].EofReached)))
      return esmFAILURE;

    if (!Segment[i].EofReached) {
      memcpy(Segment[i].CurrItem,UserDesc->basePtr+12,KeyLength);
      memcpy(&(Segment[i].CurrOid),UserDesc->basePtr,SizeofOID);
    }
  } /* for */
  
  if (sm_CreateFile(groupIndex,Volume,SortFid))
    return esmFAILURE;
 
  AllExhausted=FALSE;
  while (TRUE) {
    Count=FindItem(&AllExhausted);
    if (AllExhausted) break; /* merge complete */
 
    if (sm_ReadObject(LBgi,&(Segment[Count].CurrOid),0,READ_ALL,&LUDesc))
      return esmFAILURE;

    if (sm_CreateObject(groupIndex,SortFid,NEAR_LAST_PHYSICAL,NULL,NULL,
                        LUDesc->objectSize,LUDesc->basePtr,&TempOid))
      return esmFAILURE;

    if (sm_ReleaseObject(LUDesc))
      return esmFAILURE;

    if (sm_ScanNextObject(Segment[Count].SegSDesc,0,READ_ALL,&UserDesc,
                          &(Segment[Count].EofReached)))
      return esmFAILURE;
    
    if (!Segment[Count].EofReached) {
      memcpy(Segment[Count].CurrItem,UserDesc->basePtr+12,KeyLength);
      memcpy(&(Segment[Count].CurrOid),UserDesc->basePtr,SizeofOID);
    }
  } /* while */
  /* End of merge... */
  
  for(i=0;i<NumofSegments;i++) {
    if (sm_DestroyFile(Segment[i].SegBgi,&(Segment[i].SegFid)))
      return esmFAILURE;

    if (sm_CloseScan(Segment[i].SegSDesc))
      return esmFAILURE;

    if (sm_CloseBufferGroup(Segment[i].SegBgi))
      return esmFAILURE;
  } /* for */
  
  for(i=0;i<SEGMENTSIZE;i++)
    free(KeyList[i]);

  return esmNOERROR;
} //SortFile 


void InitGreaterFunc(SMDATATYPE KeyType)
{
  switch(KeyType) {
    case SM_int   : Greater=&GreaterInt;
                    break;
    case SM_float : Greater=&GreaterFloat;
                    break;
    case SM_char  : Greater=&GreaterChar;
                    break;
    case SM_string: Greater=&GreaterString;
                    break;
    default       : printf("Invalid Key Type!...");
                    exit(1);
  } /* switch */
} /* InitGreaterFunc */

void ArrangeHeapForward(int Parent,int HeapSize)
{
  int Changed;
  int LChild,RChild;

  Changed=TRUE;
  while (Changed && ((Parent+1)*2 <= HeapSize)) {
    LChild=2*Parent+1;
    RChild=LChild+1;
    if (Greater(KeyList[Parent],KeyList[LChild]) ||
        Greater(KeyList[Parent],KeyList[RChild])) {
      if (Greater(KeyList[RChild],KeyList[LChild])) {
        SwapLists(Parent,LChild);
        Parent=LChild;
      }
      else {
        SwapLists(Parent,RChild);
        Parent=RChild;
      }
    } /* if */
    else
      Changed=FALSE;
  } /* while */
} /* ArrangeHeapForward */

void ArrangeHeapBackward(int Child)
{
  int Changed;
  int Parent;

  Changed=TRUE;
  while (Changed && (Child > 0)) {
    Parent=(Child-1)/2;
    if (Greater(KeyList[Parent],KeyList[Child])) {
      SwapLists(Child,Parent);
      Child=Parent;
    } /* if */
    else
      Changed=FALSE;
  } /* while */
} /* ArrangeHeapBackward */

void SwapLists(int Item1,int Item2)
{
  OID TempOid;

  TempOid=OidList[Item2];
  OidList[Item2]=OidList[Item1];
  OidList[Item1]=TempOid;

  memcpy(TempKey,KeyList[Item2],LengthofKey);
  memcpy(KeyList[Item2],KeyList[Item1],LengthofKey);
  memcpy(KeyList[Item1],TempKey,LengthofKey);
} /* SwapLists */

int FindItem(BOOL *AllExhausted)
{
  int i,ItemPlace;
  
  for(i=0;Segment[i].EofReached && i<NumofSegments;i++);
  if (i==NumofSegments) {
    *AllExhausted=TRUE;
    return(0);
  }
  ItemPlace=i;
  for(;i<NumofSegments;i++) {
    if ((!Segment[i].EofReached) &&
      (Greater(Segment[ItemPlace].CurrItem,Segment[i].CurrItem)))
        ItemPlace=i;
  } /* for */
  return(ItemPlace);
} /* FindItem */

BOOL GreaterInt(int *Param1,int *Param2)
{
  if (*Param1 > *Param2)
    return TRUE;
  else
    return FALSE;
} /* GreaterInt */

BOOL GreaterFloat(float *Param1,float *Param2)
{
  if (*Param1 > *Param2)
    return TRUE;
  else
    return FALSE;
} /* GreaterFloat */

BOOL GreaterChar(char *Param1,char *Param2)
{
  if (*Param1 > *Param2)
    return TRUE;
  else
    return FALSE;
} /* GreaterChar */

BOOL GreaterString(char *Param1,char *Param2)
{
  if (strcmp(Param1,Param2) > 0)
    return TRUE;
  else
    return FALSE;
} /* GreaterString */

