#include<lists.h>
// #include <esm_ops.h>
// int Debug=0;
extern int Debug;
extern ERTYPE ErrorCheck(int e,char *routine);


//           ************** List::List  *****************
//  Constructor for a known List ( defined previously ) 
//  Just get the List in to memory if exists, otherwise 
//  List the PStatus variable to Error Code ....
//  IN  >>>    PListOid  List object's OID to get the object ...
//  OUT <<<    PTypeId   The Object's Type ID 
//  OUT <<<    PSize     The Size of Object
//  OUT <<<    PStatus   
//                       if 0 OUT ->list exists & loaded.
//                       if 1 OUT ->list  does not exist.
//          ******************************************

List::List(OID *PListOid,TYPEID *PTypeID,SIZTYP *PSize, CatBOOLEAN *PStatus) {
     // Copy Some important values to member variables ,,,,,,

     int EsmError;

     BufGrp   = BufferGroup; 
     Tid      = TransactionId; 
     ListOid   = *PListOid;
     FileId   = FileIdentifier;
     ListPos  = 0;
     

     //  Test if it Works !!!!!
     //  Read requested List  into memory....

     EsmError = sm_ReadObject(BufGrp,&ListOid,0,READ_ALL,&UserDesc);

     if(EsmError != esmNOERROR) { 
        *PStatus   = 1;
        *PSize     = 0;
        PTypeID    = NULL;
        ErrorCheck(EsmError,"List consrtuctor, getting ...\n");
     } else { 
        *PStatus=0;
        Size      = (int) UserDesc->objectSize;
        Buffer    = malloc(Size);
        memcpy(Buffer,UserDesc->basePtr,Size) ;
        ListHead  = (OID *) ( (char *)Buffer +  S_TypeId + S_INT32  ); 
        // List Members start here.......
        // Get the maximum length of list
        memcpy(&ListMax,((char *)Buffer + S_TypeId),S_INT32);           
        PTypeID   = (TYPEID *)Buffer ; 
        Size      = Size - ( S_TypeId + S_INT32 );
        Size      = Size / S_OID; 
        *PSize     = Size;
        unpin();
     };
};

//           ************** List::List  *****************
//  Constructor for List creation 
//  the List  created which has only a TypeId Field filled with MOODSSET value. 
//  List the PStatus variable to Error Code ....
//  IN  >>>  PFid      File Id for the placement of the new object.  
//  IN  >>>  PPos      Position dependency to given PNearOid. 
//  OUT <<<  PStatus   returned from constructor . 0 for success, err otherwise.
//           ******************************************

List::List(FID *PFid,NTYPE PPos,OID *PNearOid,INT32 Listmax,CatBOOLEAN *PStatus) {
     // Copy Some important values to member variables ,,,,,,

     int EsmError;
     void * buffer;
     TYPEID type  = MOODSLIST;

     BufGrp   = BufferGroup; 
     Tid      = TransactionId; 

     // Create a List object with no elements in file ................
     // Save   List's Object Id in member variable for its lifetime...
     
     buffer = malloc( S_TypeId + S_INT32 );
     memcpy(buffer,&type,S_TypeId);
     memcpy(((char *)buffer+S_TypeId),&Listmax,S_INT32);
     if (PNearOid)
        EsmError = sm_CreateObject(BufGrp,PFid,PPos,PNearOid,NULL, 
                                S_TypeId + S_INT32,buffer,&ListOid);
     else
        EsmError = sm_CreateObject(BufGrp,PFid,PPos,NEAR_LAST,NULL, 
                                S_TypeId + S_INT32,buffer,&ListOid);


     if(EsmError != esmNOERROR) {
        *PStatus = EsmError ; 
        ErrorCheck(EsmError,"creating list...\n"); }
        else {
        *PStatus = 0;
         Size = 0;
       };
//     unpin();
     ListHead = NULL;
     Size = 0;
};

ERTYPE List::updateList(OID *POid, INT32 Pos, CatBOOLEAN *PStatus) {

int Offset;
int err;

    if ( ( Pos < 0 ) && ( Pos >= Size ) ) { *PStatus = 1; return(0); }; 
    Offset = Pos * S_OID + S_TypeId + S_INT32 ;
    
    err    = sm_WriteObject(BufGrp,Offset,S_OID,POid,UserDesc,TRUE);
    if (err != esmNOERROR ) ErrorCheck(err,"Append To Object \n");

    err    = sm_ReadObject(BufGrp,&ListOid,0,READ_ALL,&UserDesc);
    if (err != esmNOERROR ) ErrorCheck(err,"Append To  Object \n");

    ListHead  = (OID *) (UserDesc->basePtr + S_TypeId + S_INT32 );
    Buffer    = UserDesc->basePtr ;
    return(0);
    
};

OID  *List::operator[] (INT32 Subscript) {

OID *tmpoid;

    tmpoid = (OID *)malloc(S_OID);
    memcpy(tmpoid,&ListHead[Subscript],S_OID);
    return(tmpoid);

};

void List::getNext(OID *POid ,CatBOOLEAN *PStatus) {

       if( (ListPos+1) < Size ) {
           memcpy(POid, &ListHead[ ++ListPos ],S_OID );;
           *PStatus = 0;
       } else *PStatus = 1;
};

void List::getPrev(OID *POid,CatBOOLEAN *PStatus) {

       if( ListPos > 0 ) {
         memcpy(POid , &ListHead[ --ListPos ],S_OID);;
         *PStatus = 0;
       } else *PStatus = 1;
};


ERTYPE List::insertToList(OID *POid,CatBOOLEAN *PStatus) {

int Offset = 0;
int err;

     BufGrp   = BufferGroup; 
     Tid      = TransactionId; 
     FileId   = FileIdentifier;
/*
     if (Debug) {
         cout << "userdesc=" << UserDesc << "\n" ;
         cout << "inserting....:" ;
         display_Oid(*POid);
     } ;
*/
/********************************************
     if ( Size  >= ListMax ) { 
          *PStatus = 1 ; 
          cout << "size of list exceeded \n";
          return(0);
     };
********************************************/  
  
     err    = sm_AppendToObject(BufGrp,&ListOid,S_OID,(void *)POid);
     if (err != esmNOERROR ) ErrorCheck(err,"Append To Object \n");
     *PStatus = (CatBOOLEAN ) 0;
/*******************************************
     err    = sm_ReadObject(BufGrp,&ListOid,0,READ_ALL,&UserDesc);
     if (err != esmNOERROR ) ErrorCheck(err,"Append To  Object \n");

     ListHead  = (OID *) (UserDesc->basePtr + S_TypeId + S_INT32 );
     Buffer    = UserDesc->basePtr ;
     Size   = Size + 1;
      unpin();
*******************************************/
     return(0);   
};

ERTYPE List::insertToList(OID *POid,INT32 Position ,CatBOOLEAN *PStatus) {

int ListIndex, Offset = 0;
int err;

     BufGrp   = BufferGroup; 
     Tid      = TransactionId; 
     FileId   = FileIdentifier;

     unpin();

     if ( Size  >= ListMax ) { 
          *PStatus = 1 ; 
          //printf("size of list exceeded \n");
          return(0);
     };
       Offset = Position * S_OID + S_TypeId + S_INT32 ;
       err    = sm_InsertInObject(BufGrp,&ListOid,Offset,S_OID,(void *)POid);
       if (err != esmNOERROR ) ErrorCheck(err,"Insert in Object \n");
       err    = sm_ReadObject(BufGrp,&ListOid,0,READ_ALL,&UserDesc);
       if (err != esmNOERROR ) ErrorCheck(err,"Insert in Object \n");
       Size   = Size + 1;
     //  return(0);   

     ListHead  = (OID *) (UserDesc->basePtr + S_TypeId + S_INT32 );
     Buffer    = UserDesc->basePtr ;
     Size   = Size + 1;
     return(0);   
};

ERTYPE List::deleteFromList(INT32 PStart,INT32 PEnd,CatBOOLEAN *PStatus) {
int  Offset ;
int ElementCount=0;
int err;

       //  Find Correct Place To delete Oid ..............

       unpin();

       if ( !( ( PStart >= 0 ) && ( PEnd < Size ) )  )  { 
             *PStatus = 1 ;
             printf( " 1 test failed exiting from routine \n") ;
             return(0); 
       } ;
       if ( PStart > PEnd ) { *PStatus = 1  ; return(0); };
       ElementCount = PEnd - PStart + 1;
       printf("must be erased\n"); 
       Offset= PStart ;
       Offset= Offset * S_OID + S_TypeId + S_INT32  ; 
       err   = sm_DeleteFromObject(BufGrp,&ListOid,Offset,ElementCount * S_OID);
       Size   = Size - ElementCount ;
       //printf("new size= %d",Size); 

       ErrorCheck(err,"Delete From Object \n");

       // Delete below line please....

       *PStatus = 0; 
       return(0);   
};

ERTYPE List::isIn(OID POid,CatBOOLEAN *PStatus) {
int ListIndex ;

       //  Find Object in list..............
    
       if (Debug) display_Oid(POid);
       *PStatus = 1 ;     // For the worst case !! .......

       for ( ListIndex=0; ListIndex < Size; ListIndex++ ) 
             if ( Test_Oid( ListHead[ListIndex] , POid , T_EQ ) ) *PStatus=0;

       //  Anything to ask ?...

       return(0);
};

ERTYPE List::getElement(OID *POid , int ListIndex ,CatBOOLEAN *PStatus) {

       if ( ListIndex > Size ) {  // If wanted more than list has ......
            *PStatus = 1 ;
            return (0) ;
       };

       memcpy(POid, &ListHead[ ListIndex ], S_OID );;
       *PStatus = 0 ;            // Everything Seems ok................
       return(0);

};

ERTYPE List::list_Union(List PList,List UList ,CatBOOLEAN *PStatus) { 

int Listdex;
OID Element;

    if ( UList.Size ) {      // Union List is not empty ..............
         *PStatus = 1 ;
          return(0);  
    };

    /*      
         Insert first the elements of current list 
                                                  */

    for (Listdex = 0 ; Listdex <  Size ; Listdex++ )  {


         getElement( &Element, Listdex, PStatus ) ;

         if ( !( *PStatus ) ) {  // If element received from current list..
             if (Debug) display_Oid(Element);
             UList.insertToList( &Element , PStatus );
         } 
    };

    /* 
         Now Insert Other list's elements ...
                                                 */

    for (Listdex = 0 ; Listdex < PList.Size ; Listdex++ )  {

         PList.getElement( &Element, Listdex, PStatus)  ;

         if ( !( *PStatus ) ) { // If element received from second  list.....
             UList.insertToList( &Element , PStatus ) ;
             if (*PStatus)  return(0);
         } 
    };
    return( 0 );
};
 
ERTYPE List::isSublist(List PList,CatBOOLEAN *PStatus) {
int  ListIndex;
OID  Element;

     ListIndex=0;
    
     if ( Size > PList.Size )  {       // If sublist, how can it be larger !... 
        *PStatus = 1;
        return(0);
     };
                                     // Check all elements in current list if
    do {                             // They are in the target..........
        getElement(&Element,ListIndex++,PStatus);
        PList.isIn(Element,PStatus);
    } while ( ( ListIndex < Size)  && ( ! (*PStatus)  ) );

    return(0);
};

ERTYPE List::isContain(List PList, CatBOOLEAN *PStatus) {
int ListIndex ;
OID Element ;


    ListIndex = 0;

    if ( Size <= PList.Size ) {    // If PList is larger no need to check...
       *PStatus =  1;
       return(0);                 // Not an ESM error.....................
    };

    PList.getFirst(&Element,PStatus);

    if ((*PStatus)) {       // If No elements in current list as can be seen !..
       return(0);           // Not an ESM error.....................
    };

    isIn(Element,PStatus);

    for (ListIndex = 1; ListIndex < PList.Size ; ListIndex++) {

        PList.getNext(&Element,PStatus); 

        isIn(Element,PStatus);

        if (*PStatus) return(0);

     } 
    
     return(0);                        // No way to get an ESM error !!!...

};


OID List::getListOid(CatBOOLEAN *PStatus) {

    if ( OID_IS_INVALID(ListOid) )  // Test if program cheats......
       *PStatus = 1 ;
    else 
       *PStatus = 0;

    return(ListOid);               // May be changed to our standard.!
};

void List::getListOid( OID *POid, CatBOOLEAN *PStatus) {
     if ( OID_IS_INVALID(ListOid) )  // test if prog. cheats..
        *PStatus = 1 ;
     else
        *PStatus = 0 ;
     memcpy( POid , &ListOid, S_OID ); 
};


ERTYPE List::getFirst(OID *POid,CatBOOLEAN *PStatus) { 

       if (!Size) { // There is no list ! or at least not in memory....
          *PStatus = 1;
          POid=NULL;
          return(0);
       };
       // Original point, elements start...(see  list def.).
       *PStatus = 0;
       memcpy(POid,ListHead,S_OID);
       return(0);

};

ERTYPE List::getLast(OID *POid,CatBOOLEAN *PStatus) {

       memcpy(POid , &ListHead[Size],S_OID);
       *PStatus  = 0;           // Think why need  this ?....
       return(0);               // Tell me if you find out !!...

};

void List::intersect(List PList,List IList,CatBOOLEAN *PStatus) {
       
int Listindex;
OID Element;

        if ( Size <= PList.Size ) {
             for ( Listindex = 0 ; Listindex < Size ; Listindex ++ ) {
                   getElement ( &Element , Listindex , PStatus ) ;
                   PList.isIn  ( Element  , PStatus );
                   if ( ! (*PStatus ) ) 
                        IList.insertToList ( &Element , PStatus );
             };
             *PStatus = 0;
        } 
        else {
             for ( Listindex = 0 ; Listindex < PList.Size ; Listindex ++ ) {
                   PList.getElement ( &Element , Listindex , PStatus ) ;
                   isIn  ( Element  , PStatus );
                   if ( ! (*PStatus ) ) 
                        IList.insertToList ( &Element , PStatus );
             };
             *PStatus = 0;
        }
};
