#include "mood_catdef.h"
#include "mood_catalog.h"

#define KEYLENGTH 15	// Key is "OID".
			// Elemenet is RefCount$Self.

typedef unsigned short int INT16;

unsigned short int bitpos[] = {	0x2000, 0x1000, 0x0800, 0x0400,
				0x0200, 0x0100, 0x0080, 0x0040, 
				0x0020, 0x0010, 0x0008, 0x0004 };

void charTOoid(char *p, OID *oid)
{
	int i;
	INT16 *si = (INT16 *)(p+12);

	for( i=0; i<12; i++ )
		if (*si & bitpos[i]) p[i]=0;
    
	memcpy( (void *)oid, (void *)p, 12 );
}

void oidTOchar(OID *oid, char *p)
{
	int i;
	INT16 *si = (INT16 *)(p+12);

	*si = 0x8001;
	memcpy((void *)p, (void *)oid, 12);
	for( i=0; i < 12; i++) 
		if (!p[i]) {
			*si |= bitpos[i];
			p[i]=1;
		}
   
	p[14]=0;
}

int IncRefCount(OID Oid)
{
	int		err, RefCount, Self;
	OID		RetOid;
	KEY		Key;
	SMCURSOR	Cursor;
	USERDESC	*UserDesc;
	char		CharOid[KEYLENGTH];
	char		Elem[8];
	BOOL		EndofFile;

	if (OID_IS_INVALID(Oid)) return TRUE;

	oidTOchar(&Oid,CharOid);
	Key.length = KEYLENGTH;
	Key.valuePtr = CharOid;

	// Check whether there is an entry...
	err = sm_FetchInit(&RefIID,BufferGroup,&Key,SM_EQ,&Key,SM_EQ,&Cursor);
	ERROR_HANDLER("Increment Ref Count");

	err = sm_FetchNext(&Cursor,&Key,(void *)Elem,&EndofFile);
	ERROR_HANDLER("Increment Ref Count");

	if (!EndofFile) { // Object already exists (Increment the ref count)

		err = sm_RemoveEntry(&RefIID,BufferGroup,&Key,(void *)Elem);
		ERROR_HANDLER("Increment Ref Count");

		RefCount = *((int *)Elem);
		RefCount++;
		memcpy(Elem,&RefCount,4);
	} // if
	else {
		RefCount=1; Self=1;
		memcpy(Elem,&RefCount,4);
		memcpy(Elem+4,&Self,4);
	} // if-else

	err = sm_InsertEntry(&RefIID,BufferGroup,&Key,(void *)Elem);
	ERROR_HANDLER("Increment Ref Count");

	return TRUE;

} // IncRefCount


int DecRefCount(OID Oid, int *RefCount, int Self)
{
	int		err, OSelf;
	OID		RetOid;
	KEY		Key;
	SMCURSOR	Cursor;
	USERDESC	*UserDesc;
	char		CharOid[KEYLENGTH];
	char		Elem[8];
	BOOL		EndofFile;

	if (OID_IS_INVALID(Oid)) return TRUE;

	oidTOchar(&Oid,CharOid);
	Key.length = KEYLENGTH;
	Key.valuePtr = CharOid;

	// Check whether there is an entry...
	err = sm_FetchInit(&RefIID,BufferGroup,&Key,SM_EQ,&Key,SM_EQ,&Cursor);
	ERROR_HANDLER("Decrement Ref Count");

	err = sm_FetchNext(&Cursor,&Key,(void *)Elem,&EndofFile);
	ERROR_HANDLER("Decrement Ref Count");

	if (!EndofFile) { // Object should exist (Decrement the ref count)

		*RefCount = *((int *)Elem);
		OSelf	  = *((int *)(Elem+4));

		if (!OSelf && Self) {
			return TRUE; // cannot delete object twice...
		}

		err = sm_RemoveEntry(&RefIID,BufferGroup,&Key,(void *)Elem);
		ERROR_HANDLER("Decrement Ref Count");

		(*RefCount)--;

		if (*RefCount > 0) {

			if (OSelf && Self)
				OSelf = 0; 

			memcpy(Elem,(char *)RefCount,4);
			memcpy(Elem+4,(char *)&OSelf,4);

			err = sm_InsertEntry(&RefIID,BufferGroup,&Key,(void *)Elem);
			ERROR_HANDLER("Increment Ref Count");

		} // if
	}
	else {
// --->         SetError...
		return FALSE;
	} // if-else

	return TRUE;

} // DecRefCount


int PrintRefCounts()
{
	int		err, i, RefCount, Self;
	SMCURSOR	Cursor;
	USERDESC	*UserDesc;
	char		CharOid[KEYLENGTH];
	char		Elem[8];
	BOOL		EndofFile;

	err = sm_FetchInit(&RefIID,BufferGroup,SM_BOF,SM_EOF,&Cursor);
	ERROR_HANDLER("Print Ref Count");

	err = sm_FetchNext(&Cursor,NULL,(void *)Elem,&EndofFile);
	ERROR_HANDLER("Print Ref Count");

	i=1;
	while (!EndofFile) {

		RefCount = *((int *)Elem);
		Self     = *((int *)(Elem+4));

		//printf("Obj: %2d  Refcount: %d Self: %d\n",i++, RefCount, Self);

		err = sm_FetchNext(&Cursor,NULL,(void *)Elem,&EndofFile);
		ERROR_HANDLER("Print Ref Count");

	} // while

	return TRUE;

} // PrintRefCounts

