/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <X11/cursorfont.h>
#include <istream.h>
#include <ostream.h>
#include <stdio.h>
#include <Xm/MainW.h>
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Mrm/MrmAppl.h>
#include <Xm/Frame.h>
#include <Xm/PushBG.h>
#include <Xm/Separator.h>
#include <Xm/CascadeBG.h>
#include <Xm/DrawnB.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include "m_class.h"
#include "libMv.h"

typedef struct AttStr {
    char name[50][150];
    char type[50][150];
} Atts;

extern char             BitMapPath[100];
extern int              language;
extern int              attmode;
extern XmStringCharSet  charset;
extern int              myAttMax;
extern Atts             myAtt;
extern Widget           CreateLabeledTextForm();
char                    *acomps[] = {"Set", "List",  "Ref"};
char                    *aatom[]  = {"integer", "float", "character",
                                     "string", "boolean"};
extern char		UidPath[100];
extern char 		BitMapPath[100];
static char             *uid_file[1];
static char             *uid_file2[1];
static MrmHierarchy     Mrmdata;
static MrmType          *dummy_class;
static MrmRegisterArg   callbackvec[] = { {"penCB",(caddr_t)penCB} };
static MrmCount         callbacknum = sizeof(callbackvec) / sizeof(MrmRegisterArg);
GC                      gc;
Widget                  top_att, popup, type_t[16], pix[16], name_t[16], globpix;
int                     glob;
Pcrclst                 newclass;


static void penCB(Widget w, XtPointer client_data, XmListCallbackStruct *cbs);
Widget      make_frame(Widget w);
Widget      make_label(Widget w, char *title);   
Widget      make_rowcol(Widget w);
XtArgVal    GetPixel(char *colorstr);
void        redraw_button(Widget button, Pixmap pixmap,
                          XmDrawnButtonCallbackStruct *cbs);
static void do_mine(Widget w, int select, XmListCallbackStruct *cbs);
static void result(Widget text_w, int i, XmListCallbackStruct *cbs);
void        attCB(Widget w, XtPointer client_data, XmListCallbackStruct *cbs);


/*--------------------------------------------------------------------------------
** redraw_button -
*-------------------------------------------------------------------------------*/
void redraw_button(Widget button, Pixmap pixmap, XmDrawnButtonCallbackStruct *cbs)
{
    int       srcx, srcy, destx, desty, pix_w, pix_h, drawsize, border;
    Dimension bdr_w, w_width, w_height;
    short     hlthick, shthick;
    Window    root;

    /* get width and height of the pixmap. don't use srcx and root */
    XGetGeometry(XtDisplay(button), pixmap, &root, &srcx, &srcx,
        &pix_w, &pix_h, &srcx, &srcx);
 
    /* get the values of all the resources that affect the entire
     * geometry of the button.  Hint: we can't use XGetGeometry()
     */
    XtVaGetValues(button,
        XmNwidth,              &w_width,
        XmNheight,             &w_height,
        XmNborderWidth,        &bdr_w,
        XmNhighlightThickness, &hlthick,
        XmNshadowThickness,    &shthick,
        NULL);
 
    /* calculate available drawing area, width 1st */
    border = bdr_w + hlthick + shthick;
 
    /* if window is bigger than pixmap, center it; else clip pixmap */
    drawsize = w_width - 2*border;
    if (drawsize > pix_w) {
        srcx = 0;
        destx = (drawsize - pix_w)/2 + border;
    } else {
        srcx = (pix_w - drawsize)/2;
        pix_w = drawsize;
        destx = border;
    }
 
    /* now the height ... */
    drawsize = w_height - 2*border;
    if (drawsize > pix_h) {
        srcy = 0;
        desty = (drawsize - pix_h)/2 + border;
    } else {
        srcy = (pix_h - drawsize)/2;
        pix_h = drawsize;
        desty = border;
    }
 
    XCopyArea(XtDisplay(button), pixmap, cbs->window, gc,
        srcx, srcy, pix_w, pix_h, destx, desty);
}

/*--------------------------------------------------------------------------------
** attCB - Display class attributes 
*-------------------------------------------------------------------------------*/
void attCB(Widget w, XtPointer client_data, XmListCallbackStruct *cbs)
{  
   extern Widget drawing_area;
   Widget main_att, att_pane, att_main_rowcol, main_w, addprc, rowcol_w, rowcol_s;
   Widget name, mstype, pen_w, rowcol_p, scroll, form, d_label, text[16], frame;
   char   *choice;
   int    i, j, k, typeint, loopn;
   Pixmap on;
   Pixel  fg, bg;
   char   typebuf[30], valstr[150];
   char   okPath[100];
   void   mydialogue(Widget w);


   glob = 0;
   globpix = NULL;

   /* Get uil path */
   uid_file[0] = (char *) malloc(100);
   uid_file2[0] = (char *) malloc(100);
   strcpy( uid_file[0], UidPath);
   strcpy( uid_file2[0], UidPath); 
   if (language == ENGLISH)
     strcat( uid_file[0],"/attpic.uid");
   if (language == GERMAN)
     strcat( uid_file[0],"/attpic2.uid");
   strcat( uid_file2[0],"/attribute.uid");

   /* Prepare widgets */
   MrmInitialize();
   top_att = XtVaCreatePopupShell(NULL, topLevelShellWidgetClass, drawing_area,
          XmNallowShellResize, True,
          XmNdeleteResponse,   XmUNMAP,
          NULL);
   main_att = XtVaCreateManagedWidget("main_w", xmMainWindowWidgetClass,
          top_att, NULL);
   att_pane = XtVaCreateWidget("panemethod", xmPanedWindowWidgetClass,
          main_att, NULL);
   att_main_rowcol = XtVaCreateWidget("mainrowcol", xmRowColumnWidgetClass,
          att_pane, NULL);
   MrmOpenHierarchy( XtNumber(uid_file), uid_file, NULL, &Mrmdata );
   MrmRegisterNames (callbackvec, callbacknum);
   MrmFetchWidget(Mrmdata,"main_widg", att_main_rowcol, &main_w,
          (short int *)&dummy_class );
   XtManageChild(main_w);
   MrmCloseHierarchy(Mrmdata);
   addprc = XtVaCreateWidget("frame", xmFrameWidgetClass, att_pane , NULL);
   rowcol_p = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, addprc,
          XmNspacing, 5,
          XmNorientation, XmVERTICAL,
          NULL);
   form = XtVaCreateWidget(NULL, xmRowColumnWidgetClass, rowcol_p,
        XmNorientation, XmHORIZONTAL,
        NULL);
   if (language == ENGLISH)
             d_label =  make_label(form, "                FIELD NAME                             DATA TYPE                                                 DESCRIPTION");
   if (language == GERMAN)
             d_label =  make_label(form, "                 FELDNAME                               DATENTYP                                                 BESCHREIBUNG");
   XtManageChild(form);
   scroll = XtVaCreateWidget("myscroll", xmScrolledWindowWidgetClass, rowcol_p,
          XmNscrollBarDisplayPolicy, XmAS_NEEDED,
          XmNscrollingPolicy, XmAUTOMATIC,
          XmNwidth, 750,
          XmNheight, 238,
          NULL);
   rowcol_s = XtVaCreateWidget(NULL, xmRowColumnWidgetClass, scroll,
          XmNorientation, XmVERTICAL,
          NULL);
   XtVaGetValues(rowcol_s, XmNforeground, &fg, XmNbackground, &bg, NULL);
   sprintf(okPath,"%s/switch_on",BitMapPath);
   on = XmGetPixmap(XtScreen(rowcol_s), okPath,
          GetPixel("black"), bg);
   if (attmode == 0) // Display mode
       loopn = myAttMax;
   else loopn = 16;  // Creation mode
   for (i=0; i<loopn; i++) {
      form = XtVaCreateWidget(NULL, xmRowColumnWidgetClass, rowcol_s,
        XmNorientation,      XmHORIZONTAL,
        NULL);
      frame = XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
        XmNshadowType, XmSHADOW_OUT,
        XmNwidth, 25,
        XmNheight, 25,
        NULL);
      pix[i] = XtVaCreateWidget("ic", xmLabelGadgetClass, frame,
        XmNleftAttachment, XmATTACH_FORM,
        XmNlabelType,       XmPIXMAP,
        XmNlabelPixmap,     on,
        XmNalignment,       XmALIGNMENT_END,
        NULL);
      name_t[i] = XtVaCreateManagedWidget("text", xmTextWidgetClass, form,
        XmNx, 400,
        XmNy, 500,
        NULL);
      if (attmode == 0) { // Display mode
          strcpy(valstr, myAtt.name[i]);
          XtVaSetValues(name_t[i], XmNvalue, valstr, NULL);   
      }
      XtAddCallback(name_t[i], XmNfocusCallback, result, i);

      type_t[i] = XtVaCreateManagedWidget("text", xmTextWidgetClass, form,
        NULL);
      if (attmode == 0) { // Display mode
          strcpy(valstr, myAtt.type[i]);
          XtVaSetValues(type_t[i], XmNvalue, valstr, NULL);
      }
      XtAddCallback(type_t[i], XmNfocusCallback, result, i);
      MrmOpenHierarchy(XtNumber(uid_file2), uid_file2, NULL, &Mrmdata );
      MrmRegisterNames (callbackvec, callbacknum);
      MrmFetchWidget(Mrmdata,"pen_w", form, &pen_w, (short int *)&dummy_class);
      MrmCloseHierarchy(Mrmdata);
      XtManageChild(pen_w);
      text[i] = XtVaCreateManagedWidget("text", xmTextWidgetClass, form,
        XmNcolumns, 30,
        NULL);
      XtAddCallback(text[i], XmNfocusCallback, result, i);
      XtManageChild(form);
   }
  
   free(uid_file[0]);
   free(uid_file2[0]);

   XtManageChild(rowcol_s);
   XtManageChild(scroll);
   mydialogue(rowcol_p);
   XtManageChild(rowcol_p);
   XtManageChild(addprc);
   XtFree(choice);
   XtManageChild(att_main_rowcol);
   XtManageChild(att_pane);
   XtPopup(top_att, XtGrabNone);
}

/*--------------------------------------------------------------------------------
** penCB - Callback for pencil button 
*-------------------------------------------------------------------------------*/
static void penCB(Widget w, XtPointer client_data, XmListCallbackStruct *cbs)
{
  Widget buttong;
  int i;
  static void ChangeRet(Widget w, char *client_data,XmListCallbackStruct *cbs);
 
  popup = XmCreatePopupMenu(XtParent(w), "popup", NULL, 0);
  for (i=0; i<5; i++) {
     buttong = XtCreateManagedWidget(aatom[i], xmPushButtonGadgetClass,
                                     popup,NULL, 0);
     XtAddCallback (buttong, XmNactivateCallback, ChangeRet, aatom[i]);
  }
  buttong = XtCreateManagedWidget("xmSep",xmSeparatorWidgetClass,
                                  popup,NULL, 0);
  for (i=0; i<3; i++) {
     buttong = XtCreateManagedWidget(acomps[i], xmPushButtonGadgetClass,
                                     popup, NULL, 0);
     XtAddCallback (buttong, XmNactivateCallback, ChangeRet, acomps[i]);
  }
  XmMenuPosition(popup,(XButtonPressedEvent *)(cbs->event));
  XtManageChild(popup);
}

/*--------------------------------------------------------------------------------
** result - callback for focusing on type or name
*-------------------------------------------------------------------------------*/
static void result(Widget text_w, int i, XmListCallbackStruct *cbs) 
{
  if (globpix && (glob != i)) XtUnmanageChild(globpix);
  XtManageChild(pix[i]);
  glob = i;
  globpix = pix[i];
}
 
/*--------------------------------------------------------------------------------
** ChangeRet - Change type
*-------------------------------------------------------------------------------*/
static void ChangeRet(Widget w, char *client_data,XmListCallbackStruct *cbs)
{
  XtVaSetValues(type_t[glob], XmNvalue, client_data, NULL);
  XtDestroyWidget(popup);
}

/*--------------------------------------------------------------------------------
** mydialogue - Ok and Cancel buttons
*-------------------------------------------------------------------------------*/
void mydialogue(Widget w)
{
    Widget pb, pb2, rowcol, rowcol2, rowcol3, frame, frame2;
    Widget label, label2;
    Pixel fg, bg;
    Pixmap pixmap, pixmap2;
    char okPath[100];
 
    rowcol = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, w,
        XmNorientation, XmHORIZONTAL,
        XmNspacing, 5,
        NULL);
    XtVaGetValues(rowcol, XmNbackground, &fg, XmNbackground, &bg, NULL);
    gc = XCreateGC(XtDisplay(rowcol), RootWindowOfScreen(XtScreen(rowcol)),
        NULL, 0);
    XSetForeground(XtDisplay(rowcol), gc, fg);
    XSetBackground(XtDisplay(rowcol), gc, bg);
    sprintf(okPath,"%s/ok",BitMapPath);
    pixmap = XmGetPixmap(XtScreen(rowcol), okPath, GetPixel("blue"), bg);
    frame =make_frame(rowcol);
    rowcol2 = make_rowcol(frame);
    pb = XtVaCreateManagedWidget("Ok", xmDrawnButtonWidgetClass, rowcol2,
        XmNwidth,             55,
        XmNheight,            45,
        XmNpushButtonEnabled, True,
        XmNshadowType,        XmSHADOW_ETCHED_OUT,
        NULL);
    label = make_label(rowcol2, "   Ok!");
    XtAddCallback(pb, XmNactivateCallback, do_mine, 0);
    XtAddCallback(pb, XmNexposeCallback, redraw_button, pixmap);
    XtAddCallback(pb, XmNresizeCallback, redraw_button, pixmap);
    XtManageChild(rowcol2);
    sprintf(okPath,"%s/cancel",BitMapPath);
    pixmap2 = XmGetPixmap(XtScreen(rowcol), okPath,
        GetPixel("red"), bg);
    frame2 = make_frame(rowcol);
    rowcol3 = make_rowcol(frame2);
    pb2 = XtVaCreateManagedWidget("pb", xmDrawnButtonWidgetClass, rowcol3,
        XmNwidth,             55,
        XmNheight,            45,
        XmNpushButtonEnabled, True,
        XmNshadowType,        XmSHADOW_ETCHED_OUT,
        NULL);
    if (language == ENGLISH)
        label2 = make_label(rowcol3, " Cancel");
    if (language == GERMAN)
        label2 = make_label(rowcol3, "Abbr.");
    XtAddCallback(pb2, XmNactivateCallback, do_mine, 1);
    XtAddCallback(pb2, XmNexposeCallback, redraw_button, pixmap2);
    XtAddCallback(pb2, XmNresizeCallback, redraw_button, pixmap2);
    XtManageChild(rowcol3);
    XtManageChild(rowcol);
}

/*--------------------------------------------------------------------------------
** do_mine - get attributes
*-------------------------------------------------------------------------------*/
static void do_mine(Widget w, int select, XmListCallbackStruct *cbs)
{
  int i;
  char *str;

  strcpy(newclass->attls, "tuple( ");
  if ((select==0) & (attmode==1)) {
     str = (char *) XmTextGetString(name_t[0]);
     for (i=0; strcmp(str,"\0")!=0; i++){ 
          // Get attributes
          str = (char *) XmTextGetString(name_t[i]);
          strcat(newclass->attls, str);
          strcat(newclass->attls, " ");
          str = (char *) XmTextGetString(type_t[i]);
          strcat(newclass->attls, str);
          strcat(newclass->attls, ",");
          str = (char *) XmTextGetString(name_t[i+1]);
      } 
     newclass->attls[strlen(newclass->attls)-1] = '\0';
     strcat(newclass->attls, " )");
  }
  printf("%s \n", newclass->attls);
  XtDestroyWidget(top_att);
}
