/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <Xm/PushBG.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/DrawingA.h>
#include <Xm/SelectioB.h>
#include <Xm/Frame.h>
#include <X11/cursorfont.h>
#include "libMv.h"
#include "chb.h"
#include "wid.h"
#include "mood_catalog.h"
#include "userreq.h"
#include "m_class.h"

int             globargc;
int             language;
char		UidPath[100];
char 		BitMapPath[100];
extern XmStringCharSet charset;
extern Window          win;
extern Display         *dpy;

UserRequest	    *userRequest;
char                globdbname[100];
Widget              main_window,drawing_area;
vertex              **tree;
vertex              *ver;
Widget              **button;
char                fName[32], fathers[12][NAMELENGTH];
typedef struct mine *Pmine;

static void acceptdblist(Widget w,Widget client_data,XtPointer call_data);
static void canceldblist(Widget w,Widget client_data,XtPointer call_data);
void        main_browser();

/*-------------------------------------------------------------
** find_fathers -
*------------------------------------------------------------*/
int find_fathers(char *name)
{
  int i;

  for(i=0;fathers[i][0];i++)
      if(strcmp(fathers[i],name)==0)
         return 1;

  return 0;
}
/*-------------------------------------------------------------
** recreate_drawing_area -
*------------------------------------------------------------*/
void recreate_drawing_area()
{
   Arg       args[6];
   XmString  label;
   GC        gc;
   XGCValues gcv;
   int       n;

   XtUnmanageChild(drawing_area);

   n =0;

   XtSetArg(args[n],XmNwidth,1900);n++;
   XtSetArg(args[n],XmNheight,1900);n++;
   XtSetArg(args[n],XmNresizePolicy, XmNONE);n++;
   XtSetArg(args[n],XmNbackground,GetPixel("LightSlateBlue"));n++;
   XtSetArg(args[n],XmNforeground,GetPixel("red"));n++;

   drawing_area = XmCreateDrawingArea(main_window,"drawing_area",args,n);
   XtManageChild(drawing_area);

   /* When scrolled, the drawing area will get expose events
      add callback for all mouse and keyboard input events */

   XtAddCallback(drawing_area,XmNexposeCallback,(XtCallbackProc) draw_lines,NULL);
   gcv.foreground = BlackPixelOfScreen(XtScreen(drawing_area));
   gc = XCreateGC(XtDisplay(drawing_area),
        RootWindowOfScreen(XtScreen(drawing_area)),GCForeground, &gcv);

   XtVaSetValues(drawing_area, XmNuserData, gc, NULL);
}

/*-------------------------------------------------------------
** AddSubClassCB - Collect superclass list
*------------------------------------------------------------*/
void AddSubClassCB(Widget w,XtPointer client_data,XtPointer call_data)
{
   extern Widget     toplevel;
   vertex            *ver1;
   unsigned          wid = (unsigned) client_data;
   int               i, n;
   Arg               args[3];
   XmString          title_string, mes_string;
   XmString          def_value;
   Widget            prompt;
   XmStringCharSet   charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

   ver = find_vertex(wid);
 
  /* Get superclass list other than immediate superclass */
  if (language == ENGLISH) {
     title_string=XmStringCreateLtoR("Superclass List",charset);
     mes_string=XmStringCreateLtoR("Superclass List: ",charset);
  }
  if (language == GERMAN) {
     title_string=XmStringCreateLtoR("Liste der Superklassen",charset);
     mes_string=XmStringCreateLtoR("Liste der Superklassen: ",charset);
  }

  def_value=XmStringCreateLtoR(ver->name, charset);
  n = 0;
  XtSetArg (args[n], XmNdialogTitle, title_string);  n++;
  XtSetArg (args[n], XmNselectionLabelString, mes_string);  n++;
  XtSetArg (args[n], XmNtextString, def_value); n++;
  prompt = XmCreatePromptDialog (XtParent(w), "SuperClasses", args, n);
  XtManageChild (prompt);
  XtAddCallback (prompt, XmNokCallback,(XtCallbackProc)acceptdblist,client_data);
  XtAddCallback (prompt, XmNcancelCallback, (XtCallbackProc)canceldblist, NULL);
  n = 0;
  XtSetArg (args[n], XmNx, 120);  n++;
  XtSetArg (args[n], XmNy, 150);  n++;
  XtSetValues (XtParent(prompt), args, n);
  XmStringFree(title_string);
  XmStringFree(mes_string);
}

/*-------------------------------------------------------------
** acceptdblist - Displayy class template
*------------------------------------------------------------*/
static void acceptdblist(Widget w,Widget client_data,XtPointer call_data) 
{
  char  *supcllist;
  Pmine pass;
  extern void ClassCB(Widget w,Pmine client_data,XtPointer call_data);


  XmSelectionBoxCallbackStruct *scb = (XmSelectionBoxCallbackStruct *)call_data;
  XmStringGetLtoR(scb->value, charset, &supcllist);
  XtDestroyWidget(w);  
 
  /* class creation... */
  pass = (Pmine) malloc(sizeof(struct mine));
  pass->flag = 1; /* new class creation mode */
  strcpy(pass->clname, supcllist); /* superclass list instead of class name */
  ClassCB(w,pass,NULL);
}

/*-------------------------------------------------------------
** canceldblist -
*------------------------------------------------------------*/
static void canceldblist(Widget w,Widget client_data,XtPointer call_data)
{
  XtDestroyWidget(w);  
}

/*-------------------------------------------------------------
** main - main for browser
*------------------------------------------------------------*/
main(int argc,char *argv[])
{
   XtAppContext app;
   extern Widget toplevel;
   char sqlstm[150];
   char *ResPath;

   //if (argc < 3)  exit(0);
   globargc = argc;
 
   ResPath = getenv("MOOD_BIN");
   if (!ResPath) {
      printf("\nMOOD_BIN environment variable is not set !\n");
      exit(1);
   }
   sprintf(UidPath,"%s/gui/uil", ResPath);
   sprintf(BitMapPath,"%s/gui/bitmap", ResPath);
   userRequest = new UserRequest();
   userRequest->initESM(argv[0]);
   if (argc == 3) { // Db selected
      if (strcmp("german", argv[2]) == 0)
       language = GERMAN;
      else
       language = ENGLISH;
      strcpy(sqlstm, "use database ");
      strcat(sqlstm, argv[1]);
      userRequest->executeMSQL(sqlstm);
      strcpy(globdbname, argv[1]);
   
      init_screen(argc,argv,&app);

      if (argc>1) strcpy(fName,argv[1]);
      else strcpy(fName,"file");
   
      dag_algorithm();
      draw_onscreen();
   }

   if (argc == 2) { // No db selected
      if (strcmp("german", argv[1]) == 0)
       language = GERMAN;
      else
       language = ENGLISH;

      init_screen(argc,argv,&app);

      if (argc>1) strcpy(fName,argv[1]);
      else strcpy(fName,"file");

      dag_algorithm();
      globargc = 3; // Continue normally
   }

   XtRealizeWidget(toplevel);
   dpy = XtDisplay(drawing_area);
   win = XtWindow(drawing_area);

   ChangeCursor(toplevel, XC_hand2);

   XtAppMainLoop(app);
}

