/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include "mood_catalog.h"
#include "mood_parser.h"
#include <Xm/List.h>
#include <Xm/LabelG.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>
#include <Xm/MainW.h>
#include "libMv.h"
#include "m_class.h"
#include "chb.h"
#include "wid.h"
#include "IntrinsicP.h"

#define NUM 20
#define USE  0
#define DROP 1

extern int  language;
extern char globdbname[100];
extern UserRequest *userRequest;
extern Widget toplevel;
extern Display  *dpy;
extern Window    win;
ListItem *p;
extern Widget drawing_area;
int actiontype=0;

void recreate_drawing_area();
static void dblistproc(Widget list_w, XtPointer client_data, XmListCallbackStruct *cbs);


/*-------------------------------------------------------------
** listdb - 
*------------------------------------------------------------*/
void listdb(int action)
{
    Widget        toplist, rowcol, list_w, text_w, label;
    Widget        main_wid;
    XtAppContext  app;
    Arg           args[1];
    char          *DbNameList[NUM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    XmString      str_list[NUM];
    int j, i=0;

    actiontype = action;
   /*
    * CAT... Get database list from Kernel.
    */ 
    userRequest->executeMSQL("list database");
    memset(DbNameList, 0, sizeof(char *)*NUM);
    for (p = ((ListDatabase *)userRequest->operation)->dbList->firstItem; p; p=p->next){ 
         DbNameList[i] = (char *) malloc(20);
         sprintf(DbNameList[i++], "%s", (char *)p->item);
    }

   /*
    * Prepare widgets
    */ 
    toplist = XtVaCreatePopupShell(NULL, topLevelShellWidgetClass, drawing_area,
            XmNallowShellResize, True,
            XmNdeleteResponse,   XmUNMAP,
            NULL);
    main_wid = XtVaCreateManagedWidget("main_wid", xmMainWindowWidgetClass, toplist,
        NULL);
    rowcol = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, main_wid, NULL);
    if (language == ENGLISH)
        label = XtVaCreateManagedWidget("Databases:", xmLabelGadgetClass, rowcol,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        label = XtVaCreateManagedWidget("Datenbank:", xmLabelGadgetClass, rowcol,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    for (j = 0; j < i; j++)
        str_list[j] = (XmString)XmStringCreateLtoR(DbNameList[j], charset);
    list_w = XmCreateScrolledList(rowcol, "list", NULL, 0);
    XtVaSetValues(list_w,
      XmNitems,            str_list,
      XmNitemCount,        i,
      XmNvisibleItemCount, 7,
      XmNselectionPolicy,  XmEXTENDED_SELECT,
      NULL);
    XtAddCallback(list_w, XmNextendedSelectionCallback,(XtCallbackProc)dblistproc, NULL);
    XtAddCallback(list_w, XmNdefaultActionCallback,(XtCallbackProc)dblistproc, NULL);
    XtManageChild(list_w);
    for (j = 0; j < i; j++)
        XmStringFree(str_list[j]);
    XtFree((char *)str_list);
    XtManageChild(rowcol);
    XtPopup(toplist, XtGrabNone);
}

/*-------------------------------------------------------------
** dblistproc - Use or drop db
*------------------------------------------------------------*/
static void dblistproc(Widget list_w, XtPointer client_data, XmListCallbackStruct *cbs)
{
   char *DbName;
   char sqlstm[30];

   XmStringGetLtoR(cbs->selected_items[0], charset, &DbName);
   switch(actiontype) {
     case USE:
         strcpy(sqlstm, "use database ");
         break;
     case DROP:
         strcpy(sqlstm, "drop database ");
         break;
   }
   strcat(sqlstm, DbName);
  
  /*
   * SQL... Get database list from Kernel.
   */
   userRequest->executeMSQL(sqlstm);
   if ((userRequest->operationType == HANDLE_ERROR) && 
       (userRequest->userError.errorType != NO_ERROR)) {
        char *msg = userRequest->getErrorText();
        Display_Error(drawing_area, msg);
        free(msg);
    } else { /* Success */
        if (actiontype == DROP) {
            XmListDeleteItem(list_w, cbs->selected_items[0]);
            //char *msg = userRequest->getMessageText();
            //strcpy(msg, "Database dropped succesfully");
            //Display_Warning(drawing_area, msg);
            //free(msg);
         }

        /*
         * Refresh the screen
         */ 
         if (actiontype == USE) {
           strcpy(globdbname, DbName);
           dispose();
           dag_algorithm();
           recreate_drawing_area();
           draw_onscreen();
           XtRealizeWidget(toplevel);
           dpy = XtDisplay(drawing_area);
           win = XtWindow(drawing_area);
         }
    }
}
