/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "chb.h"
#include <Xm/PushB.h>
#include "mood_catalog.h"
#include "userreq.h"

extern int globargc;
extern UserRequest *userRequest;
extern char fName[30];

vertex   *TOKEN_TABLE[MAX_NO_OF_TOKENS];
unsigned NO_OF_TOKENS=0,MAX_LEVEL=0;

struct queue{
    unsigned item;
    struct queue *child;
};

struct queue *frontPointer,*rearPointer;

/*-------------------------------------------------------------
** error - Prints and error message
*---------------------------------------------------------*/
void error()
{
    printf("insufficient memory\n");
    exit(1);
}

/*-------------------------------------------------------------
 * function getTokens()
 * given a string it returns the token names in an array called tokens
 * which can contain up to 50(MAX_NO_OF_TOKENS) tokens
 * tokens are separated by commas.
 *------------------------------------------------------------*/
void getTokens(char *buf,char tokens[MAX_NO_OF_TOKENS][NAMELENGTH],unsigned *noOfTokens)
{
 char *s;

 (*noOfTokens)=0;
 if (!(s=strtok(buf,":"))) 
       return;

 strcpy(tokens[0],s);

while ((s=strtok(NULL,","))!=NULL)
 {
   (*noOfTokens)++;
   strcpy(tokens[(*noOfTokens)],s);
 }

  if ((s=strtok(buf,"\0"))!=NULL){
      (*noOfTokens)++;
      strcpy(tokens[(*noOfTokens)],s);
  }
}

/*-------------------------------------------------------------
 * function foundInTable()
 * if the token is in the TOKEN_TABLE, returns its id
 * otherwise returns 0.
 *------------------------------------------------------------*/
int foundInTable(char *token)
{ unsigned i;

  for (i=1;i<=NO_OF_TOKENS;i++)
    if (strcmp(TOKEN_TABLE[i]->name,token)==0)
       return(i);     /* if found return id */
  return(0);          /* if not found  return 0*/
}


/*-------------------------------------------------------------
 * function insertIntoTable()
 * inserts a new token to the TOKEN_TABLE
 *------------------------------------------------------------*/
void insertIntoTable(char *token)
{
  NO_OF_TOKENS++;

  if ((TOKEN_TABLE[NO_OF_TOKENS]=(vertex*) malloc(sizeof(vertex)))==NULL)
     error();

  strcpy(TOKEN_TABLE[NO_OF_TOKENS]->name,token);
  TOKEN_TABLE[NO_OF_TOKENS]->id=NO_OF_TOKENS;
  TOKEN_TABLE[NO_OF_TOKENS]->level=0;
  TOKEN_TABLE[NO_OF_TOKENS]->no_of_child=0;
  TOKEN_TABLE[NO_OF_TOKENS]->no_of_parent=0;
}

/*-------------------------------------------------------------
 * function parse()
 * opens the input stream, reads the statements separated by semicolons
 * and by calling getTokens(), foundInTable(), and insertIntoTable()
 * functions forms the TOKEN_TABLE
 *------------------------------------------------------------*/
void parse()
{
   FILE *stream;
   char buf[1024],tokens[MAX_NO_OF_TOKENS][NAMELENGTH];
   char ch;
   unsigned noOfTokens,i;
   unsigned parentNo;

   Hierarchy *mvgraph; // CAT...
   CatBOOLEAN mvflag; // CAT...
   Hierarchy *gr;
   LevelInfo *li;

   NO_OF_TOKENS = 0;
   MAX_LEVEL = 0;

  if ((stream = fopen("mvgr", "wt")) == NULL)
   {
     printf("Cannot open input file.%s\n",stream);
     exit(1);
   }
 
   if (globargc == 2) fprintf(stream, "%s:;","Ghost"); // No db selected
   else {
  /*
   * CAT... Get Schema From Kernel
   */
   userRequest->startXAct();
   userRequest->parseText();
   userRequest->handleOperation();
   MoodsTypes my;
   my.getClassHierarchy("MoodsRoot", &mvgraph, &mvflag);
   userRequest->commitXAct();

   if (!mvgraph)
      fprintf(stream, "%s:;","MoodsRoot");
   else {
      fprintf(stream, "%s:",mvgraph->thislevel->ClassName);
      li = mvgraph->thislevel->next;
      while(li) {
	      printf("%s\n", li->ClassName );
              fprintf(stream, "%s,",li->ClassName);
              li = li->next;
      }
      fprintf(stream, ";\n");
      gr=mvgraph->nextlevel;

      while(gr) {
            fprintf(stream,"%s:",gr->thislevel->ClassName);
            li = gr->thislevel->next;
            while(li) {
              fprintf(stream,"%s,",li->ClassName);
              li = li->next;
               }
            fprintf(stream, ";\n");
            gr=gr->nextlevel;
    }
   }
  } // db

   fclose(stream); 
                
   if ((stream = fopen("mvgr", "rt")) == NULL)
   {
     printf("Cannot open input file.%s\n",stream);
     exit(1);
   }

   while (!feof(stream))
   {
     i=0;

     do{
       fread(&ch,1,1,stream);
       if (ch!='\n') buf[i++]=ch;
     } while ((ch!=';') && (!feof(stream)));

     buf[i-1]=0;
    
     getTokens(buf,tokens,&noOfTokens);
    
  if (noOfTokens>0)
     {
	for (i=0;i<noOfTokens;i++)
	  if (!foundInTable(tokens[i]))
	      insertIntoTable(tokens[i]);

	parentNo=foundInTable(tokens[0]);
	if (noOfTokens>1)
	{
	   TOKEN_TABLE[parentNo]->child =
	       (unsigned *)  malloc((noOfTokens-1)*sizeof(unsigned));

	   if (TOKEN_TABLE[parentNo]->child == NULL)
	       error();

	   for (i=1;i<noOfTokens;i++)
	   {
	      TOKEN_TABLE[parentNo]->child[i-1]=foundInTable(tokens[i]);
	      TOKEN_TABLE[parentNo]->no_of_child++;
	   }
	 } /* if    */
     }     /* if    */
   }       /* while */

   fclose(stream);
}


/*-------------------------------------------------------------
 * function reArrange()
 * rearranges the TOKEN_TABLE so that the first entry is a dummy token
 * whose children is the level 1 tokens.
 *-----------------------------------------------------------*/
void reArrange()
{
  unsigned i,j,k,found,counter=0;
  unsigned tokens[MAX_NO_OF_TOKENS];

  if ((TOKEN_TABLE[0]=(vertex*) malloc(sizeof(vertex)))==NULL)
      error();

  for (i=1;i<=NO_OF_TOKENS;i++){
    found=0;
    for (j=1;(j<=NO_OF_TOKENS) && !found;j++)
      for (k=0;k<TOKEN_TABLE[j]->no_of_child;k++)
	if (TOKEN_TABLE[j]->child[k]==i) found=1;
    if (!found)
       tokens[counter++]=i;
  }

  TOKEN_TABLE[0]->child=(unsigned *) malloc(counter*sizeof(unsigned));
  if (TOKEN_TABLE[0]->child == NULL)
     error();

  for (i=0;i<counter;i++)
     TOKEN_TABLE[0]->child[i]=tokens[i];

  TOKEN_TABLE[0]->no_of_child=counter;
  TOKEN_TABLE[0]->no_of_parent=0;
  TOKEN_TABLE[0]->level=0;
  TOKEN_TABLE[0]->id=0;
  strcpy(TOKEN_TABLE[0]->name,"DUMMY NAME");
}

/*-------------------------------------------------------------
 * function insertQ()
 * inserts a new item to the queue called from function BFS()
 *------------------------------------------------------------*/
void insertQ(unsigned v)
{
   struct queue *p=(struct queue *) malloc(sizeof(struct queue));
   p->item=v;
   p->child=NULL;
   rearPointer->child=p;
   rearPointer=rearPointer->child;
}

/*-------------------------------------------------------------
 * function removeQ()
 * removes an item from the queue called from function BFS()
 *------------------------------------------------------------*/
unsigned removeQ()
{
  struct queue *p=frontPointer;
  frontPointer=p->child;
  free(p);
  return(frontPointer->item);
}

/*-------------------------------------------------------------
 * function EmptyQ()
 * tests whether the queue is empty, called from function BFS()
 *------------------------------------------------------------*/
int emptyQ()
{
  return(frontPointer==rearPointer ? 1:0);
}

/*-------------------------------------------------------------
 * function foundInQ()
 * returns 1 if the item is already in queue otherwise returns 0
 -------------------------------------------------------------*/
int foundInQ(unsigned no)
{
  struct queue *p=frontPointer;

  while (p!=rearPointer)
  {
    if (p->item==no)
	return 1;           /* found */
    p=p->child;
  }
  return 0;   /* not found */
}

/*------------------------------------------------------------
 * function BFS()
 * finds the levels of each entry in the TOKEN_TABLE
 *------------------------------------------------------------*/
void BFS()
{ unsigned FLAG=50000;
  unsigned counter=0,i;
  unsigned v=0;
  unsigned po=TOKEN_TABLE[v]->no_of_child;
  unsigned o=0,t=0;

 frontPointer=rearPointer=(struct queue *) malloc(sizeof(struct queue));

  insertQ(FLAG);
  insertQ(v);
  insertQ(FLAG);

  while (!emptyQ())
  {
    v=removeQ();
      if (v==FLAG) {counter++;continue;}
    for (i=0;i<TOKEN_TABLE[v]->no_of_child;i++)
    {
      o++;
      t+=TOKEN_TABLE[TOKEN_TABLE[v]->child[i]]->no_of_child;
      if (TOKEN_TABLE[TOKEN_TABLE[v]->child[i]]->level<counter)
          TOKEN_TABLE[TOKEN_TABLE[v]->child[i]]->level=counter;

          insertQ(TOKEN_TABLE[v]->child[i]);
        if (o==po)
        {
            insertQ(FLAG);
            o=0;
            po=t;
            t=0;
        }

    }  /* for   */
  }    /* while */

  MAX_LEVEL=counter-1;
  free(frontPointer);
  free(rearPointer);
}

/*------------------------------------------------------------
 * function dummynodes()
 *
 *-----------------------------------------------------------*/
void dummynodes()
{
  unsigned i,j,no,no_of_child;

  for (i=1;i<=NO_OF_TOKENS;i++)
  {
    no_of_child=TOKEN_TABLE[i]->no_of_child;
    for (j=0;j<no_of_child;j++)
    {
      no = TOKEN_TABLE[i]->child[j];
      if ((TOKEN_TABLE[no]->level - TOKEN_TABLE[i]->level) > 1)
      {
	insertIntoTable("du");
	TOKEN_TABLE[NO_OF_TOKENS]->level = TOKEN_TABLE[i]->level+1;
	TOKEN_TABLE[NO_OF_TOKENS]->child=(unsigned *) malloc(sizeof(unsigned));
	TOKEN_TABLE[NO_OF_TOKENS]->child[0]=no;
	TOKEN_TABLE[NO_OF_TOKENS]->no_of_child = 1;
	TOKEN_TABLE[NO_OF_TOKENS]->parent=(unsigned *) malloc(sizeof(unsigned));
	TOKEN_TABLE[NO_OF_TOKENS]->parent[0]=TOKEN_TABLE[i]->child[j];
	TOKEN_TABLE[NO_OF_TOKENS]->no_of_parent = 1;

	TOKEN_TABLE[i]->child[j] = NO_OF_TOKENS;
      }
    }
  }
}

/*------------------------------------------------------------
 * function parents()
 *
 *-----------------------------------------------------------*/
void parents()
{
  typedef struct {
     unsigned parents[MAX_NO_OF_TOKENS];
     unsigned noOfParents;
   } fat;

  unsigned i,j,no;

  fat *fathers=(fat *) malloc(NO_OF_TOKENS*sizeof(fat));
  if (fathers==NULL)
     error();

  /** initialize the array **/

  for (i=0;i<NO_OF_TOKENS;i++)
     fathers[i].noOfParents=0;

  /** determine the parents of the vertexes **/

  for (i=1;i<=NO_OF_TOKENS;i++)
    for (j=0;j<TOKEN_TABLE[i]->no_of_child;j++)
    {
       no=TOKEN_TABLE[i]->child[j]-1;
       fathers[no].parents[fathers[no].noOfParents++]=i;
    }

  for (i=0;i<NO_OF_TOKENS;i++)
    if (fathers[i].noOfParents>0)
    {
       TOKEN_TABLE[i+1]->parent =
	     (unsigned *) malloc(fathers[i].noOfParents*sizeof(unsigned));
       if (TOKEN_TABLE[i+1]->parent == NULL)
	  error();
       TOKEN_TABLE[i+1]->no_of_parent = fathers[i].noOfParents;
       for (j=0;j<fathers[i].noOfParents;j++)
	 TOKEN_TABLE[i+1]->parent[j] = fathers[i].parents[j];
    }
}

/*------------------------------------------------------------
 * function levels()
 *
 *------------------------------------------------------------*/
int levels()
{
 extern vertex **tree;
 extern Widget **button;
 unsigned i,j,counter;

 reArrange();
 BFS();
 dummynodes();
 parents();

 tree=(vertex **) malloc((MAX_LEVEL+2)*sizeof(vertex *));
 button=(Widget **) malloc((MAX_LEVEL+2)*sizeof(Widget *));

 for (i=1;i<=MAX_LEVEL;i++)
 {
   counter=0;
   for (j=1;j<=NO_OF_TOKENS;j++)
     if (TOKEN_TABLE[j]->level==i)
       counter++;

   tree[i-1] = (vertex *) malloc((counter+1)*sizeof(vertex));
   button[i-1] = (Widget *) malloc((counter+1)*sizeof(Widget));

   if (tree[i-1]==NULL)
      error();

   counter=0;
   for (j=1;j<=NO_OF_TOKENS;j++)
      if (TOKEN_TABLE[j]->level==i)
      {
	strcpy(tree[i-1][counter].name,TOKEN_TABLE[j]->name);
	tree[i-1][counter].id=TOKEN_TABLE[j]->id;
	tree[i-1][counter].level=TOKEN_TABLE[j]->level-1;
	tree[i-1][counter].no_of_child=TOKEN_TABLE[j]->no_of_child;
	tree[i-1][counter].no_of_parent=TOKEN_TABLE[j]->no_of_parent;
	tree[i-1][counter].child=TOKEN_TABLE[j]->child;
	tree[i-1][counter].parent=TOKEN_TABLE[j]->parent;
	counter++;
      }

   tree[i-1][counter].id=0;
 }
  tree[MAX_LEVEL]=(vertex *)malloc(2*sizeof(vertex));
  tree[MAX_LEVEL][0].id=0;
  button[MAX_LEVEL]=(Widget *)malloc(2*sizeof(Widget));

  for(i=0;tree[i][0].id!=0;i++)
  {
    for(j=0;tree[i][j].id!=0;j++)
    {
      tree[i][j].baryUp=0;
      tree[i][j].baryDown=0;
    }
  }

 /* remove the token table from memory */
  for(i=0; TOKEN_TABLE[i][0].id;i++)
     { j = 0;
        do{
            if(TOKEN_TABLE[i]+j)
                   free(TOKEN_TABLE[i]+j);
           } while(TOKEN_TABLE[i][j++].id);
                  if(TOKEN_TABLE+i)
                     free(TOKEN_TABLE+i);
     }


 return(0);
}
