/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <Xm/PushB.h>
#include "userreq.h"
#include "chb.h"
#include "wid.h"
#include "mood_catalog.h"
#include "libMv.h"

extern FID CatalogID;

extern vertex **tree;
extern Widget **button;
extern Display *dpy;
extern Window  win;
extern Widget drawing_area;
extern UserRequest *userRequest;

/*-------------------------------------------------------------
** dag_algorithm - 
*------------------------------------------------------------*/
void dag_algorithm()
{
   parse();
   levels();
   median();
   position();
}


/*-------------------------------------------------------------
** clearCB - Clears drawing area
*------------------------------------------------------------*/
void clearCB (Widget widget,XtPointer data,XmDrawingAreaCallbackStruct *cbs)
{
    XEvent *event = cbs->event;

    XClearWindow(event->xany.display,XtWindow(drawing_area));
}


/*-------------------------------------------------------------
** dispose
*------------------------------------------------------------*/
void dispose()
{
int i,j;

   for(i=0; tree[i][0].id;i++)
     { j = 0;
        do{
            if(tree[i]+j)
                   free(tree[i]+j);
           } while(tree[i][j++].id);
                  if(tree+i)
                     free(tree+i);
     }
}

/*-------------------------------------------------------------
** destroy_buttons - Destroys the all of button widgets 
*------------------------------------------------------------*/
void destroy_buttons()
{
int i,j,cnt;

  for(i=0; tree[i][0].id;i++)
    for(j=0;tree[i][j].id;j++)
      if(tree[i][j].id)
         XtDestroyWidget(XtParent(button[i][j]));
}

/*-------------------------------------------------------------
** find_vertex - Finds the given widget id's vertex
*------------------------------------------------------------*/
vertex *find_vertex(unsigned wid)
{
  int i,j;

  for(i=0;tree[i][0].id;i++){
      for(j=0;tree[i][j].id;j++){
          if(wid==tree[i][j].wid)
             return &tree[i][j];
          }
  }

  return NULL;
}

/*-------------------------------------------------------------
** find_vertex_id - Finds the given Id's corresponding vertex
*------------------------------------------------------------*/
vertex *find_vertex_id(unsigned id)
{
int i,j;

 for(i=0;tree[i][0].id;i++){
      for(j=0;tree[i][j].id;j++){
       if(id==tree[i][j].id)
          return &tree[i][j];
      }
    }
return NULL;
}

/*-------------------------------------------------------------
** del_child - Delete the childs of the given vertex
*------------------------------------------------------------*/
void del_child(vertex *ver)
{
  unsigned i,id;

  /* Decrease the number of child of the dropped vertex`s parent */
   for( i=0;i<ver->no_of_parent;i++)
      { 
       int j = 0,k = 0;
       vertex *ver1;

          id = ver->parent[i];
          ver1 = find_vertex_id(id);
          ver1->no_of_child--;

              for(j=0;ver1->child[j]!=ver->id;j++);
                 for(k=j;k<ver1->no_of_child;k++)
                     ver1->child[k] = ver1->child[k+1];
                 free((void*)ver1->child[k]);

              while (strcmp(ver1->name,"du")==0)
              /* If parent ofthis class is dummy , 
                 then erase all dummy nodes recursively */
                 {
                  vertex *ver2;

                  id = ver1->parent[0];
                  ver2 = find_vertex_id(id);
                  ver2->no_of_child--;

                  for(j=0;ver2->child[j]!=ver1->id;j++);
                      for(k=j;k<ver2->no_of_child;k++)
                          ver2->child[k] = ver2->child[k+1];

                  free((void*)ver2->child[k]);
                  ver1=ver2;
                 }
         }
}

/*-------------------------------------------------------------
** del_button - Deletes given id's  
*------------------------------------------------------------*/
void del_button(unsigned Id)
{
  int i,j,ism;
  char sqlstm[100];

  printf("Delete Classes:\n");
  for(i=0;tree[i][0].id;i++)
      for(j=0;tree[i][j].id;j++)
          if (Id == tree[i][j].id)
              if(XtParent(button[i][j])){
                   printf("%s\n",tree[i][j].name);
                   // Delete classes
                   strcpy(sqlstm, "drop class ");
                   strcat(sqlstm, tree[i][j].name); 
                   userRequest->executeMSQL(sqlstm);
                   if ((userRequest->operationType == HANDLE_ERROR) &&
                       (userRequest->userError.errorType != NO_ERROR)) {
                       char *msg = userRequest->getErrorText();
                       Display_Error(drawing_area, msg);
                       free(msg);
                   } else {
                       del_child(&tree[i][j]);
                       XtDestroyWidget(XtParent(button[i][j]));
                       //char *msg = userRequest->getMessageText();
                       //Display_Warning(drawing_area, msg);
                       //free(msg);
                   }
                   userRequest->reset();
               }
}


/*-------------------------------------------------------------
**  coord
*------------------------------------------------------------*/
void coord(vertex *a,unsigned id,int *x,int *y)
{
   int i=0;
   for(;a[i].id != id;i++);

   *x =  a[i].x+strlen(a[i].name)*3+10;

   if(strcmp( a[i].name,"du"))
               *y =  a[i].y;
            else
               *y =  a[i].y + 16;
}


/*-------------------------------------------------------------
** draw_lines - draws lines between buttons 
*------------------------------------------------------------*/
void draw_lines()
{
    int i,j,k;
    int x,y,x1,y1,x2;
    GC gc;

    XtVaGetValues(drawing_area, XmNuserData, &gc, NULL);
    XSetForeground(dpy,gc,GetPixel("black"));

    for(i=0;tree[i][0].id;i++)
       for(j=0;tree[i][j].id;j++)
          {
            x = tree[i][j].x+strlen(tree[i][j].name)*3+10;
            if(strcmp(tree[i][j].name,"du")){
               y = tree[i][j].y+33; 
            } 
            else{
               y = tree[i][j].y + 16;
            }

               for(k=0;k<tree[i][j].no_of_child;k++)
                  {
                     coord(tree[i+1],tree[i][j].child[k],&x1,&y1);
                     XDrawLine(dpy,win,gc,x,y,x1,y1);
                   }
           }
}

void update_chb()
{
  int i,j,k,l,cnt,found=0;
  vertex *ver;
  char new_name[NAMELENGTH];
  extern char fName[32];
  FILE *f;

  if ((f = fopen("f1", "wt")) == NULL)
   {
     printf("Cannot open input file.f1\n");
     exit(1);
   }

     
    for(i=0; tree[i][0].id;i++)
      for(j=0;tree[i][j].id;j++){
        if(strcmp(tree[i][j].name,"du")!=0)
         {
          found = find_fathers(tree[i][j].name);
             if(found &&(!tree[i][j].no_of_child))
                 fprintf(f,"%s:%s;\n",tree[i][j].name,new_name);

              if(tree[i][j].no_of_child){
                 fprintf(f,"%s:",tree[i][j].name);

                   for(k=0;k<tree[i][j].no_of_child;k++){
                       cnt = tree[i][j].child[k];
                       ver = find_vertex_id(cnt);
                     
                         while(strcmp(ver->name,"du")==0){
                            cnt=ver->child[0];
                            ver = find_vertex_id(cnt);
                         } 

                         if(k!=tree[i][j].no_of_child-1)
                            fprintf(f,"%s,",ver->name);
                         else{
                            if (found)
                               fprintf(f,"%s,%s;\n",ver->name,new_name);
                            else
                               fprintf(f,"%s;\n",ver->name);
                            }
                   } 
               }
        }
      }
   
   fflush(f);
   fclose(f);
   strcpy(fName,"f1");
}
