/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <Xm/Text.h>
#include <Xm/SelectioB.h>
#include "me.h"

extern Widget new_dialog;
extern Widget menu_pane;

/*-------------------------------------------------------------
**  OpenFile
**  Open the present file.  Returns true if file
**  exists and open is successful.
*------------------------------------------------------------*/
Boolean OpenFile()
{
   struct stat statbuf; // Information on a file.
   int file_length;    
   char * file_string;  // Contents of file.   
   FILE *fp = NULL;     // Pointer to open file

   if ((fp = fopen(filename, "r+")) == NULL)
       if ((fp = fopen(filename, "r")) != NULL) {
            fprintf(stderr, "Warning: file opened read only.\n");
    } else {
      return(False);
   }

   if (stat(filename, &statbuf) == 0)
   file_length = statbuf.st_size;
   else
   file_length = 1000000; // Arbitrary file length

   // Read the file string
   file_string = (char *) XtMalloc((unsigned)file_length);
   fread(file_string, sizeof(char), file_length, fp);

   // Close the file 
   if (fclose(fp) != NULL) 
        fprintf(stderr, "Warning: unable to close file.\n");

   // Added the file string to the text widget
   XmTextSetString(text, file_string);

   file_saved = True; // Intialize to True 

   // Make appropriate item sensitive 
   XtSetSensitive(text, True);
   XtSetSensitive(cut_button, True);
   XtSetSensitive(copy_button, True);
   XtSetSensitive(copy_by_name_button, True);
   XtSetSensitive(clear_button, True);

   return(True);
}

void SaveFileCB (Widget w, XtPointer client_data, XtPointer call_data)
{
  FILE *tfp;        
  Arg args[5];
  XmString label_string,title_string;
  unsigned char * file_string;

  if (filename != NULL) {
         XmStringFree((XmString)filename);
         filename = NULL;
      }
      {
         XmSelectionBoxCallbackStruct *scb =
         (XmSelectionBoxCallbackStruct *) call_data;

         // Get the filename string from the file name prompt box 
         XmStringGetLtoR(scb->value, charset, &filename);

         // Open file if it exists, if not set items sensitive 
     if (!OpenFile()) {
              // make appropriate item sensitive
              XtSetSensitive(text, True);
              XtSetSensitive(cut_button, True);
              XtSetSensitive(copy_button, True);
              XtSetSensitive(copy_by_name_button, True);
              XtSetSensitive(clear_button, True);
         }

    // Popdown the file selection box 
           XtUnmanageChild (new_dialog);
 
    // Get the text string
    file_string = (unsigned char *)XmTextGetString(text);

  if ((tfp = fopen(filename, "w")) == NULL) {
       fprintf(stderr, "Warning: unable to open file, text not saved.\n");
    }

    // Write to file 
    fwrite(file_string, sizeof(char), strlen((char *)file_string) + 1, tfp);

    // Flush and close the temp file 
    if (fflush(tfp) != NULL) fprintf(stderr,"Warning: unable to flush file.\n");
    if (fclose(tfp) != NULL) fprintf(stderr,"Warning: unable to close file.\n");

    file_saved = True;
 }
}

/*-------------------------------------------------------------
**  SaveFile
**    Save the present file.
*------------------------------------------------------------*/
Boolean SaveFile()
{
  FILE *tfp;         
  Arg args[5];
  XmString label_string,title_string;
  unsigned char * file_string;
  int n;

  if (!filename) {
      label_string = XmStringCreateLtoR ("Enter Name of Query:", charset);
      title_string=XmStringCreateLtoR("Save Query",charset);
      n = 0;
      XtSetArg (args[n], XmNdialogTitle, title_string);  n++;
      XtSetArg(args[n], XmNselectionLabelString, label_string); n++;
      new_dialog = XmCreatePromptDialog(menu_pane,
         "new file dialog", args, n);
      XtManageChild(new_dialog);
  XtAddCallback (new_dialog, XmNokCallback,SaveFileCB, (XtPointer) DIALOG_NEW);
  XtAddCallback (new_dialog, XmNcancelCallback,DialogCancelCB, (XtPointer) DIALOG_NEW);
  XtAddCallback (new_dialog, XmNhelpCallback,DialogHelpCB,  (XtPointer)NEW_HELP);
  XmStringFree ((XmString)label_string);
      }
 else{
    if ((tfp = fopen(filename, "w")) == NULL) {
       fprintf(stderr, "Warning: unable to open file, text not saved.\n");
       return(False);;
    }

    // Get the text string
    file_string = (unsigned char *)XmTextGetString(text);

    // Write to a file
    fwrite(file_string, sizeof(char), strlen((char *)file_string) + 1, tfp);

    // Flush and close the file
    if (fflush(tfp) != NULL) fprintf(stderr,"Warning: unable to flush file.\n");
    if (fclose(tfp) != NULL) fprintf(stderr,"Warning: unable to close file.\n");

    file_saved = True;

    return(True);
}
  return True;
}
/*-------------------------------------------------------------
**      CloseFile
*------------------------------------------------------------*/
void CloseFile()
{
    /* Zero out the text string in the text widget.
       caution: is causes a value changed callack. */
    XmTextSetString(text, "");

    file_saved = True; /* reinitialize file_saved flag */

    // Free the file name
    if (filename != NULL) {
  XmStringFree((XmString)filename);
        filename = NULL;
    }

    // Set text to insensitive
    XtSetSensitive(text, False);
}

/*-------------------------------------------------------------
**  FileChangedCB
**    Process callback from Text.
*/
void FileChangedCB (Widget w, XtPointer client_data,XtPointer call_data)
{
    file_saved = False;
}

