/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/PushB.h>
#include <Xm/BulletinB.h>
#include <Xm/MessageB.h>
#include <Xm/Frame.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/DrawnB.h>
#include <Xm/LabelG.h>
#include <X11/cursorfont.h>
#include <mood_catalog.h>
#include <smex.h>
#include <userreq.h>
#include "libMv.h"
#include "opt_defs.h"
#include "dbi_gen.h"
#include "opt.h"
#include "dbi.h"
#include "me.h"
#include "mood_alg.h"

int         language;  
static char *uid_filename[1];
static MrmHierarchy Mrmdata;
static MrmType *dummy_class;
GC gc;
Widget fn_parent;
extern MoodsError Err;
char UidPath[100];

/*-------------------------------------------------------------
**  Global Variables
*------------------------------------------------------------*/

char *filename = NULL;       /* String containing file name      */
char *text_buffer;           /* Buffer for text widget text value*/
char text_query[5000];
Boolean by_name = False;     /* Copy by name flag                */
Boolean file_saved = True;   /* Indicates that the present file is saved */
    
FID CatalogID;

Widget toplevel;             /*  Application shell               */
Widget text;                 /* Multi-line text widget           */
Widget cut_button;           /* Clipboard cut button             */
Widget copy_button;          /* Clipboard copy button            */
Widget copy_by_name_button;  /* Clipboard copy by name button    */
Widget paste_button;         /* Clipboard paste button           */
Widget paste_by_name_button; /* Clipboard paste by name button   */
Widget clear_button;         /* Clipboard clear button           */
XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

int      fn_main(BULK_SCHEMA *bs, Widget Parentw);

/*-------------------------------------------------------------
** Display_Error -
*------------------------------------------------------------*/
void Display_Error(char *err_msg)
{
  Widget Err;
  Arg args[3];
  XmString title, message;
  int n;
   
  title = XmStringCreateLtoR("MOODSQL Error",charset);
  message = XmStringCreateLtoR(err_msg,charset);

  n = 0;
  XtSetArg(args[n],XmNdialogTitle,title);n++;
  XtSetArg(args[n],XmNmessageString,message);n++;
  Err = XmCreateErrorDialog(toplevel,"ERROR",args,n);
  XtManageChild(Err);
}

/*-------------------------------------------------------------
** Display_Warning -
*------------------------------------------------------------*/
void Display_Warning(char *warn_msg)
{
  Widget Warn;
  Arg args[3];
  XmString title, message;
  int n;
  
  title = XmStringCreateLtoR("MOODSQL Warning",charset);
  message = XmStringCreateLtoR(warn_msg,charset);

  n = 0;
  XtSetArg(args[n],XmNdialogTitle,title);n++;
  XtSetArg(args[n],XmNmessageString,message);n++;
  Warn = XmCreateWarningDialog(toplevel,"WARN",args,n);
  XtManageChild(Warn);
}

UserRequest *userRequest = NULL;
int             User;
extern int PutObject(OID *Oid, char *buf, int ObjSize);
extern int GetObject(OID Oid, char **buf, int *ObjSize);
extern "C" {
        int ShowGIF(unsigned char *, int );
}

/*-------------------------------------------------------------
** AcceptQueryCB - Accept the entered Query
*------------------------------------------------------------*/
void AcceptQueryCB (Widget w,Widget client_data,XtPointer call_data)
{
    int     reqStat;
    char    *reqMessage;

    ChangeCursor(toplevel, XC_gumby);

    strcpy(text_query , (char *)XmTextGetString(text));

    myalloc = fopen( "allocate.dat", "w" );
    dealloc = fopen( "deallocate.dat", "w" );
    fclose( myalloc );
    fclose( dealloc );

    userRequest->commitXAct();
    reqStat = userRequest->execute_MSQL(text_query);
    switch( reqStat )
    {
            case MS_ERROR :
                    reqMessage = userRequest->getErrorText();
                    Display_Error(reqMessage);
                    free(reqMessage);
                    break;

            case MS_WARNING :
                    Display_Warning("WARNING");
                    break;

            case MS_MESSAGE :
                    reqMessage = userRequest->getMessageText();
                    Display_Warning(reqMessage);
                    free(reqMessage);
                    break;

            case MS_NOERROR :
                    switch( userRequest->operationType )
                    {
                            case SELECT_STMT:
                                    BULK_SCHEMA *bs;
                                    bs =  userRequest->getBulkSchema();
                                    fn_parent=XtParent(w);
                                    if (!fn_main(bs, w)) {
                                       reqMessage = userRequest->getErrorText();
                                       Display_Error(reqMessage);
                                       free(reqMessage);
                                    }
                                    break;
                            case EXIT_FROM_MOODS :
                                    exit(1);
                    }
    }

   ChangeCursor(toplevel, XC_hand2);
}

/*-------------------------------------------------------------
**  ClearCB- Clears the entered Query
*------------------------------------------------------------*/
void ClearCB (Widget w,Widget client_data, XmDrawingAreaCallbackStruct *cbs)
{
    XmTextSetString(text,"");

    if (filename) {
       XmStringFree((XmString)filename);
       filename = NULL;
    }
    text_query[0] = '\0';
}

/*-------------------------------------------------------------
**  CreateText
*------------------------------------------------------------*/
static void CreateText (Widget parent)
{
  Arg    args[7]; 
  register int  n;
  Widget frame;

  frame = XtVaCreateManagedWidget("frame",
        xmFrameWidgetClass, parent,
        NULL);

  // Create text widget 
  n = 0;
  XtSetArg (args[n], XmNrows, 15);  n++;
  XtSetArg (args[n], XmNcolumns, 50);  n++;
  XtSetArg (args[n], XmNscrollVertical, True);  n++;
  XtSetArg (args[n], XmNscrollHorizontal, True);  n++;
  XtSetArg (args[n], XmNbackground,GetPixel("lightsteelblue"));n++;
  XtSetArg (args[n], XmNeditMode, XmMULTI_LINE_EDIT);  n++;
  text = XmCreateScrolledText (frame, "text", args, n);

  XtManageChild(text);
  /* Add value changed callback */
  XtAddCallback (text, XmNmodifyVerifyCallback, FileChangedCB, NULL);

  /* make appropriate item sensitive */
  XtSetSensitive(text, True);
  XtSetSensitive(cut_button, True);
  XtSetSensitive(copy_button, True);
  XtSetSensitive(copy_by_name_button, True);
  XtSetSensitive(clear_button, True);
}

/*-------------------------------------------------------------
**  CreateBBoard - creates BBoard in Main Window
*------------------------------------------------------------*/
Widget  CreateBBoard(Widget BBoard,Widget  parent)
{
  int n;
  Arg args[10];
  Widget OkButton;
  Widget ClearButton;
  

  n = 0;
  XtSetArg (args[n], XmNallowShellResize, True); n++;
  BBoard = XmCreateBulletinBoard(parent,"BBoard",args,n);

  n = 0;
  XtSetArg(args[n],XmNbackground,GetPixel("LightBlue"));n++;
  XtSetArg(args[n],XmNforeground,GetPixel("red"));n++;
  XtSetArg(args[n],XmNhighlightColor,GetPixel("red"));n++;
  XtSetArg(args[n],XmNshadowThickness,3);n++;
  XtSetArg(args[n],XmNborderWidth,1);n++;
  XtSetArg(args[n],XmNx,160);n++;
  XtSetArg(args[n],XmNy,320);n++;
  XtSetArg(args[n],XmNheight,30);n++;
  XtSetArg(args[n],XmNwidth,65);n++;
  if (language == ENGLISH)
     OkButton = XtCreateManagedWidget("Execute",xmPushButtonWidgetClass,BBoard,args ,n);
  if (language == GERMAN)
     OkButton = XtCreateManagedWidget("Ausfuehren",xmPushButtonWidgetClass,BBoard,args ,n);

  XtAddCallback (OkButton,XmNactivateCallback, (XtCallbackProc)AcceptQueryCB,NULL);

  n = 0;
  XtSetArg(args[n],XmNbackground,GetPixel("LightBlue"));n++;
  XtSetArg(args[n],XmNforeground,GetPixel("red"));n++;
  XtSetArg(args[n],XmNhighlightColor,GetPixel("red"));n++;
  XtSetArg(args[n],XmNshadowThickness,3);n++;
  XtSetArg(args[n],XmNborderWidth,1);n++;
  XtSetArg(args[n],XmNx,270);n++;
  XtSetArg(args[n],XmNy,320);n++;
  XtSetArg(args[n],XmNheight,30);n++;
  XtSetArg(args[n],XmNwidth,65);n++;
  if (language == ENGLISH)
    ClearButton = XtCreateManagedWidget("Clear",xmPushButtonWidgetClass,BBoard,args,n);
  if (language == GERMAN)
   ClearButton = XtCreateManagedWidget("Loeschen",xmPushButtonWidgetClass,BBoard,args,n); 

  XtAddCallback (ClearButton,XmNactivateCallback,(XtCallbackProc) ClearCB,NULL);
  
  XtManageChild (BBoard);
  return BBoard;
} 

/*-------------------------------------------------------------
**  main
*------------------------------------------------------------*/
int main (int argc,char **argv)
{
  Widget    main;             /*  main window         */
  Widget    BBoard;           /*  Row column          */
  Widget    menu_bar;         /*  Row column          */
  XtAppContext  app_context;  /*  Application context */
  Widget uilmain, rowcol, sep;
  char    *reqMessage;
  OID             ddd;
  char sqlstm[150];
  char *ResPath;

  Arg    args [2]; 
  register int  n;

  PutObject( NULL, NULL, 1 );
  GetObject( ddd, NULL, NULL );
  ShowGIF( NULL, 1);

  //if (argc < 3) exit(0);

  if (argc == 3) { // Db selected
    if (strcmp("german", argv[2]) == 0) 
      language = GERMAN;
    else
      language = ENGLISH;
  }
  if (argc == 2) { //No db selected
    if (strcmp("german", argv[1]) == 0)
      language = GERMAN;
    else
      language = ENGLISH;
  }

  ResPath = getenv("MOOD_BIN");
   if (!ResPath) {
      printf("\nMOOD_BIN environment variable is not set !\n");
      exit(1);
   }

  uid_filename[0] = (char *) malloc(100);
  sprintf( UidPath, "%s/gui/uil",getenv("MOOD_BIN"));
  strcpy( uid_filename[0], UidPath);
  if (language == ENGLISH)
     strcat( uid_filename[0],"/query.uid");
  if (language == GERMAN)
     strcat( uid_filename[0],"/query2.uid");

  noDBInUse = 1;

  
  userRequest = new UserRequest();
  userRequest->initESM("me");
  if (argc == 3) { /* if db selected */ 
    strcpy(sqlstm, "use db ");
    strcat(sqlstm, argv[1]);
    userRequest->executeMSQL(sqlstm);

  if ((userRequest->operationType == HANDLE_ERROR) &&
       (userRequest->userError.errorType != NO_ERROR)) {
        char *msg = userRequest->getErrorText();
        printf("Error: %s\n", msg);
        free(msg);
  }
  }
  MrmInitialize();

  /*  Initialize toolkit and open display. */
  toplevel =XtAppInitialize(&app_context,"Demos",NULL,0,&argc,argv,NULL,args,0);

  main = XtVaCreateManagedWidget("main_wid",
        xmMainWindowWidgetClass, toplevel,
        NULL);

  menu_bar = CreateMenuBar (main);

  rowcol = XtVaCreateWidget("mainrowcol", xmRowColumnWidgetClass,
        main,
        XmNorientation, XmVERTICAL,
        NULL);

  MrmOpenHierarchy( XtNumber(uid_filename), uid_filename,
                     NULL, &Mrmdata );
  MrmFetchWidget(Mrmdata,"uilmain",
       rowcol,
       &uilmain,
       (short int *) &dummy_class );
  XtManageChild(uilmain);

  sep = XtVaCreateWidget ("sep",xmSeparatorWidgetClass,
                rowcol,
                XmNorientation , XmHORIZONTAL,
                XmNleftAttachment , XmATTACH_FORM,
                XmNrightAttachment , XmATTACH_FORM,
                NULL);
  XtManageChild(sep);
 
  BBoard = CreateBBoard  (BBoard,rowcol); 

  CreateText (BBoard);

  XtManageChild (rowcol);

  /*  Realize toplevel widgets. */
  XtRealizeWidget (toplevel);

  ChangeCursor(toplevel, XC_hand2);

  /*  Process events. */
  XtAppMainLoop (app_context);

  userRequest->closeESM();

}

