#if !defined(__HASHTAB_H)
#define __HASHTAB_H

#define PRIME     97423L

#include "moods.h"

struct Bucket {

   byte   * start;
   Bucket * ovf;
   byte   * curr;
   uint   avail;     // # of available rooms for objects

};

class HashTab {

   byte   * bufferArea;    // Pointer to the primary hash area
   uint   bufferSize;      // Size of the primary hash area in bytes
   byte   * ovfArea;       // Pointer to the start of the overflow area
   uint   ovfAreaSize;     // Size of the overflow area in bytes
   uint   nBuckets;        // # of buckets
   byte   objSize;         // Size of the objects to be placed into the table
   byte   keyOffs;         // Offset of the key within each object
   byte   keySize;         // The length of the key field

   Bucket * buckets;       // Pointer to the bucket header objects
   byte   * nextOvf;       // Pointer to the next place in the overflow area
   uint   bucketSize;      // Size of a bucket in bytes
   uint   bucketCapacity;  // # of objects a bucket may contain


public :

   HashTab();
   HashTab ( byte * buf, uint bufSize,  byte * ovf, uint ovfSize,
             uint bucks, uint oSize, uint kOffs, uint kSize );
   ~HashTab();

   int Setup ( byte * buf, uint bufSize, byte * ovf, uint ovfSize,
                uint bucks, uint oSize, uint kOffs, uint kSize );
   void Destroy();

   int Put(byte *data);
   byte * Search (byte *data, Bucket **bp, int *n);
   byte * SearchNext (byte *data, Bucket **bp, int *n);

   byte   * OvfArea() { return ovfArea; }
   uint   OvfAreaSize() { return ovfAreaSize; }

};

uint Hash ( byte *key, uint keySize, uint hashMax );


#endif
