#ifndef __MOOD_PARSER
#define __MOOD_PARSER

#include "reqroot.h"
#include "linklist.h"
#include <smex.h>
#include <mood_catalog.h>
#include <keywords.h>
#include "opt_defs.h"
#include "dbi_gen.h"
#include "opt.h"
#include "dbi.h"
#include "global.h"
#include <lists.h>
#include <set.h>


	/******		DATA MANIPULATION STATEMENTS		*******/


/*************************************************************************
                CLASS FOR SELECT CLAUSE
**************************************************************************/

class Query : public RequestRoot
{
public :
        Expression      *whereP,        // where predicate
                        *startWithP,    // start with predicate
                        *havingP;       // having predicate

        ProjectionExpression
		        *projectionList,// projection list of the query
                        *groupByList,   // group by list of the query
                        *orderByList;   // order by list of the query

	BindExpression  *bindList;      // bind list of the query

	Expression      *recurseAttribute; // not null if recurse is def.

        Query           *parentQuery;   // parent query if exists
        Query           *subQueries;    // subqueries if exists
	AndTerms	*ORTerms;
	QUERY		*queryTree;
        RELATION_SCHEMA *solutionFile;
	SortedStringList *useSelfVariables; 

			Query( UserRequest *userReq );
	int		selectClause();
	int		fromClause();
	int		recurseOnClause();
	int		startWithClause();
	int		whereClause();
	int             groupByClause();
	int		havingClause();
	int		orderByClause();
	int		parse();
	void		printOn();
	void		connectUserRequest();
	void		normalizeQuery();
	int		generateTree();
	int		generateTreeForNoPredicate();
	int		generateAllTree();
	int		checkSemantic();
	int		generatePlan();
	void		executePlan();
	char		*getClassName( char *bindVar );
	char*		checkAltintas();
			~Query();
};


inline void Query::connectUserRequest()
{
	ur->operationType 	 = SELECT_STMT;
	ur->operationTypeOnError = SELECT_STMT;
	ur->operation		 = VOID this;
}


/*************************************************************************
                CLASS FOR UPDATE CLAUSE
**************************************************************************/

class UpdateClause : public RequestRoot
{
public :
        Expression      *whereP;        // where predicate
	BindExpression  *bindList;      // bind list of the update 
	LinkedList	*updateExpression; // expressions of SET statement
	LinkedList	operators;
	LinkedList	pathList;
	ProjectionExpression 
			*projectionList;
	char		*className;

			UpdateClause( UserRequest *userReq );
	int		updateClause();
	int		setClause();
	int		whereClause();
	void		connectUserRequest();
	int 		parse();
	void		printOn();
	int		parseSetClause( LinkedList *);
	int		update( OID * );
	int		update( OID *, int& );
	int		pathUpdate( OID *, int );
	int		findLinkedList( PathExpression *, LinkedList * );
	int		findLinkedLists();
	int		parseListClause( LinkedList * );
	char*		getBindVar();
	char*		getClassName();
	int		executeUpt();
	int		selectForRef( Query *, OID * );
	int		selectSet( Set *, Query * );
	int		selectList( List *, Query *);
	int		updateListItems( List *, LinkedList * );
	int		updateSetItems( Set *, LinkedList * );
	int 		updateRef( USERDESC *, int , int, char * );
	int 		updateSet( USERDESC *, int, int, char * );
	int		updateList( USERDESC *, int, int, char * );
	int		fillOpList();
	void		freeAttrList( AttrList * );
	AttrList*	findAttrList( AttrList *, char * );
			checkSemantic();
			convertToSelectClause();
			execute();
			~UpdateClause();
};


inline void UpdateClause::connectUserRequest()
{
        ur->operationType        = UPDATE_STMT;
        ur->operationTypeOnError = UPDATE_STMT;
        ur->operation            = VOID this;
}


/*************************************************************************
                CLASS FOR DELETE CLAUSE
**************************************************************************/

class DeleteClause : public RequestRoot
{
public :
        Expression      	*whereP;        // where predicate
        ProjectionExpression 	*projectionList;// projection list of the query
	BindExpression  	*bindList;      // bind list of the query

			DeleteClause( UserRequest *userReq );
	int		deleteClause();
	int		fromClause();
	int		whereClause();
	int 		parse();
	void		connectUserRequest();
	void		printOn();
	int		execute();
	int		DeleteObj(OID);
	int		DeleteRSL(char *, OID);
	void		convertToList(AttrList *, LinkedList *);
	int		DeleteObject(OID *, char *, int);
	int		DeleteSet(OID *, char *);
	int		DeleteList(OID *, char *);
			~DeleteClause();
};

inline void DeleteClause::connectUserRequest()
{
        ur->operationType        = DELETE_STMT;
        ur->operationTypeOnError = DELETE_STMT;
        ur->operation            = VOID this;
}



/*************************************************************************
                CLASS FOR NEW CLAUSE
**************************************************************************/

class NewClause : public RequestRoot
{
public :
	LinkedList	*newExpList;
	LinkedList	*attrList;
	char 		className[MAXSTRLEN];

			NewClause(UserRequest *userReq); 
	int 		newClause();
	void 		connectUserRequest();
	int 		parse();
	int 		parseSetClause(LinkedList *);
	int		parseListClause(LinkedList *);
	void 		printOn();
	int 		execute(OID *);
	int		checkConstructor(FunctionType *funcType, int *ndx, CatBOOLEAN *success);
	void		convertToList(AttrList *);
}; // NewClause

inline void NewClause::connectUserRequest()
{
	ur->operationType	 = NEW_STMT;
	ur->operationTypeOnError = NEW_STMT;
	ur->operation		 = VOID this;
} 



/*************************************************************************
                CLASS FOR DESCRIBE CLAUSE
**************************************************************************/

class DescribeClause : public RequestRoot
{
public :
	char		className[80];
	AttrList	*att;
	LinkedList	*superCls, *subCls;

	void	connectUserRequest();
	int 	parse();
	void	setClassName( char *);
	int	execute();
		DescribeClause( UserRequest *userReq ); 
		~DescribeClause();
};

inline void DescribeClause::connectUserRequest()
{
        ur->operationType        = DESCRIBE_STMT;
        ur->operationTypeOnError = DESCRIBE_STMT;
        ur->operation            = (void *) this;
}

inline void DescribeClause::setClassName( char *clsName )
{
	strcpy( className, clsName );
}


/*************************************************************************
                CLASS FOR DESCRIBE CLAUSE
**************************************************************************/

class CreateDatabase : public RequestRoot
{
private :
	char		databaseName[MAXSTRLEN];

public :
	void		setDatabaseName( char * );
	void		connectUserRequest();
			parse();
			execute();
			CreateDatabase( UserRequest *userReq );
			~CreateDatabase();
};

inline void CreateDatabase::setDatabaseName(char *cName)
{
        strcpy(databaseName, cName);
}
 
inline void CreateDatabase::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = CREATE_DATABASE;
        ur->operationTypeOnError = CREATE_DATABASE;
}


class UseDatabase : public RequestRoot
{
private :
	char		databaseName[MAXSTRLEN];

public :
	void		setDatabaseName( char * );
	void		connectUserRequest();
			parse();
			execute();
			UseDatabase( UserRequest *userReq );
			~UseDatabase();
};

inline void UseDatabase::setDatabaseName(char *cName)
{
        strcpy(databaseName, cName);
}
 
inline void UseDatabase::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = USE_DATABASE;
        ur->operationTypeOnError = USE_DATABASE;
}

/************************ USE THIS FOR ALL DROP *********************/

class GenericDrop : public RequestRoot
{
public :
	char		nameToBeDropped[MAXSTRLEN];
	
	void		setName( char * );
	void		connectUserRequest();
			GenericDrop( UserRequest * userReq) {
				RequestRoot::ur = userReq;
			}
			~GenericDrop();
};
/************************************/

class DropDatabase : public RequestRoot
{
public :
	char		databaseName[MAXSTRLEN];

	void		setDatabaseName( char * );
	void		connectUserRequest();
			parse();
			execute();
			DropDatabase( UserRequest *userReq );
			~DropDatabase();
};

class ClassDrop : public RequestRoot
{
public :
	LinkedList	*clsList;
        //char            className[MAXSTRLEN]; //name of the index to be dropped
        //int             force;
 
        void            insertClassName(char *);
        void            connectUserRequest();
                        parse();
                        execute();
                        ClassDrop( UserRequest * userReq);
                        ~ClassDrop();
};
 
inline void ClassDrop::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = DROP_CLASS;
        ur->operationTypeOnError = DROP_CLASS;
}


 
inline void DropDatabase::setDatabaseName(char *cName)
{
        strcpy(databaseName, cName);
}
 
inline void DropDatabase::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = DROP_DATABASE;
        ur->operationTypeOnError = DROP_DATABASE;
}

class ListDatabase : public RequestRoot
{
public :
	LinkedList	*dbList;

	void		connectUserRequest();
			execute();
			ListDatabase( UserRequest *userReq );
			~ListDatabase();
};

inline void ListDatabase::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = LIST_DATABASES;
        ur->operationTypeOnError = LIST_DATABASES;
}

class ListType : public RequestRoot
{
public :
	LinkedList	*typeList;

	void		connectUserRequest();
			execute();
			ListType( UserRequest *userReq );
			~ListType();
};

inline void ListType::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = LIST_TYPES;
        ur->operationTypeOnError = LIST_TYPES;
}

class CreateFunction : public RequestRoot
{
public :
	char	fileName[2*MAXSTRLEN];
	char	protoTypeStr[2*MAXSTRLEN];
	char	className[MAXSTRLEN];
	char	fSpec; // PRIVATE || PUBLIC
	
        void    connectUserRequest();
		parse();
                execute();
                CreateFunction( UserRequest *userReq, char funcSpec );
                ~CreateFunction();
};
 
inline void CreateFunction::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = CREATE_FUNCTION;
        ur->operationTypeOnError = CREATE_FUNCTION;
}

class DropFunction : public RequestRoot
{
private :
        char            delStr[2*MAXSTRLEN];
        char            className[MAXSTRLEN];
	int		deletionType;

public :
                DropFunction( UserRequest *);
        void    connectUserRequest();
        int     parse();
        int     execute();
                ~DropFunction();
};

inline void DropFunction::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = DROP_FUNCTION;
        ur->operationTypeOnError = DROP_FUNCTION;
}

typedef enum
{
	OWN_FUNCTION,
	INHERITED_FUNCTION,
	ALL_FUNCTION
} ListingType;

class ListFunction : public RequestRoot
{
private :
        char            className[MAXSTRLEN];
	ListingType	listingType;

public :
                ListFunction( UserRequest *userReq );
        void    connectUserRequest();
        int     parse();
	LinkedList	*functionList;
        int     execute();
                ~ListFunction();
};
 
inline void ListFunction::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = LIST_FUNCTION;
        ur->operationTypeOnError = LIST_FUNCTION;
}

class DisplayHierarchy : public RequestRoot
{
public :
	Hierarchy	*hier;
 
                DisplayHierarchy( UserRequest *userReq );
        void    connectUserRequest();
        int     parse();
        int     execute();
                ~DisplayHierarchy();
};
 
inline void DisplayHierarchy::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = DISP_HIER;
        ur->operationTypeOnError = DISP_HIER;
}

class StartFile : public RequestRoot
{
public :
	char	fileName[ MAXSTRLEN ];

                StartFile( UserRequest *userReq );
        void    connectUserRequest();
        int     parse();
        int     execute();
                ~StartFile();
};

inline void StartFile::connectUserRequest()
{
        ur->operation     = VOID this;
        ur->operationType = START_FILE;
        ur->operationTypeOnError = START_FILE;
}


#endif // __MOOD_PARSER
