#ifndef __USERREQ_H
#define __USERREQ_H

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "mood_error.h"
#include "keywords.h"
#include "timer.h"

extern Timer	moodTimer;

class RELATION_SCHEMA;


/*
		USER REQUEST CLASS DEFINITION
*/



// Token structure

typedef struct
{
        int type;
        char value[TOKENSIZE];
} Token;


typedef enum{
		SELECT_STMT,
		DELETE_STMT,
		UPDATE_STMT,
		INSERT_STMT,
		NEW_STMT,
		CREATE_INDEX,
		DROP_INDEX,
		CREATE_BJINDEX,
		DROP_BJINDEX,
		CREATE_TYPE,
		DROP_TYPE,
		CREATE_CLASS,
		DROP_CLASS,
		HANDLE_ERROR,
		RUN_PREVIOUS,
		EXIT_FROM_MOODS,
		DESCRIBE_STMT,
		USE_DATABASE,
		DROP_DATABASE,
		CREATE_DATABASE	,
		LIST_DATABASES,
		LIST_TYPES,
		COMMIT_XACT,
		ABORT_XACT,
		TAKE_SAVEPOINT,
		ROLLBACK_TO_SAVEPOINT,
		OS_COMMAND,
		DECLARE_STMNT,
		CREATE_FUNCTION,
		DROP_FUNCTION,
		LIST_FUNCTION,
		DISP_HIER,
		START_FILE,
		NO_OPERATION
} OperationType;

	
typedef int USERID;

typedef void * Operation;

			

class UserRequest
{

public  :

//	static unsigned MAXTEXTLENGTH;

	char 		*text, *prevText;
	unsigned	curpos,
			prevpos; // current position and previous position
	Token		token;
	OperationType   operationType,
                        operationTypeOnError;
        Operation       operation;
	MoodsError	userError; // keeps Error if exists


public :
	USERID 		userid;
	unsigned int 	textLength, // length of the text of the user request
			maxTextLength; // maximum length of the text
	float		expectedTime, // expected time to execute user request
			actualTime,   // actual time to execute user request
			parseTime, // parsing time of user request
			optimizeTime; // optimization time for user request


public :
			UserRequest();
			UserRequest(char *inText);
	int		executeMSQL( char *inText );
	int		execute_MSQL( char *inText );
	void 		readText();
	int 		handleOperation();
	int		reset();
	void 		printText();
	int 		parseText();
	int 		getToken();
	void 		backtrack();
	void		handleError();
	void		printTime();
	int 		tokenType();
	void 		setError(ErrorType type, int no, char *msg);
	char 		*tokenValue();
	char		*getErrorText();
	char		*getMessageText();
	char		*getClassName( char *bindVar );
	int		execute();
	OperationType	getOperationType();
	RELATION_SCHEMA *getBulkSchema();
			closeESM();
			initESM( char * );
			startXAct();
			commitXAct();
			abortXAct();
			~UserRequest();

	
private :
	void 		skipTabs();
	char 		*getValue();
};

inline OperationType UserRequest::getOperationType()
{
	return operationType;
}

// backtracks during the parsing
inline void UserRequest::backtrack()
{
	curpos = prevpos ;
}
 
// handle error of user request
inline void UserRequest::handleError()
{
	userError.handleError();
}


// retrieve token type
inline int UserRequest::tokenType()
{
	return token.type;
}

 
// set error if exists
inline void UserRequest::setError(ErrorType type, int no, char *msg)
{
	operationType = HANDLE_ERROR;
	userError.setError(type, no, msg);
}


inline char *UserRequest::tokenValue()
{
	return token.value;
}

#endif // __USERREQ_H
