/* 
* Volcano Optimizer Generator 
* =========================== 
*   
* $Header: /tmp_mnt/home/chainsaw/bill/optgen/optgen/RCS/grammar.y,v 1.16 92/11/18 01:32:59 bill Exp Locker: bill $ 
*
*/ 
 
%{
#include "opt_defs.h" 
#include "parse.h" 
#include "grammar.h"

#define YYMAXDEPTH 500

extern SYM_TAB 			symbol_table[] ;
extern char			yytext [] ;
extern int			max_algorithm_arity, max_operator_arity,
				oper_count, algo_count ;
CURRENT_RULE_SIDE 		current_side ;
int				input_var_id = -1, op_arg_var_id = -1,
				al_arg_var_id = -1,
				oper_var_id = -1,
				enfo_var_id = -1, algo_var_id = -1 ; 
extern int			sym_tab_index, var_tab_index, trans_rule_num,
				impl_rule_num, max_level, node_id, end_parens ;
extern RULE_GROUP_TAB		rule_group_table[];
RULE_EXPR			*lhs_root, *rhs_root ;
extern RULE_PRIORITY_INFO      	trans_rule_priority[],
                        	impl_rule_priority[] ;
int				expr_stack[MAX_EXPR_PER_RULE], top, expr_index,
				level = 0 ; 
char				temp_str[MAX_NAME] ;
RULE_EXPR			*expr_table[MAX_EXPR_PER_RULE], *make_expr () ;
RULE_DIRECTION			rule_direction ;
FUNC				func_decl_type = FUNC_NOT_SET ;
extern CURRENT_RULE_TYPE	current_rule_type ;
CURRENT_PROCESS			current_process ;
FUNC				current_var_type ;
USER_CODE			current_user_code = CODE_TYPE_NOT_SET ;
int 				num_nodes_lhs = 0 ;
int				include_it = 0, current_rule_group = -1 ;

#if defined(__cplusplus)
extern "C" void yyerror (char *) ;
#endif

#if defined (ANSI)
extern void yyerror (char *) ;
#endif

#if defined(__cplusplus) || defined (ANSI)
extern void write_trans_rule_code (RULE_EXPR *, 
                                RULE_DIRECTION) ;
extern void write_trans_rule_lhs_code (FILE *, RULE_EXPR *, int,
                                RULE_DIRECTION) ;
extern void write_trans_rule_rhs_code (FILE *) ;
extern void write_impl_rule_code (RULE_EXPR *,
                                RULE_DIRECTION) ;
extern void write_impl_rule_rhs_code (FILE *, RULE_EXPR *) ;
extern char *my_strstr( char*, char*) ;
extern void process_user_code (char*) ;
extern void free_tree (RULE_EXPR *) ;
extern RULE_EXPR *make_expr(int,int,char *,char *,FUNC,
                        int level) ;
extern void number_nodes (RULE_EXPR *, int, CURRENT_RULE_SIDE) ;
extern void clear_rule_group_table (RULE_GROUP_TAB *) ;
void print_rule_group_table (RULE_GROUP_TAB *) ;
void add_rule_group_table (RULE_GROUP_TAB *, char *, int *) ;
void find_rule_group_table (RULE_GROUP_TAB *, char *, int *,
				 BOOLEAN *) ;
#else
extern void                    	write_trans_rule_code () ;
extern void                    	write_trans_rule_lhs_code () ;
extern void                    	write_trans_rule_rhs_code () ;
extern void                    	write_impl_rule_code () ;
extern void                    	write_impl_rule_lhs_code () ;
extern void                    	write_impl_rule_rhs_code () ;
extern char                    	*my_strstr () ;
extern void                    	process_user_code () ;
extern void                    	free_tree () ;
extern RULE_EXPR               	*make_expr () ;
extern void                    	number_nodes () ;
extern void			clear_rule_group_table (), 
				print_rule_group_table(),
				add_rule_group_table(), 
				find_rule_group_table () ;
extern void yyerror () ;
#endif

%}

%token OPER_DEF ALGO_DEF ENFO_DEF IMPL_RULE_DEF TRANS_RULE_DEF OPEN_PAREN CLOSE_PAREN
%token ONE_WAY TWO_WAY INPUT_VAR IDENT OP_ARG_VAR NUMBER NIL OP_VAR AL_ARG_VAR ONE_WAY_ONCE
%token BEGIN_CC END_CC C_CODE COND_CODE APPL_CODE O_PAREN C_PAREN PRIO OG

%%

optimizer :
      user_input
    ;

user_input : 
    | user_input decl
    | user_input rule 
    ;

decl:
	decl_type {current_process = DECL_FUNC;} ident arity
		     {   
			if ($4 < 0)
			{
			    fprintf (stderr, 
			      "Error: identifier %s has negative arity(?????).\n",
				symbol_table[$3].name);
				exit (-1) ;
			}

			if (func_decl_type == IS_ENFORCER && $4 > 1)
			{
			    fprintf (stderr, 
			      "Error: Arity of enforcer (%s) must be 1.\n",
				symbol_table[$3].name);
				exit (-1) ;
			} 

			symbol_table[$3].arity = $4 ;
			if (func_decl_type == IS_OPERATOR && symbol_table[$3].arity >
			    max_operator_arity)  
			{
			    max_operator_arity = symbol_table[$3].arity ;
			    if (max_operator_arity > MAX_INPUTS)
			    {
				fprintf (stderr, 
					  "Increase MAX_INPUTS in parse.h and re-make.\n");
				exit (-1) ;
			    }
			}
			if ((func_decl_type == IS_ALGORITHM || func_decl_type == IS_ENFORCER) 
				&& symbol_table[$3].arity >
			    		max_algorithm_arity  )
			{
			    max_algorithm_arity = symbol_table[$3].arity ;
			    if (max_algorithm_arity > MAX_INPUTS)
			    {
				fprintf (stderr, 
					  "Increase MAX_INPUTS in parse.h and re-make.\n");
				exit (-1) ;
			    }
			}
			symbol_table[$3].func = func_decl_type ;
			symbol_table[$3].id = $1 ;
		    }
    ; 

decl_type:
	OPER_DEF {func_decl_type = IS_OPERATOR; $$ = oper_count; ++oper_count;}
    |   ALGO_DEF {func_decl_type = IS_ALGORITHM; $$ = algo_count; ++algo_count;}
    |   ENFO_DEF {func_decl_type = IS_ENFORCER; $$ = algo_count; ++algo_count;} 
    ;

rule_def:               { 
			  int i ;
			  for (i = sym_tab_index; ++i<=MAX_SYMBOLS;)
			  {
			      strcpy (symbol_table[i].name, "") ;
			      symbol_table[i].arity = -99 ;
			      symbol_table[i].defined_in = DEFINED_NOT_SET ;
			      symbol_table[i].var_set = VAR_NOT_SET ;
			      symbol_table[i].func = FUNC_NOT_SET ;
			      symbol_table[i].id = -99 ;
			      symbol_table[i].bind_id = -99 ;
			      symbol_table[i].rhs_count = 0 ;
                          }
                        }
	  IMPL_RULE_DEF {
			  ++impl_rule_num; current_rule_type = IMPL_RULE; 
      			  var_tab_index = sym_tab_index ; oper_var_id = -1;
		  	  op_arg_var_id = -1; input_var_id = -1; max_level = -1; 
			  al_arg_var_id = -1;
			  enfo_var_id = -1; algo_var_id = -1;}
	| TRANS_RULE_DEF {
			  ++trans_rule_num; current_rule_type = TRANS_RULE; 
      			  var_tab_index = sym_tab_index ; oper_var_id = -1;
			  al_arg_var_id = -1;
		  	  op_arg_var_id = -1; input_var_id = -1; max_level = -1; }
    ;

arity:
	NUMBER	   {$$ = atoi (yytext) ;}
    ;

rule :
      	rule_def {current_process = PROCESS_RULE; current_side = LHS_CURRENT;
			current_rule_group = -1 ;
			current_user_code = CODE_TYPE_NOT_SET ;}
	opt_group {
			BOOLEAN	found ;
			int temp_int ;
			char temp_name[MAX_NAME] ;
		/*	printf ("opt_group = %s\n", (char *) $3) ; */
			strcpy (temp_name, (char *) $3) ;
			temp_int = -1 ;
			find_rule_group_table (&(rule_group_table[0]),
					       temp_name, &(temp_int), &found) ;
			if (found == FALSE)
			{
			   /* printf ("did not find group in table\n") ; */
			    add_rule_group_table (&(rule_group_table[0]),
						  temp_name, &temp_int) ;
			}
			current_rule_group = temp_int ;
		   }
	priority {if (current_rule_type == TRANS_RULE)
		      trans_rule_priority[trans_rule_num].priority = $5 ;
		  else 
		      impl_rule_priority[impl_rule_num].priority = $5 ;}
	lhs 
		 {
		  if (lhs_root -> func != IS_OPERATOR && lhs_root -> func != IS_OPER_VARIABLE
			&& 
		      current_rule_type == IMPL_RULE)
		  {
		      fprintf (stderr, 
		        "Error: %s used instead of operator or operator variable.\n", 
				lhs_root -> name) ;
		      fprintf (stderr, "       impl. rule number = %d\n", impl_rule_num) ;
		      exit (-1) ;
		  }
		 }
	direction 
		 {current_side = RHS_CURRENT;} 
	rhs
		 {
#if DEBUG>=1
		  FILE *fp ; 
		  ((trans_rule_num == 0 && impl_rule_num == -1)
			||
		   (trans_rule_num == -1 && impl_rule_num == 0)) 
		       ?
		      (fp = fopen ("debug.out", "w")) :
		      (fp = fopen ("debug.out", "a+")) ;
		  fprintf (fp, "Rule type = ") ;  
		  if (current_rule_type == TRANS_RULE)
		  {
			fprintf (fp, "TRANS_RULE\n") ;
			fprintf (fp, "Rule number = %d\n", trans_rule_num) ;
		  }
		  else
		  {
			fprintf (fp, "IMPL_RULE\n") ;
			fprintf (fp, "Rule number = %d\n", impl_rule_num) ;
		  }
		  fprintf (fp, "Left side of rule:\n") ;
		  print_expr_tree (fp, lhs_root, 2) ;
		  fprintf (fp, "Right side of rule:\n") ;
		  print_expr_tree (fp, rhs_root, 2) ;
		  fclose (fp) ;
#endif
		  if (rhs_root -> func != IS_ALGORITHM && rhs_root -> func != IS_ENFORCER
			&& 
		      current_rule_type == IMPL_RULE)
		  {
		      fprintf (stderr, 
		        "Error: %s used instead of algorithm or enforcer.\n", 
				rhs_root -> name) ;
		      fprintf (stderr, "       impl. rule number = %d\n", impl_rule_num) ;
		      exit (-1) ;
		  }
		  if (current_rule_type == TRANS_RULE)
		      write_trans_rule_code (lhs_root, rule_direction) ;
		  else
		      write_impl_rule_code (lhs_root, rule_direction) ;
		 } 
		 user_code 
		 {
			{FILE *fp; 
			 int i; 
			 for (i = sym_tab_index; ++i<=var_tab_index;)
			 {
			     if (symbol_table[i].func == IS_OP_ARG_VARIABLE
				 && symbol_table[i].defined_in == DEFINED_IN_RHS
				 && symbol_table[i].var_set == VAR_NOT_SET)
			     { 
				  fprintf (stderr, 
				  "Error: argument var %s used in RHS of rule \n",
					       symbol_table[i].name);
				  fprintf (stderr, 
				  "       but not referenced in user application code.\n") ;

				  if (current_rule_type == TRANS_RULE)
				  {
				    fprintf (stderr, 
					"       trans. rule number = %d\n", trans_rule_num) ;
				  }
				  else
				  {
				    fprintf (stderr,
					"       impl. rule number = %d\n", impl_rule_num) ;
				  }
			    	  exit (-1) ;
			     } 
			  }
		  	 if (current_rule_type == TRANS_RULE)
			 {
				fp = fopen ("temp_match_trans_rule.c", "a+") ;
				write_trans_rule_rhs_code (fp) ;
				for (i = -1; ++i < end_parens;) 
				    fprintf (fp, "}\n") ;
				fprintf (fp, "incr_rule_iteration_%d: \n",
					      trans_rule_num) ;
				fprintf (fp, "++ RULE_ITERATION ;\n") ;
				fprintf (fp, 
				  "if (RULE_ITERATION <= FIRE_RULE_COUNT)\n") ;
                		fprintf (fp, "goto fire_rule_%d ;\n", 
					trans_rule_num) ;
				fprintf (fp, "if (found_rule == TRUE)\n") ;
				fprintf (fp, "return (OKAY) ;\n") ;
				fprintf (fp, "else goto try_rule ;\n") ;
				fclose (fp) ;
			 }
		  	 if (current_rule_type == IMPL_RULE)
			 {
				fp = fopen ("temp_match_impl_rule.c", "a+") ;
				write_impl_rule_rhs_code (fp, rhs_root) ;
				for (i = -1; ++i < end_parens;) 
				    fprintf (fp, "}\n") ;
				fprintf (fp, "if (found_rule == TRUE)\n") ;
				fprintf (fp, "return (OKAY) ;\n") ;
				fprintf (fp, "else goto try_rule ;\n") ;
				fclose (fp) ;
			 }
			}
#if DEBUG>=1
	  		dump_symbol_table () ; 
#endif
			free_tree (lhs_root) ;
			free_tree (rhs_root) ;
		 }
    ;

opt_group:  {{	
		/* printf ("og ident. empty. current table::\n") ; */
		strcpy (temp_str, "DEFAULT") ;

		$$ = (int) temp_str ; }}

    | OG 
      IDENT { 
		if (strcmp (yytext, "DEFAULT") == 0)
		{
		    fprintf (stderr, "Reserved name DEFAULT used for rule group.\n") ;
		    if (current_rule_type == TRANS_RULE)
		    {
			fprintf (stderr,
			 "       trans. rule number = %d\n", trans_rule_num);
		    }
		    else
		    {
			fprintf (stderr,
			    "       impl. rule number = %d\n", impl_rule_num) ;
		    }
		    exit (-1) ;
		}
		$$ = (int) yytext; 
	       /*
		printf ("matched on og ident. current table:\n") ;
	       */

	    } ;

priority:    {{$$ = 0;}}
    | PRIO NUMBER {{/*printf ("for prio, yytext = %s\n", yytext) ;*/ $$ = atoi (yytext); }} ; 

user_code : 
    cond_code  
    appl_code 
    ;

cond_code :
    COND_CODE 
		 { 
                   current_user_code = USER_COND_CODE ;} c_code_block 
    |            
    ;


appl_code :
     APPL_CODE 
		{ 
                  current_user_code = USER_APPL_CODE ;} c_code_block
    |           
    ;

c_code_block : 
    BEGIN_CC 
	{
	    include_it=1;
	    {
		FILE *fp ;
		if (current_rule_type == TRANS_RULE)
		{
		    fp = fopen ("temp_match_trans_rule.c", "a+") ;
		    fprintf (fp, "{\n") ;
		    fclose (fp) ;
		}
		if (current_rule_type == IMPL_RULE)
		{
		    fp = fopen ("temp_match_impl_rule.c", "a+") ;
		    fprintf (fp, "{\n") ;
		    fclose (fp) ;
		}
	    }
	} 
	c_code_line 
	{
	    include_it=0;
	    {
		FILE *fp ;
		if (current_rule_type == TRANS_RULE)
		{
		    fp = fopen ("temp_match_trans_rule.c", "a+") ;
		    fprintf (fp, "}\n") ;
		    fclose (fp) ;
		}
		if (current_rule_type == IMPL_RULE)
		{
		    fp = fopen ("temp_match_impl_rule.c", "a+") ;
		    fprintf (fp, "}\n") ;
		    fclose (fp) ;
		}
	    }
	}
	END_CC ;

c_code_line:
    | C_CODE {process_user_code (yytext);} c_code_line ;

lhs:
	{lhs_root = NULL; expr_index = -1 ; top = -1; 
	 node_id = -1; level = 0; num_nodes_lhs = 0;}
	 expr {number_nodes (lhs_root, 1, LHS_CURRENT); } ;


rhs:
      {rhs_root = NULL; expr_index = -1 ; top = -1;
	 node_id = -1; level = 0; }
	 expr {number_nodes (rhs_root, 1, RHS_CURRENT); } ;

expr :   begin_expr expr_form end_expr {$$ = $2;} ;

expr_form : expr_form1 | expr_form2 ;

expr_form1 :    
	 func 
	 arg 		     {
			      if (((symbol_table[$1].func == IS_OPERATOR ||
			          symbol_table[$1].func == IS_OPER_VARIABLE) 
				  && symbol_table[$2].func != IS_OP_ARG_VARIABLE)
					||
			          ((symbol_table[$1].func == IS_ALGORITHM ||
			            symbol_table[$1].func == IS_ENFORCER) 
				  && symbol_table[$2].func != IS_AL_ARG_VARIABLE))
			      {
				  fprintf (stderr, "Error: incorrect argument variable %s.\n",
					symbol_table[$2].name) ;
				  if (current_rule_type == TRANS_RULE)
				  {
				    fprintf (stderr, 
				   "       trans. rule number = %d\n", trans_rule_num) ;
				  }
				  else
				  {
				    fprintf (stderr, 
				   "       impl. rule number = %d\n", impl_rule_num) ;
				  }
				  exit (-1) ;
			      }}
	 OPEN_PAREN 
			     {{
			      RULE_EXPR *ex ;
			      ex = make_expr (symbol_table[$1].id,symbol_table[$2].id,
			                          symbol_table[$1].name,symbol_table[$2].name,
					     (FUNC) symbol_table[$1].func,
					     level); 
			      ++expr_index ;
			      if (expr_index >= MAX_EXPR_PER_RULE)
			      {
				  fprintf (stderr, 
				   "Increase MAX_EXPR_PER_RULE in parse.h and re-make.\n") ;
				  exit (-1) ;
			      }
			      expr_table[expr_index] = ex ;
			      ++top ;
			      expr_stack[top] = expr_index ;
			      if (current_side == LHS_CURRENT)
				  ++num_nodes_lhs ;
			      if (current_side == LHS_CURRENT && lhs_root == NULL)
			      {
				  lhs_root = ex ;
				  lhs_root -> group_id = current_rule_group ;	
			      }
			      if (current_side == RHS_CURRENT && rhs_root == NULL)
				  rhs_root = ex ;
			      ++level ;
			      if (current_side == RHS_CURRENT && level > 1
					&& current_rule_type == IMPL_RULE)
			      {
				  fprintf (stderr, 
				   "Error: nested expressions not allowed on RHS of impl. rule.\n") ;
				  fprintf (stderr, 
				   "       impl. rule number = %d\n", impl_rule_num) ;
				  exit (-1) ;
			      }
			      
			      if (level > max_level) max_level = level ;
			     }}
		             input_list CLOSE_PAREN 
					 
			     {{
			      if (expr_table[expr_stack[top]] -> arity != 
				   symbol_table[$1].arity &&
				   (symbol_table[$1].func == IS_OPERATOR || 
				    symbol_table[$1].func == IS_ALGORITHM ||
				    symbol_table[$1].func == IS_ENFORCER))
			      {
				  fprintf (stderr, 
				   "Error: inputs in rule (%d) != defined arity (%d).\n",
					       expr_table[expr_stack[top]] -> arity,
					       symbol_table[$1].arity);

				  if (current_rule_type == TRANS_RULE)
				  {
				    fprintf (stderr, 
				   "       trans. rule number = %d\n", trans_rule_num) ;
				  }
				  else
				  {
				    fprintf (stderr, 
				   "       impl. rule number = %d\n", impl_rule_num) ;
				  }

			          fprintf (stderr, 
				   "       Oper., algo. or enforcer: %s.\n",
					       symbol_table[$1].name);
			    	  exit (-1) ;
			      }
			      if (current_side == LHS_CURRENT && 
  			          symbol_table[$1].func == IS_OPER_VARIABLE &&
				  symbol_table[$1].arity == -1)
			      {
			          symbol_table[$1].arity = 
                                             expr_table[expr_stack[top]] -> arity ;
			      }
			      if (symbol_table[$1].func == IS_OPER_VARIABLE &&
			           expr_table[expr_stack[top]] -> arity != 
				   symbol_table[$1].arity &&
                                   symbol_table[$1].arity != -1)
			      {
				  fprintf (stderr, 
				   "Error: operator variable must always have same arity\n");
				  fprintf (stderr, 
				   "       in the same rule.\n");

				  if (current_rule_type == TRANS_RULE)
				  {
				    fprintf (stderr, 
				   "       trans. rule number = %d\n", trans_rule_num) ;
				  }
				  else
				  {
				    fprintf (stderr, 
				   "       impl. rule number = %d\n", impl_rule_num) ;
				  }

			          fprintf (stderr, 
				   "       Operator variable: %s.\n",
					       symbol_table[$1].name);
			    	  exit (-1) ;
			      }
			      -- level ;
			      $$ = expr_stack[top] ;
			      -- top ;
			     }}
    ;

expr_form2:
        {rhs_root = NULL; expr_index = -1 ; top = -1;
         node_id = -1; level = 0; } 
        input_var 
	      {{RULE_EXPR *input_ex ;
	       /*
		fprintf (stderr,
		 "Error: This type of rule not implemented yet.\n") ;
		if (current_rule_type == TRANS_RULE)
			fprintf (stderr, 
			     "       trans. rule %d.\n", trans_rule_num) ;
		else
			fprintf (stderr, 
			     "       impl. rule %d.\n", impl_rule_num) ;
		exit (-1) ;
	     */
              if (current_side != RHS_CURRENT || current_rule_type != TRANS_RULE) 
              {
		    fprintf (stderr,
		     "Error: Expressions with only an input variable\n") ;
		    fprintf (stderr,
		     "       are only allowed only in RHS of trans. rules.\n") ;
		    if (current_rule_type == TRANS_RULE)
			    fprintf (stderr, 
				 "       trans. rule %d.\n", trans_rule_num) ;
		    else
			    fprintf (stderr, 
				 "       impl. rule %d.\n", impl_rule_num) ;
		    exit (-1) ;
              }
	      input_ex = make_expr (symbol_table[$2].id,
	                            -1,
		                    symbol_table[$2].name, (char *) NULL,
	                            (FUNC) symbol_table[$2].func,
			            level) ;
	 rhs_root = input_ex ;}}

begin_expr:
	OPEN_PAREN ;

end_expr:
	CLOSE_PAREN ; 

ident :
	IDENT 	    {{
			int i, j = 0 ;
			for (i = -1; ++i<=sym_tab_index;)
			    if (strcmp (symbol_table[i].name, yytext) == 0)
			    { 
				j = 1 ;
			        $$ = i ;
			        break ;
			    }

			if (current_process == DECL_FUNC)
			{
			    if (j == 1)
			    {
				fprintf (stderr,
				 "Error: duplicate name decl. for oper., algo. or enforcer: %s.\n",
				 yytext) ;
				exit (-1) ;
			    }
			    else
			    {
				++sym_tab_index ;
				strcpy (symbol_table[sym_tab_index].name, yytext) ;
				$$ = sym_tab_index ;
			    }
			}

			if (current_rule_type == IMPL_RULE && j == 1 
			    && current_side == RHS_CURRENT 
			    && current_process == PROCESS_RULE
			    && (! (symbol_table[i].func == IS_ALGORITHM ||
				 symbol_table[i].func == IS_ENFORCER)))
			{
			    fprintf (stderr,
			     "Error: Identifier %s not allowed in RHS of impl. rule.\n",
			      yytext) ; 
			    fprintf (stderr, "       Must use an algorithm.\n") ;
			    fprintf (stderr, "       impl. rule %d.\n", impl_rule_num) ;
			    exit (-1) ;
			}

			if (current_rule_type == IMPL_RULE && j == 1 
			    && current_side == LHS_CURRENT 
			    && current_process == PROCESS_RULE
			    && (! (symbol_table[i].func == IS_OPERATOR ||
				 symbol_table[i].func == IS_OPER_VARIABLE)))
			{
			    fprintf (stderr,
			     "Error: Identifier %s not allowed in LHS of impl. rule.\n",
			      yytext) ; 
			    fprintf (stderr, 
				"       Must use an operator or operator variable.\n") ;
			    fprintf (stderr, "       impl. rule %d.\n", impl_rule_num) ;
			    exit (-1) ;
			}

			if (current_process == PROCESS_RULE && j == 0) 
			{
			    fprintf (stderr,
				 "Error: undefined oper., algo. or enforcer %s.\n", yytext) ;
			      if (current_rule_type == TRANS_RULE)
			      {
				    fprintf (stderr, 
					 "       trans. rule number = %d\n", trans_rule_num) ;
			      }
			      else
			      {
				    fprintf (stderr, 
					 "       impl. rule number = %d\n", impl_rule_num) ;
			      }
			     exit (-1);
			}

			if (j == 1 && current_rule_type == TRANS_RULE
			    && (symbol_table[i].func == IS_ALGORITHM ||
				 symbol_table[i].func == IS_ENFORCER))
			{
			    fprintf (stderr, 
			      "Error: algorithm or enforcer %s not allowed in trans. rule.\n", yytext) ;
			    fprintf (stderr, 
			      "       trans. rule number = %d\n", trans_rule_num) ;
			    exit (-1) ;
			}

		    }} 
	;

func :
	ident
    |   oper_var
    ;
        
oper_var :
	OP_VAR 	    {{
			int i, j = 0 ;
			for (i = sym_tab_index; ++i<=var_tab_index;)
			{
			    if (strcmp (symbol_table[i].name, yytext) == 0)
			    { 
				j = 1 ; $$ = i ; break ;
			    }
			}
			if (j == 0)
			{
			    if (current_rule_type == IMPL_RULE 
				&& current_side == RHS_CURRENT)
			    {
				fprintf (stderr,
				 "Error: Operator var. (%s) not allowed in RHS of impl. rule.\n",
				  yytext) ; 
				fprintf (stderr, "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    if (current_side == RHS_CURRENT)
			    {
				fprintf (stderr,
				 "Error: Operator var. %s used in RHS of rule not in LHS.\n",
				  yytext) ; 
				if (current_rule_type == TRANS_RULE)
				    fprintf (stderr, 
					    "       trans. rule %d.\n", trans_rule_num) ;
				else
				    fprintf (stderr, 
					"       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    ++var_tab_index ;
			    check_limit (var_tab_index, MAX_SYMBOLS) ;
			    strcpy (symbol_table[var_tab_index].name, yytext) ;
			    symbol_table[var_tab_index].id = ++oper_var_id ;
			    symbol_table[var_tab_index].bind_id = -1 ;
			    symbol_table[var_tab_index].func = IS_OPER_VARIABLE ;
			    symbol_table[var_tab_index].defined_in = DEFINED_IN_LHS ;
			    symbol_table[var_tab_index].arity = -1 ;
			    $$ = var_tab_index ;
			}
		    }} 
    ;

arg:	op_arg
    |   al_arg ;

op_arg :
	OP_ARG_VAR     {{
			int i = 0, j = 0, k = 0 ;

			for (i = sym_tab_index; ++i<=var_tab_index;)
			{
			    if (strcmp (symbol_table[i].name, yytext) == 0)
			    { 
				k = i; j = 1 ; $$ = i ; break ;
			    }
			}

			if (j == 0)
			{
			   
			    ++var_tab_index ;
			    check_limit (var_tab_index, MAX_SYMBOLS) ;
			    strcpy (symbol_table[var_tab_index].name, yytext) ;
			    symbol_table[var_tab_index].id = ++op_arg_var_id ;
			    symbol_table[var_tab_index].bind_id = -1 ;
			    symbol_table[var_tab_index].defined_in = DEFINED_IN_LHS ;
			    symbol_table[var_tab_index].rhs_count = 0 ;
                            symbol_table[var_tab_index].var_set = VAR_NOT_SET ;
			    if (current_side == RHS_CURRENT) 
			    {
				symbol_table[var_tab_index].defined_in = DEFINED_IN_RHS ;
				++(num_nodes_lhs) ;
			        if (num_nodes_lhs >= MAX_EXPR_PER_RULE)
			        {
			  	    fprintf (stderr, 
				     "Increase MAX_EXPR_PER_RULE in parse.h and re-make.\n") ;
				    exit (-1) ;
			        }
				symbol_table[var_tab_index].bind_id = num_nodes_lhs  ;
				++ (symbol_table[var_tab_index].rhs_count) ;
			    }
			    symbol_table[var_tab_index].func = IS_OP_ARG_VARIABLE ;
			    symbol_table[var_tab_index].arity = -1 ;
			    $$ = var_tab_index ;
			}
			else
			{
			    if (current_side == LHS_CURRENT)
			    {
				fprintf (stderr,
				 "Error: argument var. %s used twice in LHS.\n",
				  yytext) ; 
				if (current_rule_type == TRANS_RULE)
				    fprintf (stderr, 
				 "       trans. rule %d.\n", trans_rule_num) ;
				else
				    fprintf (stderr, 
				 "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    else
			    {
				++(symbol_table[k].rhs_count) ;
				if (symbol_table[k].rhs_count > 1)
				{
				    fprintf (stderr,
				     "Error: argument var. %s used twice in RHS.\n",
					  yytext) ; 
				    if (current_rule_type == TRANS_RULE)
					fprintf (stderr, 
					     "       trans. rule %d.\n", trans_rule_num) ;
				    else
					fprintf (stderr, 
					     "       impl. rule %d.\n", impl_rule_num) ;
				    exit (-1) ;
				}
		
				fprintf (stderr,
				 "Error: argument var. %s used in LHS and RHS of rule.\n",
				      yytext) ; 
				fprintf (stderr,
				 "       Define a new argument var. for the RHS.\n") ;
				if (current_rule_type == TRANS_RULE)
					fprintf (stderr, 
					     "       trans. rule %d.\n", trans_rule_num) ;
				else
					fprintf (stderr, 
					     "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			}
		    }} 
    ;

al_arg :
	AL_ARG_VAR     {{
			int i, j = 0 ;
			if (current_rule_type == TRANS_RULE)
			{
			    fprintf (stderr, "Error: argument type %s not allowed in trans. rule.\n",
					yytext) ;
			    fprintf (stderr, 
				 "       trans. rule %d.\n", trans_rule_num) ;
			}
			for (i = sym_tab_index; ++i<=var_tab_index;)
			{
			    if (strcmp (symbol_table[i].name, yytext) == 0)
			    { 
				j = 1 ; $$ = i ; break ;
			    }
			}
			if (j == 0)
			{
			    ++var_tab_index ;
			    check_limit (var_tab_index, MAX_SYMBOLS) ;
			    strcpy (symbol_table[var_tab_index].name, yytext) ;
			    symbol_table[var_tab_index].id = ++al_arg_var_id ;
			    symbol_table[var_tab_index].bind_id = -1 ;
			    symbol_table[var_tab_index].defined_in = DEFINED_IN_LHS ;
			    symbol_table[var_tab_index].rhs_count = 0 ;
			    if (current_side == RHS_CURRENT)
			    {
				symbol_table[var_tab_index].defined_in = DEFINED_IN_RHS ;
				++(num_nodes_lhs) ;
			        if (num_nodes_lhs >= MAX_EXPR_PER_RULE)
			        {
			  	    fprintf (stderr, 
				     "Increase MAX_EXPR_PER_RULE in parse.h and re-make.\n") ;
				    exit (-1) ;
			        }
				symbol_table[var_tab_index].bind_id = num_nodes_lhs  ;
				++ (symbol_table[var_tab_index].rhs_count) ;
			    }
			    symbol_table[var_tab_index].func = IS_AL_ARG_VARIABLE ;
			    symbol_table[var_tab_index].arity = -1 ;
			    $$ = var_tab_index ;
			}
			else
			{
			    if (current_side == LHS_CURRENT)
			    {
				fprintf (stderr,
				 "Error: argument var. %s used twice in LHS.\n",
				  yytext) ; 
				fprintf (stderr, 
				     "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    else
			    {
				++(symbol_table[i].rhs_count) ;
				if (symbol_table[i].rhs_count > 1)
				{
				    fprintf (stderr,
				     "Error: argument var. %s used twice in RHS.\n",
				      yytext) ; 
				    fprintf (stderr, 
					 "       impl. rule %d.\n", impl_rule_num) ;
				    exit (-1) ;
			        }
				fprintf (stderr,
				 "Error: argument var. %s used in LHS and RHS of rule.\n",
				      yytext) ; 
				fprintf (stderr,
				 "       Define a new argument var. for the RHS.\n") ;
				fprintf (stderr, 
				     "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			}
		    }} 
    ;

direction :
	ONE_WAY		{rule_direction = L_TO_R ;} 
    |   ONE_WAY_ONCE    {
			    if (current_rule_type == IMPL_RULE)
			    {
				fprintf (stderr, "Error: only -> allowed in impl. rule.\n") ;
				fprintf (stderr, 
				    "       impl. rule number %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    rule_direction = L_TO_R1;
			}	
    ;

input_list :  	   
		    {{
		      RULE_EXPR *current_exp ;
		      int current_expr = expr_stack[top] ;
		      current_exp = expr_table [current_expr] ;
		      current_exp -> input[current_exp -> arity] = NULL ;
		    }}
    | 	NIL	    {{
		      RULE_EXPR *current_exp, *input_ex ;
		      int current_expr = expr_stack[top] ;
		      current_exp = expr_table [current_expr] ;
		      input_ex = make_expr (-1,
		                            -1,
			                    (char *) NULL, (char *) NULL,
		                            (FUNC) IS_NULL,
					    level) ;
		      current_exp -> input[current_exp -> arity] = input_ex ;
		      input_ex -> parent = current_exp ;
		      if (current_side == LHS_CURRENT)
			  ++num_nodes_lhs ;
		      ++(current_exp -> arity) ; 
		    }}
      	input_list  
		    
    | 	input_var   {{ 
		      RULE_EXPR *current_exp, *input_ex ;
		      int current_expr = expr_stack[top] ;
		      current_exp = expr_table [current_expr] ;
		      input_ex = make_expr (symbol_table[$1].id,
		                                -1,
			                         symbol_table[$1].name, (char *) NULL,
		                                (FUNC) symbol_table[$1].func,
						level) ;
		      current_exp -> input[current_exp -> arity] = input_ex ;
		      input_ex -> parent = current_exp ;
		      ++(current_exp -> arity) ; 
		      if (current_side == LHS_CURRENT)
			  ++num_nodes_lhs ;
		    }}
	input_list   
		    
    |	expr    {{
		      RULE_EXPR *current_exp, *input_ex ;
		      int current_expr = expr_stack[top] ;
		      current_exp = expr_table [current_expr] ;
		      input_ex = expr_table [$1] ;
		      input_ex -> parent = current_exp ;
		      current_exp -> input[current_exp -> arity] = input_ex ;
		      ++(current_exp -> arity) ; 
		    }}
	input_list 
    ;

input_var :
	INPUT_VAR
		    {{
			int i, j = 0 ;
			for (i = sym_tab_index; ++i<=var_tab_index;)
			{
			    if (strcmp (symbol_table[i].name, yytext) == 0)
			    { 
				j = 1 ; $$ = i ; break ;
			    }
			}
			if (j == 0)
			{
			    if (current_side == RHS_CURRENT)
			    {
				fprintf (stderr,
				 "Error: Input var. %s used in RHS of rule not in LHS.\n",
				  yytext) ; 
				if (current_rule_type == TRANS_RULE)
				    fprintf (stderr, 
				 "       trans. rule %d.\n", trans_rule_num) ;
				else
				    fprintf (stderr, 
				 "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    ++var_tab_index ;
			    check_limit (var_tab_index, MAX_SYMBOLS) ;
			    strcpy (symbol_table[var_tab_index].name, yytext) ;
			    symbol_table[var_tab_index].id = ++input_var_id ;
			    symbol_table[var_tab_index].bind_id = -1 ;
			    symbol_table[var_tab_index].func = IS_INPUT_VARIABLE ;
			    symbol_table[var_tab_index].arity = -1 ;
			    symbol_table[var_tab_index].rhs_count = 0 ;
			    $$ = var_tab_index ;
			}
			else
			{
			    if (current_side == LHS_CURRENT)
			    {
				fprintf (stderr,
				 "Error: input var. %s used twice in LHS.\n",
				  yytext) ; 
				if (current_rule_type == TRANS_RULE)
				    fprintf (stderr, 
				 "       trans. rule %d.\n", trans_rule_num) ;
				else
				    fprintf (stderr, 
				 "       impl. rule %d.\n", impl_rule_num) ;
				exit (-1) ;
			    }
			    else
			    {
				++(symbol_table[i].rhs_count) ;
				if (symbol_table[i].rhs_count > 1)
				{
				    fprintf (stderr,
				     "Error: input var. %s used twice in RHS.\n",
				      yytext) ; 
				    if (current_rule_type == TRANS_RULE)
					fprintf (stderr, 
				 "       trans. rule %d.\n", trans_rule_num) ;
				    else
					fprintf (stderr, 
				 "       impl. rule %d.\n", impl_rule_num) ;
				    exit (-1) ;
				}
			    }
			}
		    }}  
    ;
