/*----------------------------- dbi.c -----------------*/
#include "opt_defs.h"
#include "global.h"
#include "dbi_gen.h"
#include "dbi.h"
#include "opt.h" 
#include <errno.h>

#include <string.h>

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif
#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

/* Selectivity in cases where we cannot determine it from other sources */
#define SELECTIVITY     0.1

/* DBI has to define zero cost and infinite_cost to be variables. */
COST zero_cost = { 0.0 };
COST infinite_cost = {9999999999999999999.0};

//extern double  log ( double) ; 
//extern double  ceil ( double) ; 
extern int  indent_value ; 

/* Global Variables */
char *Internal_set = "InternalSet";

/* DBI cost constants. */
#define PAGE_SIZE     4096              /* page size in bytes */
/* costs are in miliseconds */
#define SEEK          16     /* time to seek page on disk */
#define EBT           0.84   /* time to transfer page from disk */
#define ROTATION      8.3    /* rotational latency */
#define COPY_COST     0.0001 /* time to copy 1 byte */
#define TRANSMIT_COST 0.01   /* time to transmit 1 byte on network */
#define RANDOM        25.1 
//#define MEM           100   /* available memory for each operation */
#define MAX_COST      999999999999999999.0

#if defined(__cplusplus) || defined(ANSI)
 extern char * error_text (int ) ;
#endif

extern int MEM;

void Print_Join_Arg( JOIN_ARG *p);

/****************************************************************************
* CNMR()
*
* This routine finds the unique values of the attributes after selection
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
int CNMR(int n, int m, int r) 
#else
int CNMR(n,m,r)
int n,m,r;
#endif
{
   if (r<m/2) return(r);
   if (m/2<=r && r<=2*m) return((int) ceil(1+((double)(r+m))/3));
   return(m);	
}

/****************************************************************************
* Find_Join_Arg 
*
* This routine finds the join argument having an attribute.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
JOIN_ARG  *Find_Join_Arg(JOIN_ARG *p)
#else
JOIN_ARG  *Find_Join_Arg(p)
JOIN_ARG *p;
#endif
{
   JOIN_ARG *q;

    if (p==NULL) return (NULL);
    if (p->type == ATTRIB) return(p);
    q=Find_Join_Arg(p->left);
    if ( q != NULL) return(q);
    q=Find_Join_Arg(p->right);
    return(q);
}

/****************************************************************************
* derive_JOIN_sys_prop
*
* This routine determines the resulting system properties of a join, given
* the input operator argument, input stream system properties and the input
* stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_JOIN_sys_prop(
OPERATOR_ARGUMENT *oa,
SYSTEM_PROPERTY   *input_sys_prop[],
LOGICAL_PROPERTY  *input_log_prop[],
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_JOIN_sys_prop(oa, input_sys_prop, input_log_prop, 
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_JOIN_sys_prop */

/****************************************************************************
* derive_FIX_sys_prop
*
* This routine determines the resulting system properties of a fix, given
* the input operator argument, input stream system properties and the input
* stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_FIX_sys_prop(
OPERATOR_ARGUMENT *oa,
SYSTEM_PROPERTY   *input_sys_prop[],
LOGICAL_PROPERTY  *input_log_prop[],
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_FIX_sys_prop(oa, input_sys_prop, input_log_prop,
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_FIX_sys_prop */


/****************************************************************************
* derive_UNION_sys_prop
*
* This routine determines the resulting system properties of a union, given
* the input operator argument, input stream system properties and the input
* stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_UNION_sys_prop(
OPERATOR_ARGUMENT *oa,
SYSTEM_PROPERTY   *input_sys_prop[],
LOGICAL_PROPERTY  *input_log_prop[],
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_UNION_sys_prop(oa, input_sys_prop, input_log_prop,
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_UNION_sys_prop */

/****************************************************************************
* derive_MAT_sys_prop
*
* This routine determines the resulting system properties of a materialize,
* given the input operator argument, input stream system properties and
* the input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI) 
STATUS derive_MAT_sys_prop(
OPERATOR_ARGUMENT *oa, 
SYSTEM_PROPERTY   *input_sys_prop[], 
LOGICAL_PROPERTY  *input_log_prop[], 
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_MAT_sys_prop(oa, input_sys_prop, input_log_prop,
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_MAT_sys_prop */

/****************************************************************************
* derive_UNNEST_sys_prop
*
* This routine determines the resulting system properties of a unnest, given
* the input operator argument, input stream system properties and the input
* stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_UNNEST_sys_prop(
OPERATOR_ARGUMENT *oa,
SYSTEM_PROPERTY   *input_sys_prop[],
LOGICAL_PROPERTY  *input_log_prop[],
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_UNNEST_sys_prop(oa, input_sys_prop, input_log_prop,
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_UNNEST_sys_prop */

/****************************************************************************
* derive_SELECT_sys_prop
*
* This routine determines the resulting system properties of a select, given
* the input operator argument, input stream system properties and the input
* stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI) 
STATUS derive_SELECT_sys_prop(
OPERATOR_ARGUMENT *oa, 
SYSTEM_PROPERTY   *input_sys_prop[], 
LOGICAL_PROPERTY  *input_log_prop[], 
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_SELECT_sys_prop(oa, input_sys_prop, input_log_prop, 
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_SELECT_sys_prop */

/****************************************************************************
* derive_GET_SET_sys_prop
*
* This routine determines the resulting system properties of a get set, given
* the input operator argument, input stream system properties and the input
* stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI) 
STATUS derive_GET_SET_sys_prop(
OPERATOR_ARGUMENT *oa, 
SYSTEM_PROPERTY   *input_sys_prop[], 
LOGICAL_PROPERTY  *input_log_prop[], 
SYSTEM_PROPERTY   *derived_sys_prop)
#else
STATUS derive_GET_SET_sys_prop(oa, input_sys_prop, input_log_prop, 
        derived_sys_prop)
OPERATOR_ARGUMENT *oa;
SYSTEM_PROPERTY   *input_sys_prop[];
LOGICAL_PROPERTY  *input_log_prop[];
SYSTEM_PROPERTY   *derived_sys_prop;
#endif
{
    return (OKAY) ;
} /* derive_GET_SET_sys_prop */


/****************************************************************************
* print_log_prop
****************************************************************************/
void print_log_prop( LOGICAL_PROPERTY *lp )
{
      FILE *fp;
      char buffer[100];
      int  m;
      
      fp = fopen("schema","at");
      sprintf(buffer,"%s %d %d %d %d\n",lp->name,lp->set_id,
                                        lp->arity,lp->cardinality,
                                        lp->record_width);
      fputs(buffer,fp);
      for (m = 0; m != lp->arity; m++) {
         sprintf(buffer,"%s %d %d %d %d %d %d %d %d %f %f\n",
                         lp->attr[m]->name,
                         lp->attr[m]->attr_type,
                         lp->attr[m]->set_id,
                         lp->attr[m]->domain_min.int_value,
                         lp->attr[m]->domain_max.int_value,
                         lp->attr[m]->unique,
                         lp->attr[m]->attr_width,
                         lp->attr[m]->offset,
                         lp->attr[m]->use,
                         lp->attr[m]->fan,
                         lp->attr[m]->totref);
         fputs(buffer,fp);
      }
      fclose(fp);
}


/****************************************************************************
* derive_JOIN_log_prop
*
* This routine determines the resulting logical properties of a join,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI) 
STATUS derive_JOIN_log_prop(
OPERATOR_ARGUMENT *oa,
LOGICAL_PROPERTY  *input_log_prop[],
LOGICAL_PROPERTY  *derived_log_prop)
#else
STATUS derive_JOIN_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
   int        k = 0, offs = 0, CardZero = 0; 
   int        i,j,max_unique,skip_size,s_i,m;
   double     min_max,max_min,d, d0, d1;
   int        u1,u0;        /* resulting input unuque values */ 
   double     c0,c1;        /* resulting input cardinalities */
   ATTRIBUTE  a0, a1;       /* join attributes of inputs */
   char       tstr[MAX_ATTR_NAME_LEN];
   JOIN_ARG   *ja;          /* pointer to the join argument attribute node */
   JOIN_ARG   *current;
   int l=0;

   /* Sanity check - insure no null pointers */
   if (oa == NULL || derived_log_prop == NULL || input_log_prop == NULL)
       return (ERR_NULL_POINTER) ;

   /* Initialize the derived logical property vector */
   init_logical_prop(derived_log_prop);
   derived_log_prop->type = EXTENT;

   strcpy(derived_log_prop->name, input_log_prop[0]->name);
   	

/* Burayi certan icin yazdim 

   i = strlen(input_log_prop[0]->name);
   j = strlen(input_log_prop[1]->name);
   m=u1=u0=0;
   for (m=0;m<i+j;m++) {
      if (u0==i) {
           derived_log_prop->name[m]=input_log_prop[1]->name[u1];
           u1++;
      }
      else if (u1==j) {
           derived_log_prop->name[m]=input_log_prop[0]->name[u0];
           u0++;
      }
      else if (input_log_prop[0]->name[u0]<input_log_prop[1]->name[u1]){
           derived_log_prop->name[m]=input_log_prop[0]->name[u0];
           u0++;
      }
      else {
           derived_log_prop->name[m]=input_log_prop[1]->name[u1];
           u1++;
      }
   }
   derived_log_prop->name[m+1]=(char)0;
 buraya kadar */


   derived_log_prop->record_width = 0;
   derived_log_prop->arity =0;
   derived_log_prop->set_id = 1;
   s_i = input_log_prop[0]->attr[0]->set_id;

   /* Copy over all attributes from both input streams */
   for (i = 0; i != 2; i++)
   {
       for (j = 0; j != input_log_prop[i]->arity; j++)
       {
           if (input_log_prop[i]->attr[j]->use) {
              derived_log_prop->attr[k]= new ATTRIBUTE;
              *(derived_log_prop->attr[k]) = *(input_log_prop[i]->attr[j]);
              derived_log_prop->attr[k]->offset=derived_log_prop->record_width;
              derived_log_prop->attr[k]->btree_avail = FALSE;
              derived_log_prop->arity++;
              derived_log_prop->record_width += input_log_prop[i]->attr[j]->attr_width;
              k++;
           }
       }
   }

   current = &(oa->join);
   while (1) {
      if (current->data.op ==AND_OP) {
      
          /* Find the attribute node on the left hand side */
          ja = Find_Join_Arg(current->left->left);

          /* find the left join attribute in the input set */
          for(i=0; i<input_log_prop[0]->arity; i++)
             if (strcmp( ja->data.attr,input_log_prop[0]->attr[i]->name)==0)
                 a0 = *(input_log_prop[0]->attr[i]);

/********************* Burasi method projection icin eklendi **********/
          if (a0.attr_type == METHOD ) {
             for (i=0;i<derived_log_prop->arity;i++) {
                if (derived_log_prop->attr[i]->set_no == a0.set_no)
                   derived_log_prop->attr[i]->use --;
             }
          }
          else {
             for ( i=0;i<derived_log_prop->arity; i++) {
               if ( strcmp(a0.name,derived_log_prop->attr[i]->name)==0)
                  derived_log_prop->attr[i]->use--;
             }
          }
/********************* Buraya kadar **********/

          /* Find the attribute node on the right hand side */
          ja = Find_Join_Arg(current->left->right);

          /* find the right join attribute in the input set */
          for(j=0; j<input_log_prop[1]->arity; j++)
            if (strcmp( ja->data.attr,input_log_prop[1]->attr[j]->name)==0)
                 a1 = *(input_log_prop[1]->attr[j]);

/********************* Burasi method projection icin eklendi **********/
          if (a1.attr_type == METHOD ) {
             for (i=0;i<derived_log_prop->arity;i++) {
                if (derived_log_prop->attr[i]->set_no == a1.set_no)
                   derived_log_prop->attr[i]->use --;
             }
          }
          else {
             for ( i=0;i<derived_log_prop->arity; i++) {
               if ( strcmp(a1.name,derived_log_prop->attr[i]->name)==0)
                  derived_log_prop->attr[i]->use--;
             }
          }
/********************* Buraya kadar **********/

/* skip the attributes if their use is 0 */
          i=0;
          while (i<derived_log_prop->arity) {
              if ( derived_log_prop->attr[i]->use < 1 ) {
                 skip_size = derived_log_prop->attr[i]->attr_width;
                 derived_log_prop->record_width -= skip_size;
                 for ( j=i;j<derived_log_prop->arity-1;j++) {
                     derived_log_prop->attr[j] = derived_log_prop->attr[j+1];
                     derived_log_prop->attr[j]->offset -= skip_size;
                 }
                 derived_log_prop->arity --;
              }
              else i++;
          }


          /* if the join if pointer join ( converted from materialize) and the 
          pointer is on the left derived cardinality is the same as input 1 */
          i = strlen(a0.name) - 4;
          if (strcmp(&(a0.name[i]),"self")==0 )  {
             derived_log_prop->cardinality = (int)(a1.fan * input_log_prop[1]->cardinality);
             if ( (!derived_log_prop->cardinality) || CardZero ) 
                CardZero = 1;
             else if (input_log_prop[1]->cardinality>derived_log_prop->cardinality) {
                for(i=0; i< derived_log_prop->arity;i++) 
                   if (s_i != derived_log_prop->attr[i]->set_id ) {
                      derived_log_prop->attr[i]->unique = CNMR(
	    	            input_log_prop[1]->cardinality,
	    	            derived_log_prop->attr[i]->unique,
	                    derived_log_prop->cardinality);
                   }
             }
             current=current->right;
             continue;

          }

          /* if the join if pointer join ( converted from materialize)an the 
         pointer is on the right derived cardinality is the same as input 0 */
          i = strlen(a1.name) - 4;
          if (strcmp(&(a1.name[i]),"self")== 0 ) {
             derived_log_prop->cardinality = (int)(a0.fan*input_log_prop[0]->cardinality);
             if ( (!derived_log_prop->cardinality) || CardZero )
                CardZero = 1;
             else if (input_log_prop[0]->cardinality>derived_log_prop->cardinality){
                for(i=0; i<derived_log_prop->arity;i++)  {
                   if (s_i == derived_log_prop->attr[i]->set_id ) {	
	                derived_log_prop->attr[i]->unique = CNMR(
	   	               input_log_prop[0]->cardinality,
	   	               derived_log_prop->attr[i]->unique,
		               derived_log_prop->cardinality);
                   }
                 }
              }
             current=current->right;
             continue;

          }
   

          if ( a0.attr_type == TEXT  || a0.attr_type == REF ) {
             max_unique = max(a0.unique,a1.unique);
             max_unique = 2000;
             derived_log_prop->cardinality = (int)((float)input_log_prop[0]->cardinality*
                                  ((float)input_log_prop[1]->cardinality/max_unique)); 
             if ( (!derived_log_prop->cardinality) || CardZero ) {
                CardZero = 1;
                current = current->right;
                continue;
             }
             for(i=0; i<derived_log_prop->arity;i++)  {
                if ( s_i  == derived_log_prop->attr[i]->set_id ) {	
	            derived_log_prop->attr[i]->unique = CNMR(
	                input_log_prop[0]->cardinality,
	                derived_log_prop->attr[i]->unique,
	                (int)(derived_log_prop->cardinality*1.0
                             /input_log_prop[1]->cardinality*a1.unique));
                 }
                 else {	
	            derived_log_prop->attr[i]->unique = CNMR(
	                 input_log_prop[1]->cardinality,
	                 derived_log_prop->attr[i]->unique,
	                 (int)(derived_log_prop->cardinality*1.0
                               /input_log_prop[0]->cardinality*a0.unique));
                 }
             }
             current=current->right;
             continue;
          }
          if ( a0.attr_type == METHOD  || a1.attr_type == METHOD ) {
             derived_log_prop->cardinality =
                 input_log_prop[0]->cardinality+input_log_prop[1]->cardinality;
             
             if ( !derived_log_prop->cardinality)
                CardZero = 1;
             current = current->right;
             continue;
          }
       
          if ( a0.attr_type == INTEGER ) {
             min_max=min(a0.domain_max.int_value,a1.domain_max.int_value);
             max_min=max(a0.domain_min.int_value,a1.domain_min.int_value);
             d0 = a0.domain_max.int_value - a0.domain_min.int_value+1;
             d1 = a1.domain_max.int_value - a1.domain_min.int_value+1;
             d = min_max - max_min+1;
          }
          else {
             min_max=min(a0.domain_max.real_value,a1.domain_max.real_value);
             max_min=max(a0.domain_min.real_value,a1.domain_min.real_value);
             d0 =a0.domain_max.real_value-a0.domain_min.real_value+0.000001;
             d1 =a1.domain_max.real_value-a1.domain_min.real_value+0.000001;
             d = min_max - max_min+0.000001;
          }
       
          if ( (min_max < max_min) || CardZero ) {
             CardZero =1;
             derived_log_prop -> cardinality = 0;
             current = current->right;
             continue;
          }
          
          d0 = d / d0;
          d1 = d / d1;
          c0 = (d0 * input_log_prop[0]->cardinality);
          c1 = (d1 * input_log_prop[1]->cardinality);
          u0 = (int)ceil(d0 * a0.unique);
          u1 = (int)ceil(d1 * a1.unique);
          derived_log_prop->cardinality=(int)ceil(c0*(c1*1.0/max(u1,u0)));
          if ( !derived_log_prop->cardinality  ) {
                CardZero=1;
                current = current->right;
                continue;
          }
          for(i=0; i<derived_log_prop->arity;i++)  {
            if (s_i  == derived_log_prop->attr[i]->set_id ) {	
              if (strcmp(derived_log_prop->attr[i]->name,a0.name)==0){
                if (a0.attr_type == INTEGER) {
                  derived_log_prop->attr[i]->domain_max.int_value=(int)min_max;
                  derived_log_prop->attr[i]->domain_min.int_value=(int)max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
                else if (a0.attr_type == REAL) {
                  derived_log_prop->attr[i]->domain_max.real_value=min_max;
                  derived_log_prop->attr[i]->domain_min.real_value=max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
              }
              else {
	         derived_log_prop->attr[i]->unique = CNMR(
	            input_log_prop[0]->cardinality,
	            derived_log_prop->attr[i]->unique,
	            (int)(derived_log_prop->cardinality*1.0
                         /input_log_prop[1]->cardinality*a1.unique));
              }
            }
            else {	
              if (strcmp(derived_log_prop->attr[i]->name,a1.name)==0){
                if (a1.attr_type == INTEGER) {
                  derived_log_prop->attr[i]->domain_max.int_value=(int)min_max;
                  derived_log_prop->attr[i]->domain_min.int_value=(int)max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
                else if (a1.attr_type == REAL) {
                  derived_log_prop->attr[i]->domain_max.real_value=min_max;
                  derived_log_prop->attr[i]->domain_min.real_value=max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
              }
              else {
	        derived_log_prop->attr[i]->unique = CNMR(
	             input_log_prop[1]->cardinality,
	             derived_log_prop->attr[i]->unique,
	             (int)(derived_log_prop->cardinality*1.0
                           /input_log_prop[0]->cardinality*a0.unique));
              }
            }
          }
          current=current->right;
          continue;
      }
      else {
          /* Find the attribute node on the left hand side */
          ja = Find_Join_Arg(current->left);
       
          /* find the left join attribute in the input set */
          for(i=0; i<input_log_prop[0]->arity; i++)
             if (strcmp( ja->data.attr,input_log_prop[0]->attr[i]->name)==0)
                 a0 = *(input_log_prop[0]->attr[i]);

/********************* Burasi method projection icin eklendi **********/
          if (a0.attr_type == METHOD ) {
             for (i=0;i<derived_log_prop->arity;i++) {
                if (derived_log_prop->attr[i]->set_no == a0.set_no)
                   derived_log_prop->attr[i]->use --;
             }
          }
          else {
             for ( i=0;i<derived_log_prop->arity; i++) {
               if ( strcmp(a0.name,derived_log_prop->attr[i]->name)==0)
                  derived_log_prop->attr[i]->use--;
             }
          }
/********************* Buraya kadar **********************/


          /* Find the attribute node on the right hand side */
          ja = Find_Join_Arg(current->right);

          /* find the right join attribute in the input set */
          for(j=0; j<input_log_prop[1]->arity; j++)
             if (strcmp( ja->data.attr,input_log_prop[1]->attr[j]->name)==0)
                 a1 = *(input_log_prop[1]->attr[j]);

/********************* Burasi method projection icin eklendi **********/
          if (a1.attr_type == METHOD ) {
             for (i=0;i<derived_log_prop->arity;i++) {
                if (derived_log_prop->attr[i]->set_no == a1.set_no)
                   derived_log_prop->attr[i]->use --;
             }
          }
          else {
             for ( i=0;i<derived_log_prop->arity; i++) {
               if ( strcmp(a1.name,derived_log_prop->attr[i]->name)==0)
                  derived_log_prop->attr[i]->use--;
             }
          }
/********************* Buraya kadar **********/
         
/*   projected attributes are removed */
          i=0;
          while (i<derived_log_prop->arity) {
              if ( derived_log_prop->attr[i]->use < 1 ) {
                 skip_size = derived_log_prop->attr[i]->attr_width;
                 derived_log_prop->record_width -= skip_size;
                 for ( j=i;j<derived_log_prop->arity-1;j++) {
                     derived_log_prop->attr[j] = derived_log_prop->attr[j+1];
                     derived_log_prop->attr[j]->offset -= skip_size;
                 }
                 derived_log_prop->arity --;
              }
              else i++;
          }


          /* if the join if pointer join ( converted from materialize) and the 
          pointer is on the left derived cardinality is the same as input 1 */
          i = strlen(a0.name) - 4;
          if (strcmp(&(a0.name[i]),"self")==0 )  {
             derived_log_prop->cardinality = (int)(a1.fan * input_log_prop[1]->cardinality);
             if( (derived_log_prop->cardinality==0) || CardZero) 
               derived_log_prop->cardinality = 0;
             else if (input_log_prop[1]->cardinality>derived_log_prop->cardinality){
                for(i=0; i< derived_log_prop->arity;i++)
                   if (s_i != derived_log_prop->attr[i]->set_id ) {
                      derived_log_prop->attr[i]->unique = CNMR(
                            input_log_prop[1]->cardinality,
                            derived_log_prop->attr[i]->unique,
                            derived_log_prop->cardinality);
                      derived_log_prop->attr[i]->set_id = s_i;

                   }
             }

             return(OKAY);
          }

         /* if the join if pointer join ( converted from materialize)an the 
         pointer is on the right derived cardinality is the same as input 0 */
          i = strlen(a1.name) - 4;
          if (strcmp(&(a1.name[i]),"self")== 0 ) {
             derived_log_prop->cardinality =(int)(a0.fan*input_log_prop[0]->cardinality);
             if ( (derived_log_prop->cardinality == 0) || CardZero)
                derived_log_prop->cardinality = 0;
             else if (input_log_prop[0]->cardinality>derived_log_prop->cardinality){ 
                for(i=0; i< derived_log_prop->arity;i++) {
                   if (s_i == derived_log_prop->attr[i]->set_id ) {
                      derived_log_prop->attr[i]->unique = CNMR(
                            input_log_prop[0]->cardinality,
                            derived_log_prop->attr[i]->unique,
                            derived_log_prop->cardinality);
                   }
                   else 
                       derived_log_prop->attr[i]->set_id = s_i ;
                }
             }
             return(OKAY);

          }
   

          if ( a0.attr_type == TEXT  || a0.attr_type == REF ) {
             max_unique = max(a0.unique,a1.unique);
             max_unique = 2000;
             derived_log_prop->cardinality =
                   (int)((float) input_log_prop[0]->cardinality*
                        ((float)input_log_prop[1]->cardinality/max_unique)); 
             if ((!derived_log_prop->cardinality) || CardZero )  {
		derived_log_prop->cardinality = 0;
                for(i=0; i<derived_log_prop->arity;i++)  
                  derived_log_prop->attr[i]->set_id =s_i; 
                return (OKAY);
             }

             for(i=0; i<derived_log_prop->arity;i++)  {
                if ( s_i == derived_log_prop->attr[i]->set_id ) {
                    derived_log_prop->attr[i]->unique = CNMR(
                        input_log_prop[0]->cardinality,
                        derived_log_prop->attr[i]->unique,
                        (int)(derived_log_prop->cardinality*1.0/
                              input_log_prop[1]->cardinality*a1.unique));
                 }
                 else {
                    derived_log_prop->attr[i]->unique = CNMR(
                         input_log_prop[1]->cardinality,
                         derived_log_prop->attr[i]->unique,
                         (int)(derived_log_prop->cardinality*1.0
                             /input_log_prop[0]->cardinality*a0.unique));
                    derived_log_prop->attr[i]->set_id =s_i; 

                 }
             }
             return (OKAY);
          }
          if ( a0.attr_type == METHOD  || a1.attr_type == METHOD ) {
             for(i=0;i<derived_log_prop->arity;i++)
                derived_log_prop->attr[i]->set_id =s_i; 
             derived_log_prop->cardinality =
                 input_log_prop[0]->cardinality+input_log_prop[1]->cardinality;
             if ((!derived_log_prop->cardinality) || CardZero )
                 derived_log_prop->cardinality = 0;
             return(OKAY);
          }
       
          if ( a0.attr_type == INTEGER ) {
             min_max=min(a0.domain_max.int_value,a1.domain_max.int_value);
             max_min=max(a0.domain_min.int_value,a1.domain_min.int_value);
             d0 = a0.domain_max.int_value - a0.domain_min.int_value+1;
             d1 = a1.domain_max.int_value - a1.domain_min.int_value+1;
             d = min_max - max_min+1;
          }
          else {
             min_max=min(a0.domain_max.real_value,a1.domain_max.real_value);
             max_min=max(a0.domain_min.real_value,a1.domain_min.real_value);
             d0 =a0.domain_max.real_value-a0.domain_min.real_value+0.000001;
             d1 =a1.domain_max.real_value-a1.domain_min.real_value+0.000001;
             d = min_max - max_min+0.000001;
          }
 
          if ( (min_max < max_min) || CardZero ) {
             CardZero =1;
             derived_log_prop -> cardinality = 0;
  //           print_log_prop(derived_log_prop);
             for(i=0;i<derived_log_prop->arity;i++)
                derived_log_prop->attr[i]->set_id =s_i; 
             return(OKAY);
          }
       
          d0 = d / d0;
          d1 = d / d1;
          c0 = (d0 * input_log_prop[0]->cardinality);
          c1 = (d1 * input_log_prop[1]->cardinality);
          u0 = (int)ceil (d0 * a0.unique);
          u1 = (int)ceil (d1 * a1.unique);
          if (!(u0 || u1))
             u1 = 1;
          derived_log_prop->cardinality=(int)ceil(c0 * (c1*1.0/ max(u1,u0)));

          if (!derived_log_prop->cardinality)   {
       //      print_log_prop(derived_log_prop);
             for(i=0;i<derived_log_prop->arity;i++)
                derived_log_prop->attr[i]->set_id =s_i; 
              return(OKAY);
          }

          for(i=0; i<derived_log_prop->arity;i++)  {
            if ( s_i == derived_log_prop->attr[i]->set_id ) {
              if (strcmp(derived_log_prop->attr[i]->name,a0.name)==0){
                if (a0.attr_type == INTEGER) {
                  derived_log_prop->attr[i]->domain_max.int_value=(int)min_max;
                  derived_log_prop->attr[i]->domain_min.int_value=(int)max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
                else if (a0.attr_type == REAL) {
                  derived_log_prop->attr[i]->domain_max.real_value =min_max;
                  derived_log_prop->attr[i]->domain_min.real_value =max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
              }
              else {
                 derived_log_prop->attr[i]->unique = CNMR(
                     input_log_prop[0]->cardinality,
                     derived_log_prop->attr[i]->unique,
                     (int)(derived_log_prop->cardinality*1.0
                           /input_log_prop[1]->cardinality*a1.unique));
              }
            }
            else {
              if (strcmp(derived_log_prop->attr[i]->name,a1.name)==0){
                if (a1.attr_type == INTEGER) {
                  derived_log_prop->attr[i]->domain_max.int_value=(int)min_max;
                  derived_log_prop->attr[i]->domain_min.int_value=(int)max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
                else if (a1.attr_type == REAL) {
                  derived_log_prop->attr[i]->domain_max.real_value = min_max;
                  derived_log_prop->attr[i]->domain_min.real_value = max_min;
                  derived_log_prop->attr[i]->unique = min(u0,u1);
                }
              }
              else {
                derived_log_prop->attr[i]->unique = CNMR(
                     input_log_prop[1]->cardinality,
                     derived_log_prop->attr[i]->unique,
                     (int)(derived_log_prop->cardinality*1.0
                          /input_log_prop[0]->cardinality*a0.unique));
              }
              derived_log_prop->attr[i]->set_id = s_i; 
            }
          }
      //       print_log_prop(derived_log_prop);
          return(OKAY);
      }
   }
   /* That's all */
   return (OKAY) ;
}
/****************************************************************************
* derive_UNION_log_prop
*
* This routine determines the resulting logical properties of a union,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_UNION_log_prop(
OPERATOR_ARGUMENT *oa,
LOGICAL_PROPERTY  *input_log_prop[],
LOGICAL_PROPERTY  *derived_log_prop)
#else
STATUS derive_UNION_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
    int i;                              /* loop counter */
    int j;                              /* loop counter */
    int k = 0;                          /* index counter */
    int offs;
    char temp[MAX_ATTR_NAME_LEN];

    /* Sanity check - insure no null pointers */
    if (oa == NULL || derived_log_prop == NULL || input_log_prop == NULL)
        return (ERR_NULL_POINTER) ;

    /* Initialize the derived logical property vector */
    init_logical_prop(derived_log_prop);
    derived_log_prop->type = EXTENT;
    strcpy(derived_log_prop->name,input_log_prop[0]->name);
    strcat(derived_log_prop->name,input_log_prop[1]->name);
    derived_log_prop->record_width =input_log_prop[0]->record_width;
    derived_log_prop->arity = input_log_prop[0]->arity;
    derived_log_prop->cardinality = input_log_prop[0]->cardinality +input_log_prop[1]->cardinality;
    derived_log_prop->set_id = 1;

     for (i=0;i<input_log_prop[0]->arity;i++)  {
         derived_log_prop->attr[i]=new ATTRIBUTE;
         *(derived_log_prop->attr[i]) = *(input_log_prop[0]->attr[i]);
     }
     return(OKAY);
}

/****************************************************************************
* derive_FIX_log_prop
*
* This routine determines the resulting logical properties of a fix,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_FIX_log_prop(
OPERATOR_ARGUMENT *oa,
LOGICAL_PROPERTY  *input_log_prop[],
LOGICAL_PROPERTY  *derived_log_prop)
#else
STATUS derive_FIX_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
    int i;                              /* loop counter */
    int j;                              /* loop counter */
    int k = 0;                          /* index counter */
    int offs;
    char temp[MAX_ATTR_NAME_LEN];

    /* Sanity check - insure no null pointers */
    if (oa==NULL || derived_log_prop==NULL || input_log_prop==NULL)
        return (ERR_NULL_POINTER) ;

    /* Initialize the derived logical property vector */
    init_logical_prop(derived_log_prop);
    derived_log_prop->type = EXTENT;
    strcpy(derived_log_prop->name,input_log_prop[0]->name);
    derived_log_prop->record_width =0;
    derived_log_prop->arity = 0;
    derived_log_prop->set_id = 1;

    derived_log_prop->cardinality = input_log_prop[0]->cardinality;

   /* Copy over all attributes from both input streams */

   for (j = 0; j != input_log_prop[0]->arity; j++)
    {
       offs = -1;
       if ( strcmp(input_log_prop[0]->attr[j]->name,oa->fix.attr)==0) {
         offs =j;
         derived_log_prop->cardinality=(int)(input_log_prop[0]->attr[j]->fan*derived_log_prop->cardinality);
       }
       if ( input_log_prop[0]->attr[j]->use >0 ) {
           if ( offs ==j) {
              if (input_log_prop[0]->attr[j]->use > 1) {
                 derived_log_prop->attr[k] = new ATTRIBUTE;
                 *(derived_log_prop->attr[k]) = *(input_log_prop[0]->attr[j]);
                 derived_log_prop->attr[k]->use--;
                 derived_log_prop->attr[k]->btree_avail = FALSE;
                 derived_log_prop->attr[k]->offset=derived_log_prop->record_width;
                 derived_log_prop->record_width+=derived_log_prop->attr[k]->attr_width;
                 derived_log_prop->arity++ ;
                 k++;
             }
           }
           else {
              derived_log_prop->attr[k] = new ATTRIBUTE;
              *(derived_log_prop->attr[k]) = *(input_log_prop[0]->attr[j]);
              derived_log_prop->attr[k]->btree_avail = FALSE;
              derived_log_prop->attr[k]->offset=derived_log_prop->record_width;
              derived_log_prop->record_width+=derived_log_prop->attr[k]->attr_width;
              derived_log_prop->arity++ ;
              k++;
           }
        }
     }

    j= oa->fix.input.arity;
    for (j = 0; j != oa->fix.input.arity; j++) {
        if ( oa->fix.input.attr[j]->use > 0) {
           derived_log_prop->attr[k] = new ATTRIBUTE;
           *(derived_log_prop->attr[k]) = *(oa->fix.input.attr[j]);
           derived_log_prop->attr[k]->offset =  derived_log_prop->record_width;
           derived_log_prop->attr[k]->btree_avail = FALSE;
           strcpy(temp,oa->fix.attr);
           strcat(temp,".");
           strcat(temp,derived_log_prop->attr[k]->name);
           strcpy(derived_log_prop->attr[k]->name,temp);
           derived_log_prop->record_width += derived_log_prop->attr[k]->attr_width;
           derived_log_prop->arity++ ;

           k++;
        }
    }

    /* That's all */
    return (OKAY) ;
}


/****************************************************************************
* derive_MAT_log_prop
*
* This routine determines the resulting logical properties of a mat,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_MAT_log_prop(
OPERATOR_ARGUMENT *oa, 
LOGICAL_PROPERTY  *input_log_prop[], 
LOGICAL_PROPERTY  *derived_log_prop) 
#else 
STATUS derive_MAT_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
    int i,j;                              /* loop counter */
    int k = 0;                          /* index counter */
    char  set_no;
    int l=0,sl=0;
    int offs;
    int skip_size;
    int meth;
    char temp[MAX_ATTR_NAME_LEN];
    float fn;

    /* SANITY CHECK - INSURE NO NULL POINTERS */
    if (oa == NULL || derived_log_prop == NULL || input_log_prop == NULL)
        return (ERR_NULL_POINTER) ;

    /* Initialize the derived logical property vector */
    init_logical_prop(derived_log_prop);
    derived_log_prop->type = EXTENT;
    strcpy(derived_log_prop->name,input_log_prop[0]->name);


/* Burasi Certan Icin Eklendi
    i=0;
    j= strlen(input_log_prop[0]->name);
    while(i<j && input_log_prop[0]->name[i] < oa->mat.input[0].name[0]){
      derived_log_prop->name[i] = input_log_prop[0]->name[i];
      i++;
    }
    derived_log_prop->name[i]= oa->mat.input[0].name[0];
    while (i<j) {
       derived_log_prop->name[i+1]=input_log_prop[0]->name[i];
       i++;
    }
    derived_log_prop->name[i+1]=(char)0;
Buraya kadar*/

    derived_log_prop->record_width =0; 
    derived_log_prop->arity = 0;
    derived_log_prop->set_id = 1;
    derived_log_prop->cardinality = input_log_prop[0]->cardinality; 


   /* Copy over all attributes from both input streams */
   k=0;
   for (j = 0; j != input_log_prop[0]->arity; j++) {
     if (input_log_prop[0]->attr[j]->use) {
        derived_log_prop->attr[k]= new ATTRIBUTE;
        *(derived_log_prop->attr[k]) = *(input_log_prop[0]->attr[j]);
        derived_log_prop->attr[k]->offset=derived_log_prop->record_width;
        derived_log_prop->attr[k]->btree_avail = FALSE;
        derived_log_prop->arity++;
        derived_log_prop->record_width+=derived_log_prop->attr[k]->attr_width;
        k++;
     }
   }
   for (i=0;i<oa->mat.num;i++) {
     for (j = 0; j != oa->mat.input[i].arity; j++) {
       if (oa->mat.input[i].attr[j]->use) {
         derived_log_prop->attr[k]= new ATTRIBUTE;
         *(derived_log_prop->attr[k]) = *(oa->mat.input[i].attr[j]);
         strcpy(derived_log_prop->attr[k]->name,oa->mat.attr[i]);
         strcat(derived_log_prop->attr[k]->name,".");
         strcat(derived_log_prop->attr[k]->name,oa->mat.input[i].attr[j]->name);
         derived_log_prop->attr[k]->offset=derived_log_prop->record_width;
         derived_log_prop->attr[k]->btree_avail = FALSE;
         derived_log_prop->arity++;
         derived_log_prop->record_width+=derived_log_prop->attr[k]->attr_width;
         k++;
       }
     }
   }

   derived_log_prop->cardinality = 2000;
   for (j=0;j<oa->mat.num;j++) {
     for (i=0;i<derived_log_prop->arity;i++) {
       if ( ! strcmp(derived_log_prop->attr[i]->name , oa->mat.attr[j]) ){ 
         // derived_log_prop->cardinality *= derived_log_prop->attr[i]->fan;
/* 
         if ( derived_log_prop->attr[i]->attr_type == SET )         
            derived_log_prop->cardinality *= 3;   
*/
         if ( derived_log_prop->attr[i]->attr_type == METHOD ) {
            set_no =  derived_log_prop->attr[i]->set_no;
            for (k=0;k<derived_log_prop->arity;k++){
               if ( derived_log_prop->attr[k]->set_no == set_no )
                   derived_log_prop->attr[k]->use --;
            }
         }
         else 
            derived_log_prop->attr[i]->use --;
       }
     }
   }

/* skip the attributes if their use is 0 */
   i=0;
   while (i<derived_log_prop->arity) {
     if ( derived_log_prop->attr[i]->use < 1 ) {
        skip_size = derived_log_prop->attr[i]->attr_width;
        derived_log_prop->record_width -= skip_size;
        for ( j=i;j<derived_log_prop->arity-1;j++) {
           derived_log_prop->attr[j] = derived_log_prop->attr[j+1];
           derived_log_prop->attr[j]->offset -= skip_size;
        }
        derived_log_prop->arity --;
     }
     else i++;
   }

   return (OKAY) ;
}
/****************************************************************************
* derive_SELECT_log_prop
*
* This routine determines the resulting logical properties of a select,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_SELECT_log_prop(
OPERATOR_ARGUMENT *oa, 
LOGICAL_PROPERTY  *input_log_prop[], 
LOGICAL_PROPERTY  *derived_log_prop) 
#else 
STATUS derive_SELECT_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
    int i,j,k,m;                        /* loop counter */
    int attr,skip_size;                           /* attribute index */
    int minv,maxv;
    double fraction,frc;
    char tstr[MAX_ATTR_NAME_LEN];
    JOIN_ARG *ja;
    ATTRIBUTE *atr, *atr0;
    double min_max,max_min,max_max,min_min;
    int   CardZero = 0;

    /* Sanity check - insure no null pointers */
    if (oa == NULL || derived_log_prop == NULL || 
        input_log_prop == NULL)
        return (ERR_NULL_POINTER) ;

    init_logical_prop(derived_log_prop);
    derived_log_prop->type = EXTENT;
    strcpy(derived_log_prop->name,input_log_prop[0]->name);
    derived_log_prop->record_width =0;
    derived_log_prop->arity = 0;
    derived_log_prop->set_id = 1;

    derived_log_prop->cardinality = input_log_prop[0]->cardinality;

    if ( derived_log_prop->cardinality == 0)
       CardZero=1;
   /* Copy over all attributes from both input streams */
    k=0;
    for (j = 0; j != input_log_prop[0]->arity; j++)
    {
       if ( input_log_prop[0]->attr[j]->use >0 ) {
         derived_log_prop->attr[k] = new ATTRIBUTE;
         *(derived_log_prop->attr[k]) = *(input_log_prop[0]->attr[j]);
         derived_log_prop->attr[k]->btree_avail = FALSE;
         derived_log_prop->attr[k]->offset = derived_log_prop->record_width;
         derived_log_prop->record_width+=derived_log_prop->attr[k]->attr_width;
         derived_log_prop->arity++ ;
         k++;
       }
    }
    frc = 1.0;
    /* For now, this code assumes only conjunctions */
    for (i = 0; i != oa->select.terms; i++)
    {
       /* Set attr to the attribute index for this term */
       /* ja = Find_Join_Arg(oa->select.cond[i]->left); */
       ja = Find_Join_Arg(oa->select.cond[i]);
 
       attr= -1;
       for (j = 0; j < derived_log_prop->arity; j++)
       {
           if (strcmp( ja->data.attr, derived_log_prop->attr[j]->name) == 0) 
               attr = j;
       } /* for */
       if (attr == -1)
           FAIL(ERR_INVALID_VALUE);

/* BURASI METHOD PROJ ICIN YAZILDI */

       atr = derived_log_prop->attr[attr];
       if ( atr->attr_type== METHOD) {
          for ( j=0;j<derived_log_prop->arity;j++)
             if ( derived_log_prop->attr[j]->set_no == atr->set_no )
                derived_log_prop->attr[i]->use --;
       }
       else 
         atr->use-- ;

       if ( atr->attr_type == METHOD) {
	    frc = frc * SELECTIVITY;
	    continue;
	}
 
       ja = Find_Join_Arg(oa->select.cond[i]->right);
       /* This means that this selection is join */
       if ( ja != NULL ) {
         attr = -1;
	 for (j = 0; j < derived_log_prop->arity; j++)
         {
            if(strcmp( ja->data.attr, derived_log_prop->attr[j]->name) == 0)
            attr = j;
         } /* for */
         if (attr == -1)
             FAIL(ERR_INVALID_VALUE);

         atr0 =  derived_log_prop->attr[attr];
         if ( atr0->attr_type== METHOD) {
           for ( j=0;j<derived_log_prop->arity;j++)
              if ( derived_log_prop->attr[j]->set_no == atr0->set_no )
                 derived_log_prop->attr[i]->use --;
         }
         else
           atr0->use-- ;

         if ( atr0->attr_type == METHOD || atr0->attr_type == TEXT || 
              atr0->attr_type == REF) {
             frc = frc * SELECTIVITY;
             continue;
         }
         if ( atr->attr_type == INTEGER ) {
           min_max=min(atr0->domain_max.int_value,atr->domain_max.int_value);
           max_min=max(atr0->domain_min.int_value,atr->domain_min.int_value);
           max_max=max(atr0->domain_max.int_value,atr->domain_max.int_value);
           min_min=min(atr0->domain_min.int_value,atr->domain_min.int_value);
         }
         else {
           min_max=min(atr0->domain_max.real_value,atr->domain_max.real_value);
           max_min=max(atr0->domain_min.real_value,atr->domain_min.real_value);
           max_max=max(atr0->domain_max.real_value,atr->domain_max.real_value);
           min_min=min(atr0->domain_min.real_value,atr->domain_min.real_value);
         }
         frc = frc* (min_max - max_min) / (max_max - min_min);
         if ( frc < 0.0 ) frc *=(-1);
         continue;
       }

       /* Adjust the domain minimum and maximum based on the operator */
       switch (oa->select.cond[i]->data.op)
       {
       case EQUAL_OP:
           if (atr->attr_type == INTEGER || atr->attr_type==REAL) {
             atr->domain_min = oa->select.cond[i]->right->data.cnst.value;
             atr->domain_max = atr->domain_min; 
           }
           if (atr->unique)
              fraction = 1.0 / atr->unique;
           else
              fraction = 0.5;
           break;
       case LESS_THAN_OP:
           if (atr->attr_type == INTEGER)
           {
               fraction=((double)
                       (oa->select.cond[i]->right->data.cnst.value.int_value-
	   	       atr->domain_min.int_value-1))/
                       (atr->domain_max.int_value - atr->domain_min.int_value);

               atr->domain_max.int_value =
                       oa->select.cond[i]->right->data.cnst.value.int_value -1;
               break;
           }
           else if (atr->attr_type == REAL)
           {
              fraction=oa->select.cond[i]->right->data.cnst.value.real_value;
              fraction=atr->domain_min.real_value;
              fraction=atr->domain_max.real_value;
              fraction=fabs(((double)
                    (oa->select.cond[i]->right->data.cnst.value.real_value-
                     atr->domain_min.real_value ))/
                    (atr->domain_max.real_value - atr->domain_min.real_value));

               atr->domain_max.real_value =
                    oa->select.cond[i]->right->data.cnst.value.real_value;
               break;
           }
           else {
               fraction = SELECTIVITY;
               break;
           }
       case LESS_THAN_EQUAL_OP:
           if (atr->attr_type == INTEGER)
           {
               fraction=((double)
		(oa->select.cond[i]->right->data.cnst.value.int_value-
                        atr->domain_min.int_value))/
                       (atr->domain_max.int_value - atr->domain_min.int_value);

               atr->domain_max.int_value =
                       oa->select.cond[i]->right->data.cnst.value.int_value;
               break;
           }
           else if (atr->attr_type == REAL)
           {
               fraction=((double)
		(oa->select.cond[i]->right->data.cnst.value.real_value-
                     atr->domain_min.real_value))/
                     (atr->domain_max.real_value - atr->domain_min.real_value);

                atr->domain_max.real_value =    
                       oa->select.cond[i]->right->data.cnst.value.real_value;
               break;
           }
           else fraction = SELECTIVITY;
           atr->domain_max =
                       oa->select.cond[i]->right->data.cnst.value;
           break;
       case GREATER_THAN_OP:
           if (atr->attr_type == INTEGER)
           {
               fraction=((double)(atr->domain_max.int_value-
                  oa->select.cond[i]->right->data.cnst.value.int_value + 1 )) /
                  (atr->domain_max.int_value - atr->domain_min.int_value);

               atr->domain_min.int_value =
                     oa->select.cond[i]->right->data.cnst.value.int_value + 1;
                break;
           }
           else if (atr->attr_type == REAL)
           {
             fraction=((double)(atr->domain_max.real_value-
              oa->select.cond[i]->right->data.cnst.value.real_value+0.000001))/
                    (atr->domain_max.real_value - atr->domain_min.real_value); 
           }
           else fraction = SELECTIVITY;
           atr->domain_min =
                       oa->select.cond[i]->right->data.cnst.value;
               break;
       case GREATER_THAN_EQUAL_OP:
           if (atr->attr_type == INTEGER)
           {
              fraction=((double)(atr->domain_max.int_value-
                     oa->select.cond[i]->right->data.cnst.value.int_value )) /
                     (atr->domain_max.int_value - atr->domain_min.int_value); 
           }
           else if (atr->attr_type == REAL)
           {
               fraction=((double)(atr->domain_max.real_value- 
                  oa->select.cond[i]->right->data.cnst.value.real_value + 1 ))/
                  (atr->domain_max.real_value -  atr->domain_min.real_value);  
           } 
           else fraction = SELECTIVITY;
           atr->domain_min =
                   oa->select.cond[i]->right->data.cnst.value;
           break;
       default : fraction = SELECTIVITY;
       } /* switch */
       frc *=fraction;
       /* Note that we can't really handle NOT_EQUAL in the above switch */
    } /* for */

    i=0;
    while (i<derived_log_prop->arity) {
       if ( derived_log_prop->attr[i]->use < 1 ) {
          skip_size = derived_log_prop->attr[i]->attr_width;
          derived_log_prop->record_width -= skip_size;
          for ( j=i;j<derived_log_prop->arity-1;j++) {
             derived_log_prop->attr[j] = derived_log_prop->attr[j+1];
             derived_log_prop->attr[j]->offset -= skip_size;
          }
          derived_log_prop->arity --;
       }
       else i++;
    }      

    derived_log_prop->cardinality=(int)(input_log_prop[0]->cardinality*frc)+1;
    if (derived_log_prop->cardinality == 0) 
       return (OKAY);
    for (k=0; k<derived_log_prop->arity;k++) 
        derived_log_prop->attr[k]->unique = 
              CNMR(input_log_prop[0]->cardinality,
                   derived_log_prop->attr[k]->unique,
                   derived_log_prop->cardinality); 

    /* That's all */
    return (OKAY) ;
} 

/****************************************************************************
* derive_GET_SET_log_prop
*
* This routine determines the resulting logical properties of a get set,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_GET_SET_log_prop(
OPERATOR_ARGUMENT *oa, 
LOGICAL_PROPERTY  *input_log_prop[], 
LOGICAL_PROPERTY  *derived_log_prop) 
#else 
STATUS derive_GET_SET_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
    int i,j;
    char temp[MAX_ATTR_NAME_LEN];
    /* Sanity check - insure no null pointers */
    if (oa == NULL || derived_log_prop == NULL)
        return (ERR_NULL_POINTER) ;

    strcpy(derived_log_prop->name,oa->get_set.input.name);
    derived_log_prop->cardinality = oa->get_set.input.cardinality;
    derived_log_prop->set_id=oa->get_set.input.set_id;
    derived_log_prop->type = EXTENT;
    derived_log_prop->fid = oa->get_set.input.fid;
    if ( oa->get_set.use_self){
        derived_log_prop->record_width=sizeof(MOODS_REF);
        derived_log_prop->arity=1;
        derived_log_prop->attr[0] = new ATTRIBUTE;
        strcpy(derived_log_prop->attr[0]->name,oa->get_set.attr);
        strcat(derived_log_prop->attr[0]->name,".self");
        derived_log_prop->attr[0]->attr_type = REF;
        derived_log_prop->attr[0]->attr_width = sizeof(MOODS_REF);
        derived_log_prop->attr[0]->unique=oa->get_set.input.cardinality;
        derived_log_prop->attr[0]->btree_avail = FALSE;
        derived_log_prop->attr[0]->offset = 0;
        derived_log_prop->attr[0]->set_id = oa->get_set.input.set_id;
        derived_log_prop->attr[0]->use = 10;
        j=1;
        for (i=0;i<oa->get_set.input.arity;i++) {
           derived_log_prop->attr[j]=new ATTRIBUTE;
           *(derived_log_prop->attr[j]) = *(oa->get_set.input.attr[i]);
           derived_log_prop->attr[j]->offset +=  sizeof(MOODS_REF);
/*
           if (derived_log_prop->attr[j]->attr_type == METHOD) 
              derived_log_prop->attr[j]->offset +=  derived_log_prop->record_width;
           else if (derived_log_prop->attr[j]->attr_type == DB_TYPE)
              derived_log_prop->attr[j]->offset +=  sizeof(MOODS_REF);
           else
              derived_log_prop->attr[j]->offset =  derived_log_prop->record_width;
*/
           derived_log_prop->attr[j]->btree_avail= FALSE;
           strcpy(temp,oa->get_set.attr);
           strcat(temp,".");
           strcat(temp,derived_log_prop->attr[j]->name);
           strcpy(derived_log_prop->attr[j]->name,temp);
           derived_log_prop->record_width += derived_log_prop->attr[j]->attr_width; 
	   derived_log_prop->arity++ ;
           j++;
        }
    }
    else {
        derived_log_prop->record_width= oa->get_set.input.record_width;
        derived_log_prop->arity= oa->get_set.input.arity;
        for (i=0;i<oa->get_set.input.arity;i++)  {
           derived_log_prop->attr[i]=new ATTRIBUTE;
           *(derived_log_prop->attr[i]) = *(oa->get_set.input.attr[i]);
           strcpy(temp,oa->get_set.attr);
           strcat(temp,".");
           strcat(temp,oa->get_set.input.attr[i]->name);
           strcpy(derived_log_prop->attr[i]->name,temp);
        }
    }
    /* That's all */
    return (OKAY) ;
} 

/****************************************************************************
* derive_UNNEST_log_prop
*
* This routine determines the resulting logical properties of a unnest ,
* given the input operator argument and the  input stream logical properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_UNNEST_log_prop(
OPERATOR_ARGUMENT *oa, 
LOGICAL_PROPERTY  *input_log_prop[], 
LOGICAL_PROPERTY  *derived_log_prop) 
#else 
STATUS derive_UNNEST_log_prop(oa, input_log_prop, derived_log_prop)
OPERATOR_ARGUMENT *oa;
LOGICAL_PROPERTY  *input_log_prop[];
LOGICAL_PROPERTY  *derived_log_prop;
#endif
{
    int i,j,k;
    int skip_size;
    char set_no;
    char temp[MAX_ATTR_NAME_LEN];

    /* Sanity check - insure no null pointers */
    if (oa == NULL || derived_log_prop == NULL)
        return (ERR_NULL_POINTER) ;

    init_logical_prop(derived_log_prop);
    derived_log_prop->type = EXTENT;
    strcpy(derived_log_prop->name, Internal_set);
    derived_log_prop->record_width =0;
    derived_log_prop->arity = 0;
    derived_log_prop->set_id = 1;


    j=0;
    for (i=0;i<input_log_prop[0]->arity;i++) {
      if ( input_log_prop[0]->attr[i]->use ) {
         derived_log_prop->attr[j]=new ATTRIBUTE;
         *(derived_log_prop->attr[j]) = *(input_log_prop[0]->attr[i]);
         derived_log_prop->arity++;
         derived_log_prop->attr[j]->offset = derived_log_prop->record_width;
         derived_log_prop->record_width+=derived_log_prop->attr[j]->attr_width;
         j++;
      }
    }
    for (i=0;i<oa->unnest.input.arity;i++) {
      if ( oa->unnest.input.attr[i]->use ) {
         derived_log_prop->attr[j]=new ATTRIBUTE;
         *(derived_log_prop->attr[j]) = *(oa->unnest.input.attr[i]);
         strcpy(derived_log_prop->attr[j]->name,oa->unnest.attr);
         strcat(derived_log_prop->attr[j]->name,".");
         strcat(derived_log_prop->attr[j]->name,oa->unnest.input.attr[i]->name);
         derived_log_prop->arity++;
         derived_log_prop->attr[j]->offset = derived_log_prop->record_width;
         derived_log_prop->record_width+=derived_log_prop->attr[j]->attr_width;
         j++;
      }
    }
    for (i=0;i<derived_log_prop->arity;i++) {
      if ( ! strcmp(derived_log_prop->attr[i]->name , oa->unnest.attr) ){
        if ( derived_log_prop->attr[i]->attr_type == METHOD ) {
           set_no =  derived_log_prop->attr[i]->set_no;
           for (k=0;k<derived_log_prop->arity;k++){
              if ( derived_log_prop->attr[k]->set_no == set_no )
                  derived_log_prop->attr[k]->use --;
           }
        }
        else
           derived_log_prop->attr[i]->use --;
      }
    }

/* skip the attributes if their use is 0 */
   i=0;
   while (i<derived_log_prop->arity) {
     if ( derived_log_prop->attr[i]->use < 1 ) {
        skip_size = derived_log_prop->attr[i]->attr_width;
        derived_log_prop->record_width -= skip_size;
        for ( j=i;j<derived_log_prop->arity-1;j++) {
           derived_log_prop->attr[j] = derived_log_prop->attr[j+1];
           derived_log_prop->attr[j]->offset -= skip_size;
        }
        derived_log_prop->arity --;
     }
     else i++;
   }

    derived_log_prop->cardinality *=3;
    /* That's all */
    return (OKAY) ;
}

/****************************************************************************
* derive_FIX_ALG_phy_prop
*
* This routine determines the resulting physical properties of a fix alg.,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_FIX_ALG_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_FIX_ALG_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);
    init_property_vec(derived_pv);
    return (OKAY) ;
}

/****************************************************************************
* derive_UNION_ALG_phy_prop
*
* This routine determines the resulting physical properties of a hash join,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_UNION_ALG_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_UNION_ALG_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);
    init_property_vec(derived_pv);
    return (OKAY) ;
}

/****************************************************************************
* derive_HASH_JOIN_phy_prop
*
* This routine determines the resulting physical properties of a hash join,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_HASH_JOIN_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_HASH_JOIN_phy_prop(aa, input_pv, sys_prop, log_prop, 
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Initialize the derived property vector */
    init_property_vec(derived_pv);

    /* That's all */
    return (OKAY) ;
 }

/****************************************************************************
* derive_NESTED_LOOP_JOIN_phy_prop
*
* This routine determines the resulting physical properties of a hash join,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_NESTED_LOOP_JOIN_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_NESTED_LOOP_JOIN_phy_prop(aa, input_pv, sys_prop, log_prop, 
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Initialize the derived property vector */
    init_property_vec(derived_pv);

    /* That's all */
    return (OKAY) ;
 }

/*****************************************************************************
* derive_UNNEST_ALG_phy_prop
*
* This routine determines the resulting physical properties of a unnest,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_UNNEST_ALG_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_UNNEST_ALG_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Initialize the derived property vector */
    *derived_pv=*input_pv[0];

    /* That's all */
    return (OKAY) ;
}

/*****************************************************************************
* derive_PTR_HH_JOIN_phy_prop
*
* This routine determines the resulting physical properties of a unnest,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_PTR_HH_JOIN_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_PTR_HH_JOIN_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Initialize the derived property vector */
    *derived_pv=*input_pv[0];

    /* That's all */
    return (OKAY) ;
}

/*****************************************************************************
* derive_PTR_SM_JOIN_phy_prop
*
* This routine determines the resulting physical properties of a unnest,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_PTR_SM_JOIN_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_PTR_SM_JOIN_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Initialize the derived property vector */
    *derived_pv=*input_pv[0];

    /* That's all */
    return (OKAY) ;
}


/****************************************************************************
* derive_TRAVERSE_phy_prop
*
* This routine determines the resulting physical properties of traverse,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_TRAVERSE_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_TRAVERSE_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    *derived_pv=*input_pv[0];

    /* That's all */
    return (OKAY) ;
}

/****************************************************************************
* derive_MERGE_JOIN_phy_prop
*
* This routine determines the resulting physical properties of a merge join,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_MERGE_JOIN_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_MERGE_JOIN_phy_prop(aa, input_pv, sys_prop, log_prop, 
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;
    JOIN_ARG *ja;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);
    derived_pv->sort_arity = 2;
    ja = Find_Join_Arg(aa->merge_join.left);
    strcpy(derived_pv->sort_attr[0], ja->data.attr);
    ja = Find_Join_Arg(aa->merge_join.right);
    strcpy(derived_pv->sort_attr[1], ja->data.attr);
    /* That's all */
    return (OKAY) ;
} 

/****************************************************************************
* derive_FILTER_phy_prop
*
* This routine determines the resulting physical properties of a filter,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_FILTER_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_FILTER_phy_prop(aa, input_pv, sys_prop, log_prop, input_sys_prop, 
        input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* The property vector matches that of the input */
    *derived_pv = *input_pv[0];

    /* That's all */
    return (OKAY) ;
} 

/****************************************************************************
* derive_FILE_SCAN_phy_prop
*
* This routine determines the resulting physical properties of a file scan,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_FILE_SCAN_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_FILE_SCAN_phy_prop(aa, input_pv, sys_prop, log_prop, 
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Files are assumed unsorted */
    init_property_vec(derived_pv);

    /* That's all */
    return (OKAY) ;
} 

/****************************************************************************
* derive_BTREE_SELECT_phy_prop
*
* This routine determines the resulting physical properties of a btree scan,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_BTREE_SELECT_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_BTREE_SELECT_phy_prop(aa, input_pv, sys_prop, log_prop,
        input_sys_prop, input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;
    JOIN_ARG *ja;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* Set the sort attribute */
    derived_pv->sort_arity = 1;
    ja = Find_Join_Arg(aa->btree_select.cond);
    if (ja == NULL ) exit(8);
    strcpy(derived_pv->sort_attr[0], ja->data.attr);

    /* That's all */
    return (OKAY);
}

/****************************************************************************
* derive_SORT_phy_prop
*
* This routine determines the resulting physical properties of a sort,
* given the input algorithm argument, the input stream logical, system and
* physical properties, and the output logical and system properties.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS derive_SORT_phy_prop(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *input_pv[],
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv)
#else
STATUS derive_SORT_phy_prop(aa, input_pv, sys_prop, log_prop, input_sys_prop, 
        input_log_prop, needed_pv, derived_pv)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *input_pv[];
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
#endif
{
    int i;

    /* Sanity check */
    if (aa==NULL || input_pv==NULL || input_log_prop==NULL || derived_pv==NULL)
        return (ERR_NULL_POINTER);

    /* The sort property comes from the algorithm argument */
    derived_pv->sort_arity = aa->sort.sort_arity;
    for (i = 0; i != aa->sort.sort_arity; i++)
        strcpy(derived_pv->sort_attr[i], aa->sort.sort_attr[i]);

    /* That's all */
    return (OKAY);
} 
/****************************************************************************
* cost_UNION_ALG
*
* This routine determines the cost of a union .  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_UNION_ALG(
ALGORITHM_ARGUMENT *aa,
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv,
PROPERTY_VECTOR    *input_pv[],
COST               *cost )
#else
STATUS cost_UNION_ALG(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double  pages,pages1,pages2;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL) {
        derived_pv = needed_pv;
    } /* if */

    /* Find cost at each node */
    pages= 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages1= 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages2= 1.0/PAGE_SIZE*input_log_prop[1]->cardinality*input_log_prop[1]->record_width;
    cost->cost.io = (pages+pages1+pages2)*RANDOM+0.000001;
    return(OKAY);
}
   
/****************************************************************************
* cost_FIX_ALG
*
* This routine determines the cost of a union .  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_FIX_ALG(
ALGORITHM_ARGUMENT *aa,
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv,
PROPERTY_VECTOR    *input_pv[],
COST               *cost )
#else
STATUS cost_FIX_ALG(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double  pages,pages1,pages2;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL) {
        derived_pv = needed_pv;
    } /* if */

    /* Find cost at each node */
    pages= 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages1= 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages2= 1.0/PAGE_SIZE*aa->fix_alg.input.cardinality*aa->fix_alg.input.record_width;
    cost->cost.io = (pages+pages1+pages2)*RANDOM+0.000001;
    return(OKAY);
}


/****************************************************************************
* cost_MERGE_JOIN
*
* This routine determines the cost of a merge join.  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_MERGE_JOIN(
ALGORITHM_ARGUMENT *aa,
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv,
PROPERTY_VECTOR    *input_pv[],
COST               *cost )
#else
STATUS cost_MERGE_JOIN(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double  pages,pages1,pages2;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL) {
        derived_pv = needed_pv;
    } /* if */

    /* Find cost at each node */
    pages= 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages1= 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages2= 1.0/PAGE_SIZE*input_log_prop[1]->cardinality*input_log_prop[1]->record_width;
    if ((pages2==0.0)|| (pages2==0.0)){
       cost->cost.io = 0.000001;
       return(OKAY);
    }
   
    cost->cost.io = (pages+pages1+pages2)*RANDOM+0.000001;
    
    /* That's all */

    return (OKAY) ;
} /* cost_MERGE_JOIN */

/****************************************************************************
* cost_UNNEST_ALG
*
* This routine determines the cost of a merge join.  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_UNNEST_ALG(
ALGORITHM_ARGUMENT *aa,
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv,
PROPERTY_VECTOR    *input_pv[],
COST               *cost )
#else
STATUS cost_UNNEST_ALG(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double  pages;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL) {
        derived_pv = needed_pv;
    } /* if */

    pages  = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages += 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    cost->cost.io = RANDOM*pages+0.000001;

    /* That's all */
    return (OKAY) ;
} /* cost_UNNEST_ALG */


/****************************************************************************
* cost_HASH_JOIN
*
* This routine determines the cost of a hash join.  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_HASH_JOIN(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_HASH_JOIN(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double pages,pages1,pages2,nsg;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL)
    {
        derived_pv = needed_pv;
    } /* if */


    pages =1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages1=1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages2=1.0/PAGE_SIZE*input_log_prop[1]->cardinality*input_log_prop[1]->record_width;
    if ((pages1==0.0) || (pages2 == 0.0)) {
       cost->cost.io = 0.000001;
       return(OKAY);
    }
//    nsg = ceil(2.4*pages1/MEM);
//    cost->cost.io=3*RANDOM*(pages1+pages2)+(1.0+pages2/pages1)*nsg*nsg*(ROTATION+SEEK)+0.000001;
    cost->cost.io=3*RANDOM*(pages1+pages2);
    cost->cost.io += (pages*RANDOM);  
    /* That's all */
    return (OKAY) ;
} /* cost_HASH_JOIN */

/****************************************************************************
* cost_NESTED_LOOP_JOIN
*
* This routine determines the cost of a hash join.  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_NESTED_LOOP_JOIN(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_NESTED_LOOP_JOIN(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double pages,pages1,pages2,nsg;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL)
    {
        derived_pv = needed_pv;
    } /* if */

    pages = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages2= 1.0/PAGE_SIZE*input_log_prop[1]->cardinality*input_log_prop[1]->record_width;

    pages1= 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
 
    if ((pages1==0.0) || (pages2==0.0)) {
       cost->cost.io = 0.000001;
       return(OKAY);
    }
      nsg = ceil(2.4*pages1/MEM);
    cost->cost.io = pages2*RANDOM*nsg+pages1*RANDOM+0.000001;    
    cost->cost.io += (pages*RANDOM);  
  
    /* That's all */
    return (OKAY) ;
} /* cost_NESTED_LOOP_JOIN */

/****************************************************************************
* cost_PTR_HH_JOIN
*
* This routine determines the cost of a hash join.  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_PTR_HH_JOIN(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_PTR_HH_JOIN(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double pages;
    int  i;
    float fanc;
    char *temp;

    temp = aa->ptr_hh_join.attr;
    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL)
    {
        derived_pv = needed_pv;
    } /* if */

    fanc = 0.0;
    for(i=0;i<input_log_prop[0]->arity;i++) {
       temp = input_log_prop[0]->attr[i]->name;
       if (strcmp(input_log_prop[0]->attr[i]->name,aa->ptr_hh_join.attr)==0) 
           fanc = input_log_prop[0]->attr[i]->fan;
    }
    if ( fanc == 0.0 ) {
       cost->cost.io = MAX_COST;
       return(OKAY);
    }
       
    pages = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    cost->cost.io = pages+3.0/PAGE_SIZE* input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages = 1.0/PAGE_SIZE*aa->ptr_hh_join.input.cardinality*aa->ptr_hh_join.input.record_width;
    cost->cost.io +=CNMR(aa->ptr_hh_join.input.cardinality,
                         (int)pages,
                         log_prop->cardinality);
    cost->cost.io *= RANDOM;
    cost->cost.io += 0.000001;
    /* That's all */
    return (OKAY) ;
} /* cost_PTR_HH_JOIN */

/****************************************************************************
* cost_PTR_SM_JOIN
*
* This routine determines the cost of a hash join.  It assumes no overflow.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_PTR_SM_JOIN(
ALGORITHM_ARGUMENT *aa,
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv,
PROPERTY_VECTOR    *input_pv[],
COST               *cost )
#else
STATUS cost_PTR_SM_JOIN(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double pages;
    int  i;
    float fanc;
    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL)
    {
        derived_pv = needed_pv;
    } /* if */

    fanc = 0.0;
    for(i=0;i<input_log_prop[0]->arity;i++)
       if (strcmp(input_log_prop[0]->attr[i]->name,aa->ptr_hh_join.attr)==0) 
           fanc = input_log_prop[0]->attr[i]->fan;
    if ( fanc == 0.0 ) {
       cost->cost.io = MAX_COST;
       return(OKAY);
    }

    pages = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    cost->cost.io = pages+3.0/PAGE_SIZE* input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages = 1.0/PAGE_SIZE*aa->ptr_hh_join.input.cardinality*aa->ptr_hh_join.input.record_width;
    cost->cost.io +=CNMR(aa->ptr_hh_join.input.cardinality,
                         (int)pages,
                         log_prop->cardinality);
    cost->cost.io *=RANDOM;
    cost->cost.io += 0.00001;
    /* That's all */
    return (OKAY) ;
} /* cost_PTR_HH_JOIN */


/****************************************************************************
* cost_TRAVERSE
*
* This routine determines the cost of traverse 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_TRAVERSE(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_TRAVERSE(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double pages,pages1;
    float  totr;
    float fanc,K,A,slope;
    int   i,j;
    int ind;

    fanc = 0.0;
    for(i=0;i<input_log_prop[0]->arity;i++) {
       for ( j=0;j<aa->traverse.num;j++)
          if (strcmp(input_log_prop[0]->attr[i]->name,aa->traverse.attr[j])==0){
              fanc += input_log_prop[0]->attr[i]->fan;
              totr = input_log_prop[0]->attr[i]->totref;
      }
    }

/* Burasi duzeltilecek !  */
    totr = 10;
    fanc = 1000;
/* Buraya kadar !!! done by mehmet */
    if ( fanc == 0.0 ) {
       cost->cost.io = MAX_COST;
       return(OKAY);
    }

/* 

    for ( i=0; i<input_log_prop[0]->arity; i++ ) {
        if ( input_log_prop[0]->attr[i]->name == aa->traverse.attr[0]) {
           ind = i;
           break;
        }
    }
    if ( input_log_prop[0]->attr[ind]->attr_type == SET ) {
        
    }
*/

    pages = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    pages1= 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    cost->cost.io = (pages1+pages) *RANDOM;
    pages = 1.0/PAGE_SIZE*aa->traverse.input[0].cardinality*
            aa->traverse.input[0].record_width;
    A = input_log_prop[0]->cardinality*fanc;


    K = CNMR(aa->traverse.input[0].cardinality,(int)pages,(int)totr);
    if (K<MEM ) pages = K;
    else {
       slope =(K-A)/(K-1);
       pages = slope*(MEM-1)+A;
    }
    cost->cost.io += RANDOM*pages+ 0.00001;


    /* That's all */
    return (OKAY) ;
} /* cost_TRAVERSE*/


/****************************************************************************
* cost_FILTER
*
* This routine determines the cost of a filter operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_FILTER(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_FILTER(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double pages;
    int  i;
    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL)
    {
        derived_pv = needed_pv;
    } /* if */

    pages  = 1.0/PAGE_SIZE*input_log_prop[0]->cardinality*input_log_prop[0]->record_width;
    pages  += 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    cost->cost.io = pages*RANDOM;
    cost->cost.io +=0.000001;

    /* That's all */
    return (OKAY) ;
} /* cost_FILTER */

/****************************************************************************
* cost_FILE_SCAN
*
* This routine determines the cost of a file scan.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_FILE_SCAN(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_FILE_SCAN(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    long    pages;                      /* pages in file to scan */

/*    if (aa->file_scan.use_self) {
       pages = 1+log_prop->cardinality / (PAGE_SIZE / log_prop->record_width);
       cost->cost.io = (SEEK + EBT) * pages; 
    }
*/
    cost->cost.io = 0.000001;

    /* That's all */
    return (OKAY) ;
} /* cost_FILE_SCAN */


/****************************************************************************
* cost_BTREE_SELECT
*
* This routine determines the cost of a B+ tree scan.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_BTREE_SELECT(
ALGORITHM_ARGUMENT *aa,
SYSTEM_PROPERTY    *sys_prop,
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *derived_pv,
PROPERTY_VECTOR    *input_pv[],
COST               *cost )
#else
STATUS cost_BTREE_SELECT(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
   double    pages;

   cost->cost.io=log_prop->cardinality *0.6;
   pages = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
   cost->cost.io +=  pages+0.000001;
    cost->cost.io *=RANDOM;
   return (OKAY) ;
} /* cost_BTREE_SCAN */


/****************************************************************************
* cost_SORT
*
* This routine determines the cost of a sort.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS cost_SORT(
ALGORITHM_ARGUMENT *aa, 
SYSTEM_PROPERTY    *sys_prop, 
LOGICAL_PROPERTY   *log_prop,
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *derived_pv, 
PROPERTY_VECTOR    *input_pv[], 
COST               *cost )    
#else
STATUS cost_SORT(aa, sys_prop, log_prop, input_sys_prop, input_log_prop,
        needed_pv, derived_pv, input_pv, cost)
ALGORITHM_ARGUMENT *aa;
SYSTEM_PROPERTY    *sys_prop;
LOGICAL_PROPERTY   *log_prop;
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *derived_pv;
PROPERTY_VECTOR    *input_pv[];
COST               *cost ;
#endif
{
    double  pages,nsg;

    /* If derived_pv == NULL, then this is just an "estimated" cost. */
    if (derived_pv == NULL)
    {
        derived_pv = needed_pv;
    } /* if */

    pages = 1.0/PAGE_SIZE*log_prop->cardinality*log_prop->record_width;
    nsg = ceil(pages/MEM);
//    cost->cost.io = pages*2*RANDOM+2*nsg*nsg*(SEEK+ROTATION)+0.000001;
    cost->cost.io = pages*2*RANDOM;
    /* That's all */
    return (OKAY) ;
}

/****************************************************************************
* do_any_good_FIX_ALG
*
* This routine determines whether the given fix will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_FIX_ALG(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_FIX_ALG(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{

    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.fix_alg = oa->fix;
    ret_value->process_input_count = 1;

    return (OKAY);
}


/****************************************************************************
* do_any_good_UNION_ALG
*
* This routine determines whether the given merge join will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_UNION_ALG(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_UNION_ALG(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.un_alg = oa->un;
    ret_value->process_input_count = 1;

    return (OKAY);
}



/****************************************************************************
* do_any_good_MERGE_JOIN
*
* This routine determines whether the given merge join will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_MERGE_JOIN(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_MERGE_JOIN(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    JOIN_ARG *ja;

    ja = Find_Join_Arg(oa->join.left);

    if (excl_pv->sort_arity == 1 &&
            strcmp(excl_pv->sort_attr[0], ja->data.attr) == 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    }

    if (needed_pv->sort_arity > 1 ||
            (needed_pv->sort_arity == 1 &&
            strcmp(needed_pv->sort_attr[0], ja->data.attr) != 0))
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } 

    if (needed_pv->sort_arity != 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } /* if */
    /* We can use merge join, so set up the return values */
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.merge_join = oa->join;
    ret_value->process_input_count = 1;

    return (OKAY);
}       

/****************************************************************************
* do_any_good_HASH_JOIN
*
* This routine determines whether the given hash join will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_HASH_JOIN(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_HASH_JOIN(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;    
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{

    /* Hash join cannot meet any needed physical sort */
    if (needed_pv->sort_arity != 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } /* if */

   
    /* We can use hash join, so set up the return values */
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.hash_join = oa->join;
    ret_value->process_input_count = 1;
    
    return (OKAY);
}       


/****************************************************************************
* do_any_good_NESTED_LOOP_JOIN
*
* This routine determines whether the given hash join will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_NESTED_LOOP_JOIN(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_NESTED_LOOP_JOIN(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;    
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{

    /* Hash join cannot meet any needed physical sort */
    if (needed_pv->sort_arity != 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } /* if */

   
    /* We can use hash join, so set up the return values */
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.hash_join = oa->join;
    ret_value->process_input_count = 1;
    
    return (OKAY);
}       

/****************************************************************************
* do_any_good_PTR_HH_JOIN
*
* This routine determines whether the given hash join will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_PTR_HH_JOIN(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_PTR_HH_JOIN(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    JOIN_ARG *ja;

    ja = Find_Join_Arg(oa->join.left);
 
    /* Hash join has no excluded physical properties */

    /* Hash join cannot meet any needed physical sort */
    if (needed_pv->sort_arity != 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } /* if */

    /* We can use hash join, so set up the return values */
    ret_value->apply_truth_value = TRUE;
    strcpy(ret_value->algorithm_argument.ptr_hh_join.attr, ja->data.attr); 
    ret_value->process_input_count = 1;
 
    return (OKAY);
}
/****************************************************************************
* do_any_good_PTR_SM_JOIN
*
* This routine determines whether the given hash join will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_PTR_SM_JOIN(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_PTR_SM_JOIN(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    JOIN_ARG *ja;

    ja = Find_Join_Arg(oa->join.left);

    /* Hash join has no excluded physical properties */

    /* Hash join cannot meet any needed physical sort */
    if (needed_pv->sort_arity != 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } /* if */

    /* We can use hash join, so set up the return values */
    ret_value->apply_truth_value = TRUE;
    strcpy(ret_value->algorithm_argument.ptr_sm_join.attr, ja->data.attr);
    ret_value->process_input_count = 1;

    return (OKAY);
}
 

/****************************************************************************
* do_any_good_TRAVERSE
*
* This routine determines whether the given traverse will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_TRAVERSE(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_TRAVERSE(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    if (needed_pv->sort_arity != 0)
    {
        ret_value->apply_truth_value = FALSE;
        return (OKAY);
    } /* if */
 
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.traverse= oa->mat;
    ret_value->process_input_count = 1;
 
    return (OKAY);
}

/****************************************************************************
* do_any_good_UNNEST_ALG
*
* This unnest always done by unnest_alg
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_UNNEST_ALG(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_UNNEST_ALG(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    /* The filter can always be used */
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.unnest_alg= oa->unnest;
    ret_value->process_input_count = 1;

    return (OKAY);
}

/****************************************************************************
* do_any_good_FILTER
*
* This routine determines whether the given filter will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_FILTER(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_FILTER(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;    
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    /* The filter can always be used */
    ret_value->apply_truth_value = TRUE;
    ret_value->algorithm_argument.filter= oa->select;
    ret_value->process_input_count = 1;

    return (OKAY);
}

/****************************************************************************
* do_any_good_FILE_SCAN
*
* This routine determines whether the given file scan will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_FILE_SCAN(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_FILE_SCAN(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    /* We assume for now that files are never stored sorted */
    if (needed_pv->sort_arity != 0)
        ret_value->apply_truth_value = FALSE;
    else
    {
        ret_value->apply_truth_value = TRUE;
	ret_value->algorithm_argument.file_scan = oa->get_set;
        ret_value->process_input_count = 1;
    } /* if */

    return (OKAY);
}

/****************************************************************************
* do_any_good_BTREE_SELECT
*
* This routine determines whether the given btree scan will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_BTREE_SELECT(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_BTREE_SELECT(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    int         i, j;
    JOIN_ARG    *ja;

    ret_value->apply_truth_value = FALSE;
    if (needed_pv->sort_arity == 0) {
       ret_value->process_input_count = 1;
       ret_value->apply_truth_value = TRUE;
    }
    else if(needed_pv->sort_arity==1 ) {
       ja = Find_Join_Arg(oa->select.cond[0]->left);
       if (strcmp(needed_pv->sort_attr[0],ja->data.attr)==0) {
          ret_value->process_input_count = 1;
          ret_value->apply_truth_value = TRUE;
       }
    }

    return (OKAY);
}

/****************************************************************************
* do_any_good_SORT
*
* This routine determines whether the given sort will satisfy the
* physical property requirements.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS do_any_good_SORT(
OPERATOR          *current_operator,
OPERATOR_ARGUMENT *oa,
PROPERTY_VECTOR   *needed_pv,
SYSTEM_PROPERTY   *sys_prop,
LOGICAL_PROPERTY  *log_prop,
PROPERTY_VECTOR   *excl_pv,
OPERATOR          input_operator[],
ALGO_APPL_RETURN  *ret_value)
#else
STATUS do_any_good_SORT(current_operator, oa, needed_pv, sys_prop,
        log_prop, excl_pv, input_operator, ret_value)
OPERATOR          *current_operator;
OPERATOR_ARGUMENT *oa;
PROPERTY_VECTOR   *needed_pv;
SYSTEM_PROPERTY   *sys_prop;
LOGICAL_PROPERTY  *log_prop;
PROPERTY_VECTOR   *excl_pv;
OPERATOR          input_operator[];
ALGO_APPL_RETURN  *ret_value;
#endif
{
    int i;

    /* Make sure the needed_pv has at least 1 item */
    if (needed_pv->sort_arity == 0)
    {
        ret_value->apply_truth_value = FALSE;
    }
    else
    {
        /* A sort can always satisfy the needed property vectors */
        ret_value->apply_truth_value = TRUE;
        /* Set up the sort property vector */
        ret_value->algorithm_argument.sort.sort_arity = needed_pv->sort_arity;
        for (i = 0; i != needed_pv->sort_arity; i++)
            strcpy(ret_value->algorithm_argument.sort.sort_attr[i],
                    needed_pv->sort_attr[i]);
        ret_value->process_input_count = 1;
    } /* if */

    return (OKAY);
}

/****************************************************************************
* get_input_pv_MERGE_JOIN
*
* This routine requests property vectors for the input to a merge join
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_MERGE_JOIN(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *excl_pv,
int                input_id,
int                iteration,
PROPERTY_VECTOR    *input_pv[],
PROPERTY_VECTOR    *input_excl_pv[],
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PHY_EXPR_LIST_ELEM *input_ret_pele[],
COST               *input_ret_cost[],
BOOLEAN            *fail)
#else
STATUS get_input_pv_MERGE_JOIN(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    JOIN_ARG *ja;

    /* Initialize property vectors */
    init_property_vec(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    if (input_id == 0) {
        ja = Find_Join_Arg(aa->merge_join.left);
 
        input_pv[0]->sort_arity = 1;
        strcpy(input_pv[0]->sort_attr[0], ja->data.attr);
    }
    else if (input_id == 1) {
        ja = Find_Join_Arg(aa->merge_join.right);
 
        input_pv[1]->sort_arity = 1;
        strcpy(input_pv[1]->sort_attr[0], ja->data.attr);
    }
    else
        return (ERR_SANITY_CHECK);
    return (OKAY);
}

/****************************************************************************
* get_input_pv_FIX_ALG
*
* This routine requests property vectors for the input to a fix oper. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_FIX_ALG(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *excl_pv,
int                input_id,
int                iteration,
PROPERTY_VECTOR    *input_pv[],
PROPERTY_VECTOR    *input_excl_pv[],
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PHY_EXPR_LIST_ELEM *input_ret_pele[],
COST               *input_ret_cost[],
BOOLEAN            *fail)
#else
STATUS get_input_pv_FIX_ALG(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);


    return (OKAY);
}


/****************************************************************************
* get_input_pv_UNION_ALG
*
* This routine requests property vectors for the input to a union oper.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_UNION_ALG(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *excl_pv,
int                input_id,
int                iteration,
PROPERTY_VECTOR    *input_pv[],
PROPERTY_VECTOR    *input_excl_pv[],
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PHY_EXPR_LIST_ELEM *input_ret_pele[],
COST               *input_ret_cost[],
BOOLEAN            *fail)
#else
STATUS get_input_pv_UNION_ALG(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    return (OKAY);
}


/****************************************************************************
* get_input_pv_HASH_JOIN
*
* This routine requests property vectors for the input to a hash join
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_HASH_JOIN(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_HASH_JOIN(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    return (OKAY);
}

/****************************************************************************
* get_input_pv_NESTED_LOOP_JOIN
*
* This routine requests property vectors for the input to a hash join
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_NESTED_LOOP_JOIN(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_NESTED_LOOP_JOIN(aa, needed_pv, excl_pv, input_id, 
        iteration, input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    return (OKAY);
}

/****************************************************************************
* get_input_pv_UNNEST_ALG
*
* This routine requests property vectors for the input to unnest_alg operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_UNNEST_ALG(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *excl_pv,
int                input_id,
int                iteration,
PROPERTY_VECTOR    *input_pv[],
PROPERTY_VECTOR    *input_excl_pv[],
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PHY_EXPR_LIST_ELEM *input_ret_pele[],
COST               *input_ret_cost[],
BOOLEAN            *fail)
#else
STATUS get_input_pv_UNNEST_ALG(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    return (OKAY);
}

/****************************************************************************
* get_input_pv_PTR_HH_JOIN
*
* This routine requests property vectors for the input to a hash join
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_PTR_HH_JOIN(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_PTR_HH_JOIN(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);
 
    return (OKAY);
}

/****************************************************************************
* get_input_pv_PTR_SM_JOIN
*
* This routine requests property vectors for the input to a hash join
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_PTR_SM_JOIN(
ALGORITHM_ARGUMENT *aa,
PROPERTY_VECTOR    *needed_pv,
PROPERTY_VECTOR    *excl_pv,
int                input_id,
int                iteration,
PROPERTY_VECTOR    *input_pv[],
PROPERTY_VECTOR    *input_excl_pv[],
SYSTEM_PROPERTY    *input_sys_prop[],
LOGICAL_PROPERTY   *input_log_prop[],
PHY_EXPR_LIST_ELEM *input_ret_pele[],
COST               *input_ret_cost[],
BOOLEAN            *fail)
#else
STATUS get_input_pv_PTR_SM_JOIN(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    return (OKAY);
}


/****************************************************************************
* get_input_pv_TRAVERSE
*
* This routine requests property vectors for the input to a traverse 
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_TRAVERSE(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_TRAVERSE(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);
    copy_property_vec(input_pv[0], needed_pv);
    copy_property_vec(input_excl_pv[0], excl_pv);

    return (OKAY);
}

/****************************************************************************
* get_input_pv_FILTER
*
* This routine requests property vectors for the input to a filter
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_FILTER(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_FILTER(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);
    init_property_vec(input_excl_pv[input_id]);

    /* Filter doesn't care about input sort property vectors, but passes them */
    /* through from the output.  Excluded property vectors are likewise. */
    copy_property_vec(input_pv[0], needed_pv);
    copy_property_vec(input_excl_pv[0], excl_pv);

    return (OKAY);
}

/****************************************************************************
* get_input_pv_FILE_SCAN
*
* This routine requests property vectors for the input to a file scan
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_FILE_SCAN(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_FILE_SCAN(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* File scan has no inputs */

    return (OKAY);
}

/****************************************************************************
* get_input_pv_BTREE_SELECT
*
* This routine requests property vectors for the input to a btree scan
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_BTREE_SELECT(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_BTREE_SELECT(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    /* Btree scan has no inputs */

    return (OKAY);
}

/****************************************************************************
* get_input_pv_SORT
*
* This routine requests property vectors for the input to a sort
* operator.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS get_input_pv_SORT(
ALGORITHM_ARGUMENT *aa, 
PROPERTY_VECTOR    *needed_pv, 
PROPERTY_VECTOR    *excl_pv, 
int                input_id, 
int                iteration, 
PROPERTY_VECTOR    *input_pv[], 
PROPERTY_VECTOR    *input_excl_pv[], 
SYSTEM_PROPERTY    *input_sys_prop[], 
LOGICAL_PROPERTY   *input_log_prop[], 
PHY_EXPR_LIST_ELEM *input_ret_pele[], 
COST               *input_ret_cost[], 
BOOLEAN            *fail) 
#else 
STATUS get_input_pv_SORT(aa, needed_pv, excl_pv, input_id, iteration,
        input_pv, input_excl_pv, input_sys_prop, input_log_prop,
        input_ret_pele, input_ret_cost, fail)
ALGORITHM_ARGUMENT *aa;
PROPERTY_VECTOR    *needed_pv;
PROPERTY_VECTOR    *excl_pv;
int                input_id;
int                iteration;
PROPERTY_VECTOR    *input_pv[];
PROPERTY_VECTOR    *input_excl_pv[];
SYSTEM_PROPERTY    *input_sys_prop[];
LOGICAL_PROPERTY   *input_log_prop[];
PHY_EXPR_LIST_ELEM *input_ret_pele[];
COST               *input_ret_cost[];
BOOLEAN            *fail;
#endif
{
    int i;

    /* Initialize property vectors */
    property_vec_dont_care(input_pv[input_id]);

    input_excl_pv[0]->sort_arity = needed_pv->sort_arity;
    for (i = 0; i != needed_pv->sort_arity; i++)
        strcpy(input_excl_pv[0]->sort_attr[i], needed_pv->sort_attr[i]);

    return (OKAY);
}

/****************************************************************************
* long_cmp
*
* This routine compares two long values.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
int long_cmp( long *l1, long *l2)
#else
int long_cmp(l1, l2)
long *l1;
long *l2;
#endif
{
    if (*l1 > *l2) return 1;
    else if (*l1 < *l2) return -1;
    else return 0;
}

/****************************************************************************
* compare_cost
*
* This routine compares two costs, returning the cost comparison value.
* Special Note: If the cost2 value is actually the zero_cost value, then this
* comparison is being made to determine if cost1 is still valid.  In this case,
* if any of cost1's values are negative, return LESS_THAN.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
void compare_cost( COST *cost1, COST *cost2, COMPARE *result)
#else
void compare_cost(cost1, cost2, result)
COST    *cost1;
COST    *cost2;
COMPARE *result;
#endif
{
  if (cost1->cost.io < cost2->cost.io)
    *result = LESS_THAN;
  else if (cost1->cost.io == cost2->cost.io)
    *result = EQUAL_TO;
  else
    *result = GREATER_THAN;
}

/****************************************************************************
* add_cost
*
* This routine adds together two cost variables.  The result is in cost3.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
void add_cost( COST *cost1, COST *cost2, COST *cost3)
#else
void add_cost(cost1, cost2, cost3)
COST    *cost1;
COST    *cost2;
COST    *cost3;
#endif
{
    /* Set up the cost3 values */
    cost3->cost.io = cost2->cost.io + cost1->cost.io;
}

/****************************************************************************
* subtract_cost
*
* This routine subtracts two cost variables.  The result of cost1-cost2 is in
* cost3. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
void subtract_cost( COST    *cost1, COST    *cost2, COST    *cost3)
#else
void subtract_cost(cost1, cost2, cost3)
COST    *cost1;
COST    *cost2;
COST    *cost3;
#endif
{
    /* Set up the cost3 values */
    cost3->cost.io = cost1->cost.io - cost2->cost.io;
}

/****************************************************************************
* compare_property_vec
*
* This routine compares two property vectors, and determines if either
* covers the other.  The conditions for coverage are as follows: A covers B
* iff:
* 1) Neither A nor B is sorted, or A is sorted and B is not, or the major sort
*   order for A matches the full sort order for B, and
* 2) Neither A or B have distribution requirements, or A has distribution
*   requirements and B does not, or A and B have identical distribution
*   requirements, and
* 3) All nodes used by A are available in B.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS compare_property_vec(
PROPERTY_VECTOR  *pv1,
PROPERTY_VECTOR  *pv2,
SYSTEM_PROPERTY  *sp,
LOGICAL_PROPERTY *lp,
PV_COMPARE       *result)
#else
STATUS compare_property_vec(pv1, pv2, sp, lp, result)
PROPERTY_VECTOR  *pv1;
PROPERTY_VECTOR  *pv2;
SYSTEM_PROPERTY  *sp;
LOGICAL_PROPERTY *lp,
PV_COMPARE       *result;
#endif
{
    int i;
    int match;
    int pv1_cov_pv2 = FALSE;
    int pv2_cov_pv1 = FALSE;

    /* Perform a quick sanity check */
    if (pv1 == NULL || pv2 == NULL || lp == NULL || sp == NULL)
         return(ERR_NULL_POINTER);

    if ( pv1->sort_arity == -2) {
       if (pv1->sort_arity == -2) *result = PV_EQUAL;
       else  *result = PV2_COV_PV1;
       return(OKAY);
    }
    else if  ( pv2->sort_arity == -2) {
       *result = PV1_COV_PV2;
       return(OKAY);
    }
    /* 1) Check sort coverage. */
    /* See if the two vectors are identical */
    if (pv1->sort_arity == pv2->sort_arity)
    {
        match = TRUE;
        for (i = 0; i != pv1->sort_arity; i++)
            match = match &&
                    strcmp(pv1->sort_attr[i], pv2->sort_attr[i]) == 0;
        pv1_cov_pv2 = match; pv2_cov_pv1 = match;
    }
    /* See if pv1 covers pv2 */
    else if (pv1->sort_arity > pv2->sort_arity)
    {
        match = TRUE;
        for (i = 0; i != pv2->sort_arity; i++)
            match = match &&
                    strcmp(pv1->sort_attr[i], pv2->sort_attr[i]) == 0;
        pv1_cov_pv2 = match;
    }
    /* See if pv2 covers pv1 */
    else if (pv2->sort_arity > pv1->sort_arity)
    {
        match = TRUE;
        for (i = 0; i != pv1->sort_arity; i++)
            match = match &&
                    strcmp(pv1->sort_attr[i], pv2->sort_attr[i]) == 0;
        pv2_cov_pv1 = match;
    } /* if */

    /* Combine s_result, d_result, n_result. */
    if (pv1_cov_pv2 && pv2_cov_pv1)
        *result = PV_EQUAL;
    else if (pv1_cov_pv2)
        *result = PV1_COV_PV2;
    else if (pv2_cov_pv1)
        *result = PV2_COV_PV1;
    else
        *result = PV_NOT_COMPARABLE;

    return (OKAY);
}

/****************************************************************************
* compare_excl_property_vec
*
* This routine compares two property vectors, and determines if pv1 excludes
* pv2.  In other words, pv1 is the derived property vector for the plan, and
* pv2 is the properties to be excluded.  Conflict occurs whenever:
* 1) pv2 distribution is set and pv1 distribution matches pv2, or
* 2) pv2 sort is set and pv1 sort major portion is the same as pv2 sort.
* If conflict occurs, return PV_EXCL_FALSE, otherwise PV_EXCL_TRUE, in the
* result field.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS compare_excl_property_vec(
PROPERTY_VECTOR  *pv1,
PROPERTY_VECTOR  *pv2,
SYSTEM_PROPERTY  *sp,
LOGICAL_PROPERTY *lp,
PV_EXCL_COMPARE  *result)
#else
STATUS compare_excl_property_vec(pv1, pv2, sp, lp, result)
PROPERTY_VECTOR  *pv1;
PROPERTY_VECTOR  *pv2;
SYSTEM_PROPERTY  *sp;
LOGICAL_PROPERTY *lp;
PV_EXCL_COMPARE  *result;
#endif
{
    int i;

    /* Quick sanity check */
    if (pv1 == NULL || pv2 == NULL || lp == NULL)
         return (ERR_NULL_POINTER) ;


    /* 2) Check sort. */
    if (pv2->sort_arity != 0 && pv1->sort_arity >= pv2->sort_arity)
    {
        *result = PV_EXCL_FALSE;
        for (i = 0; i != pv2->sort_arity; i++)
            if (strcmp(pv1->sort_attr[i], pv2->sort_attr[i]) != 0)
                *result = PV_EXCL_TRUE;
        if (*result == PV_EXCL_FALSE)
            return (OKAY);
    } /* if */

    /* No conflicts between excluded property vectors. */
    *result = PV_EXCL_TRUE;

    return (OKAY);
}


/****************************************************************************
*
* Compare_Join_Args
*
* This routine compares two join arguments, testing for equality.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
int  Compare_Join_Args( JOIN_ARG  *p, JOIN_ARG *q)
#else
int  Compare_Join_Args( p, q )
JOIN_ARG  *p,*q;
#endif
{
    if (p==NULL && q==NULL) return (TRUE);
    else if (p==NULL && q!=NULL) return (FALSE);
    else if (p!=NULL && q==NULL) return (FALSE); 
    else {
      if (p->type == q->type ) {
         switch(p->type) {
         case CONST :
             if (p->data.cnst.kind == REAL) {
                if (p->data.cnst.value.real_value !=
                         q->data.cnst.value.real_value) return(FALSE);
             }
             else if (p->data.cnst.kind == INTEGER) {
                if (p->data.cnst.value.int_value != 
                         q->data.cnst.value.int_value) return(FALSE); 
             }	
             else  if (p->data.cnst.kind == TEXT){
                if(strcmp(p->data.cnst.value.text_value,
                         p->data.cnst.value.text_value) != 0) return(FALSE);
             }
             else return(FALSE);
             break;
          case ATTRIB : 
             if (strcmp(p->data.attr, q->data.attr) != 0 ) return (FALSE);
             break; 
          case OPER :
             if ( p->data.op != q->data.op ) return(FALSE);
             break;
          default :
             return (FALSE);
          }              
          if ( Compare_Join_Args(p->left,q->left) == FALSE) return (FALSE);
          if ( Compare_Join_Args(p->right,q->right) == FALSE) return (FALSE);
          return (TRUE);
      }
      else return(FALSE);
    }
}
          
          
 
/****************************************************************************
* compare_operator_args
*
* This routine compares two operator arguments, testing for equality. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS compare_operator_args(
OPERATOR_ARGUMENT *oa1,
OPERATOR          *op1,
OPERATOR_ARGUMENT *oa2,
OPERATOR          *op2,
COMPARE           *comp)
#else
STATUS compare_operator_args(oa1, op1, oa2, op2, comp)
OPERATOR_ARGUMENT *oa1;
OPERATOR          *op1;
OPERATOR_ARGUMENT *oa2;
OPERATOR          *op2;
COMPARE           *comp;
#endif
{
    int i;
    int match;

    /* Sanity checks */
    if (oa1 == NULL || oa2 == NULL || op1 == NULL || op2 == NULL)
        return (ERR_NULL_POINTER);

    if (*op1 != *op2)
        return (ERR_SANITY_CHECK);


    /* Default to not-equal, reset later if equality is found */
    *comp = NOT_EQUAL_TO ;

    /* Base the comparison on the operator type */
    switch (*op1)
    {
    case JOIN:
        if ( Compare_Join_Args(&(oa1->join),&(oa2->join)) )
            *comp = EQUAL_TO;
        break;
    case MAT:
        match = FALSE;
        if (oa1->mat.num==oa2->mat.num ) {
           match = TRUE;
           for (i=0; i< oa1->mat.num; i++) {
               if (strcmp(oa1->mat.attr[i],oa2->mat.attr[i])  ||
                   strcmp(oa1->mat.input[i].name,oa2->mat.input[i].name)) 
                     match = FALSE;
           }  
        }
        if (match)  *comp = EQUAL_TO;
        break;
    case FIX:
        if (strcmp(oa1->fix.attr, oa2->fix.attr) == 0 &&
              strcmp(oa1->fix.input.name, oa2->fix.input.name) == 0) 
            *comp = EQUAL_TO;
        break;
    case UNION :
        *comp = EQUAL_TO;
        break; 
    case SELECT:
        if (oa1->select.terms == oa2->select.terms)
        {
            match = TRUE;
            for (i = 0; i != oa1->select.terms; i++)
            {
                if ((!oa2->select.cond[i]->left) && (!oa1->select.cond[i]->left)) {
                   match = TRUE;
                   continue;
                }
                if (!oa2->select.cond[i]->left) {
                   match = FALSE;
                   break;
                }
                if (!oa1->select.cond[i]->left) {
                   match = FALSE;
                   break;
                }
                if (!Compare_Join_Args(oa1->select.cond[i]->left,oa2->select.cond[i]->left) || 
                        oa1->select.cond[i]->data.op != oa2->select.cond[i]->data.op || 
                        oa1->select.cond[i]->right->data.cnst.kind != oa2->select.cond[i]->right->data.cnst.kind ) 
                    match = FALSE;
                else if (oa1->select.cond[i]->right->data.cnst.kind == REAL)
                    match = oa1->select.cond[i]->right->data.cnst.value.real_value ==
                            oa2->select.cond[i]->right->data.cnst.value.real_value;
                else if (oa1->select.cond[i]->right->data.cnst.kind == INTEGER)
                    match = oa1->select.cond[i]->right->data.cnst.value.int_value ==
                            oa2->select.cond[i]->right->data.cnst.value.int_value;
                else if (oa1->select.cond[i]->right->data.cnst.kind == METHOD)
                    match = TRUE;
                else 
                    match = strcmp(oa1->select.cond[i]->right->data.cnst.value.text_value,
                            oa2->select.cond[i]->right->data.cnst.value.text_value) == 0;
                if (! match)
                    break;
            } /* for */
            if (match)
                *comp = EQUAL_TO;
        } /* if */
        break;
    case GET_SET:
        /* Use only the set name */
        if ( oa1->get_set.use_self == oa2->get_set.use_self &&
             strcmp(oa1->get_set.input.name, oa2->get_set.input.name) == 0 &&
             strcmp(oa1->get_set.attr, oa2->get_set.attr) == 0)
            *comp = EQUAL_TO;
        break;
    default:
            return (ERR_UNKNOWN_OPERATOR) ;
    } /* switch */

    return (OKAY);
}

/****************************************************************************
* print_attribute
*
* This routine prints out a field attribute record. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
void print_attribute( ATTRIBUTE *attr)
#else
void print_attribute(attr)
ATTRIBUTE *attr;
#endif
{
    printi("-%s: ", attr->name);
    switch(attr->attr_type)
    {
    case REAL:
	printf("real %f to %f, ", attr->domain_min.real_value,
		attr->domain_max.real_value);
	printf("width=%d", attr->attr_width);
	if (attr->btree_avail)
	    printf("  btree exists");
        printf(" unique=%d",attr->unique);
	break;
    case INTEGER:
	printf("int. %d to %d ", attr->domain_min.int_value,
		attr->domain_max.int_value);
	printf("width=%d use=%d", attr->attr_width,attr->use);
	if (attr->btree_avail)
	    printf(" btree exists");
        printf("  unique=%d",attr->unique);
	break;
    case TEXT:
	printf("text %s to %s ", attr->domain_min.text_value,
		attr->domain_max.text_value);
	printf("width=%d", attr->attr_width);
	if (attr->btree_avail)
	    printf(" btree exists");
        printf("  unique=%d",attr->unique);
	break;
    case REF:
        printf("ref.");
        break;
    case SET:
        printf("set.");
        break;
    case LIST:
        printf("list.");
        break;
    case METHOD:
        printf("method");
        break;
    } /* switch */
    printf(" off:%d\n",attr->offset);
}

/****************************************************************************
* print_relation
*
* This routine prints out a field relation value. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
void print_relation( RELATION rel)
#else
void print_relation(rel)
RELATION rel;
#endif
{
    switch (rel)
    {
    case EQUAL_OP: printf("="); break;
    case LESS_THAN_OP: printf("<"); break;
    case GREATER_THAN_OP: printf(">"); break;
    case LESS_THAN_EQUAL_OP: printf("<="); break;
    case GREATER_THAN_EQUAL_OP: printf(">="); break;
    case NOT_EQUAL_OP: printf("!="); break;
    case NO_OP : printf(" no_operator "); break;
    case PLUS  : printi(" + "); break;
    case MINUS : printi(" - "); break;
    case DIV  : printi(" * "); break;
    case MUL  : printi(" / "); break;
    case AND_OP : printi (" AND "); break;
    } /* switch */
}

/****************************************************************************
* print_logical_prop
*
* This routine prints out a logical property record. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS print_logical_prop( LOGICAL_PROPERTY *lp)
#else
STATUS print_logical_prop(lp)
LOGICAL_PROPERTY *lp;
#endif
{
    int i;

    if (lp == NULL)
        return (ERR_NULL_POINTER) ; 

    if (logical_prop_not_set (lp))
        return (ERR_NOT_INITIALIZED) ;

    printi("Relation %s\n", lp->name);
    printi("-arity = %d\n", lp->arity);
    printi("-width = %d\n",lp->record_width);
    printi("-cardinality = %d\n",lp->cardinality);
    for (i = 0; i != lp->arity; i++)
        print_attribute(lp->attr[i]);

    return (OKAY);
}

/****************************************************************************
* print_property_vec
*
* This routine prints out a property vector. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS print_property_vec( PROPERTY_VECTOR *pv)
#else
STATUS print_property_vec(pv)
PROPERTY_VECTOR *pv;
#endif
{
    int i ;

    /* Sanity checks */
    if (pv == NULL)
         return (ERR_NULL_POINTER);

    if (property_vec_not_set (pv))
        return (ERR_NOT_INITIALIZED);

    /* Perform the print */
    if (pv->sort_arity == 0)
    {
        printi("Unsorted\n");
    }
    else if (pv->sort_arity == -2)
    {
        printi("dont care\n");
    }
    else
    {
        printi("Arity = %d\n", pv->sort_arity);
        for (i = 0; i != pv->sort_arity; i++)
        {
            printi("-sort attribute %d: %s\n", i + 1, pv->sort_attr[i]);
        } /* for */
    } /* if */
    printf("\n");

    return (OKAY);
}

/****************************************************************************
* Print_Join_Arg 
*
* This routine prints out an join argument.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
void Print_Join_Arg( JOIN_ARG *p)
#else
void Print_Join_Arg(p)
JOIN_ARG *p;
#endif
{
   if (p==NULL) return;
   Print_Join_Arg(p->left);
   switch (p->type) {
      case OPER : print_relation(p->data.op);
                      break;
      case ATTRIB :   printi(" %s ",p->data.attr);
  		      break;
      case CONST  :
            switch (p->data.cnst.kind)
            {
            case REAL:
                printf(" %f", p->data.cnst.value.real_value);
                break;
            case INTEGER:
                printf(" %d", p->data.cnst.value.int_value);
                break;
            case TEXT:
                printf(" %s", p->data.cnst.value.text_value);
                break;
            }
            break;
   }
   Print_Join_Arg(p->right);
}

/****************************************************************************
* print_operator_arg
*
* This routine prints out an operator argument. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS print_operator_arg ( OPERATOR_ARGUMENT *oa, OPERATOR  *op)
#else
STATUS print_operator_arg (oa, op)
OPERATOR_ARGUMENT *oa;
OPERATOR          *op;
#endif
{
    int i;

    /* Sanity checks */
    if (oa == NULL || op == NULL)
        return (ERR_NULL_POINTER) ;

    /* Print based on operator type */
    switch (*op)
    {
    case JOIN:
        Print_Join_Arg(&(oa->join));
        printf(" join\n");
        break;
    case MAT:
        for (i=0;i<oa->mat.num;i++)
         printi("materialize %s on %s\n",oa->mat.attr[i],oa->mat.input[i].name);
        break;
    case SELECT:
        printi("");
        for (i = 0; i != oa->select.terms; i++)
        {
            /* printf("%s ", oa->select.cond[i]->data.attr);*/
            Print_Join_Arg(oa->select.cond[i]->left);
            print_relation(oa->select.cond[i]->data.op);
            Print_Join_Arg(oa->select.cond[i]->right);
            if (i != oa->select.terms - 1)
            {
                switch (oa->select.op[i])
                {
                case AND_OP: printf(" and "); break;
                case OR_OP: printf(" or "); break;
                } /* switch */
            } /* if */
        } /* for */
        printf("\n");
        break;
    case GET_SET:
        printi("read set %s", oa->get_set.input.name);
        printi("with name %s\n",oa->get_set.attr);
        break;
    case UNNEST:
	printi("attribute %s\n",oa->unnest.attr);
        break;
    default:
        return (ERR_UNKNOWN_OPERATOR);
    } /* switch */

    return (OKAY);
}

/****************************************************************************
* print_algorithm_arg
*
* This routine prints out an algorithm argument. 
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
STATUS print_algorithm_arg(
ALGORITHM_ARGUMENT *aa,
ALGORITHM          *al)
#else
STATUS print_algorithm_arg(aa, al)
ALGORITHM_ARGUMENT *aa;
ALGORITHM          *al;
#endif
{
    int i;

    /* Sanity checks */
    if (aa == NULL || al == NULL)
        return (ERR_NULL_POINTER) ;

    /* Switch based on algorithm type */
    switch (*al) 
    {
    case HASH_JOIN:
        Print_Join_Arg(&(aa->hash_join));
        printf("\n");
        break;
    case NESTED_LOOP_JOIN:
        Print_Join_Arg(&(aa->nested_loop_join));
        printf("\n");
        break;
    case PTR_SM_JOIN:
        printi("%s ", aa->ptr_hh_join.attr);
        printf("= self of %s ",aa->ptr_sm_join.input.name);
        printf("\n");
        break;
    case PTR_HH_JOIN:
        printi("%s ", aa->ptr_hh_join.attr);
        printf("= self of %s ",aa->ptr_hh_join.input.name);
        printf("\n");
        break;
    case MERGE_JOIN:
        Print_Join_Arg(&(aa->merge_join));
        printf("\n");
        break;
    case TRAVERSE:
        for (i=0;i<aa->traverse.num;i++)
           printi("%s on %s\n",aa->traverse.attr[i],aa->traverse.input[i].name);
        break;
    case FIX_ALG:
        printi(" on %s\n",aa->fix_alg.attr);
        break;
    case UNNEST_ALG:
	printi("on %s\n",aa->unnest_alg.attr);
        break;
    case UNION_ALG:
	printi("\n");
        break;
    case FILTER:
        printi("");
        for (i = 0; i != aa->filter.terms; i++)
        {
            Print_Join_Arg(aa->filter.cond[i]->left);
            print_relation(aa->filter.cond[i]->data.op);
            Print_Join_Arg(aa->filter.cond[i]->right);
            if (i != aa->filter.terms - 1)
            {
                switch (aa->filter.op[i])
                {
                case AND_OP: printf(" and "); break;
                case OR_OP: printf(" or "); break;
                } /* switch */
            } /* if */
        } /* for */
        printf("\n");
        break;
    case FILE_SCAN:
        printi("scan file %s\n", aa->file_scan.input.name);
        break;
    case BTREE_SELECT:
        printf(" btree select on set %s\n",aa->btree_select.input.name);
        Print_Join_Arg(aa->btree_select.cond->left);
        print_relation(aa->btree_select.cond->data.op);
        Print_Join_Arg(aa->btree_select.cond->right);
        printf("\n");
        break;
    case SORT:
        if (aa->sort.sort_arity != 0)
        {
            printi("sort on attribute%s",
                    (aa->sort.sort_arity == 1) ? " " : "s ");
            for (i = 0; i != aa->sort.sort_arity - 1; i++)
                printf("%s, ", aa->sort.sort_attr[i]);
            printf("%s.\n", aa->sort.sort_attr[i]);
        }
        else {
            printi("unsorted.\n");
        } /* if */
        break;
    default:
        printi("Invalid algorithm.\n");
        return (ERR_UNKNOWN_ALGORITHM);
    } /* switch */

    return (OKAY);
}

/* Support functions for the model.input file */

/****************************************************************************
* attr_in_set
*
* This function returns TRUE if the given attribute name is in the input set.
* It also works for sequences, and any other bulk type with the set attribute
* layout in its schema.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
int attr_in_set(
char        *attr,
BULK_SCHEMA *set)
#else
int attr_in_set(attr, set)
char        *attr;
BULK_SCHEMA *set;
#endif
{
    int i,k,j;
    char  tstr[MAX_ATTR_NAME_LEN];

    for (i = 0; i != set->arity; i++)
        if (strcmp(attr, set->attr[i]->name) == 0)
           return TRUE;
    return FALSE;
} /* attr_in_set */

/****************************************************************************
* select_in_set
*
* This function returns TRUE if the given select operator uses attributes from
* the given set.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
int select_in_set(
SELECT_ARG  *select,
BULK_SCHEMA *set)
#else
int select_in_set(select, set)
SELECT_ARG  *select;
BULK_SCHEMA *set;
#endif
{
    int       i;
    JOIN_ARG *ja;
    char left_or_right;

    for (i = 0; i != select->terms; i++) {
/*
        ja = Find_Join_Arg(select->cond[i]->left);
        if (ja == NULL) {
           printf("ja NULL\n");
           exit(8);
        }
        if (attr_in_set(ja->data.attr, set))
            return TRUE;
*/
        left_or_right = 0;
        if (select->cond[i]->left) {
           left_or_right = 1;
           ja = Find_Join_Arg(select->cond[i]->left);
           if (ja ) {
              if (! attr_in_set(ja->data.attr, set))
                 return FALSE;
              
           }
        }
        if (select->cond[i]->right) {
           left_or_right = 1;
           ja = Find_Join_Arg(select->cond[i]->right);
           if (ja ) {
              if (! attr_in_set(ja->data.attr, set))
                 return FALSE;
           }
        }
        if (!left_or_right) {
           ja = Find_Join_Arg(select->cond[i]);
           if (ja ) {
              if (! attr_in_set(ja->data.attr, set))
                 return FALSE;
           }
        }

    }
    return TRUE;
} /* select_in_set */
/****************************************************************************
* all_select_in_set
*
* This function returns TRUE if the given select operator uses attributes from
* the given set.
****************************************************************************/
#if defined(__cplusplus) || defined(ANSI)
int all_select_in_set(
SELECT_ARG  *select,
BULK_SCHEMA *set)
#else
int all_select_in_set(select, set)
SELECT_ARG  *select;
BULK_SCHEMA *set;
#endif
{
    int i;
    JOIN_ARG *ja;
    char left_or_right;

    for (i = 0; i != select->terms; i++) {
  //      ja = Find_Join_Arg(select->cond[i]->left);
        left_or_right = 0;
        if (select->cond[i]->left) {
           left_or_right = 1;
           ja = Find_Join_Arg(select->cond[i]->left);
           if (ja ) {
              if (! attr_in_set(ja->data.attr, set))
                 return FALSE;
           }
        }
        if (select->cond[i]->right) {
           left_or_right = 1;
           ja = Find_Join_Arg(select->cond[i]->right);
           if (ja ) {
              if (! attr_in_set(ja->data.attr, set))
                 return FALSE;
           }
        }
        if (!left_or_right) {
           ja = Find_Join_Arg(select->cond[i]);
           if (ja ) {
              if (! attr_in_set(ja->data.attr, set))
                 return FALSE;
           }
        }
       
    }

    return TRUE;
} /* select_in_set */


