#include "opt_defs.h"
#include "dbi_gen.h"
#include "dbi.h"
#include "opt.h"
#include "global.h"
#include <math.h>
#include <string.h>
#include "vol_util.h"

 extern char *error_text (int ) ;
 extern STATUS print_plan (PLAN *) ;
 extern STATUS volcano_optimizer (QUERY*, PROPERTY_VECTOR*,
                         PV_COMPARE, PROPERTY_VECTOR*,
                         PV_EXCL_COMPARE, COST*, PLAN**);

 extern QUERY     *get_qry( FILE   *fp );
 extern void       prnt_qry( QUERY * node, FILE   *fp );
 extern void       PutUse(QUERY *q, SEARCH *s);
 extern void       prnt_qry ( QUERY * node, FILE   *fp );
 extern QUERY     *get_qry (FILE   *fp );
 extern void       prnt_pln ( PLAN *node, FILE   *fp );
 extern PLAN      *get_pln ( FILE   *fp );

extern COST		infinite_cost ;

extern int              indent_value ;

SEARCH *s;
SEARCH s1;
QUERY  *q;
int MEM;


char *myAlloc( int size )
{
        char *buf;
        buf = malloc( size );
        return buf;
}

myFree( char *buf )
{
        if( !buf ) return 1;
        free( buf );
}

void printJoinArg( JOIN_ARG* pred )
{
	if( !pred ) return;
	printJoinArg( (JOIN_ARG *) pred->left );
	switch( pred->type )
	{
		case OPER :
			switch( pred->data.op )
			{
				case EQUAL_OP : printf(" = "); break;
				case LESS_THAN_OP : printf(" < "); break;
				case GREATER_THAN_OP : printf(" > "); break;
				case LESS_THAN_EQUAL_OP : printf(" <= "); break;
				case GREATER_THAN_EQUAL_OP : printf(" >= "); break;
				case NOT_EQUAL_OP : printf(" <> "); break;
				case AND_OP : printf(" AND ");
				default : printf(" UNDEF OP ");
			}
			break;
		case ATTRIB :
			printf(" %s ", pred->data.attr );
			break;
			
		case CONST :
			switch( pred->data.cnst.kind )
			{
				case REAL :
				   printf(" %f ",pred->data.cnst.value.real_value );
				   break;
				case TEXT :
				    printf(" %s ",pred->data.cnst.value.text_value );
				    break;
				case INTEGER :
				   printf(" %d ",pred->data.cnst.value.int_value );
				   break;
				default :
					printf(" UNDEF CONS ");
			}	
		default :
			printf(" UNDEF TYPE  : %d ", pred->type );
	}
	printJoinArg( (JOIN_ARG *) pred->right );
	
}

void RELATION_SCHEMA::copy( RELATION_SCHEMA *copyFrom )
{
	int i = 0;
	*this = *copyFrom;
	for(i=0; i<copyFrom->arity; i++) 
	{
		attr[i] = (ATTRIBUTE * ) myAlloc( sizeof(ATTRIBUTE));
		*attr[i] = *(copyFrom->attr[i]);
	}

}

ATTRIBUTE* RELATION_SCHEMA::contains( char *attName )
{
	int i;
	for(i=0; i<arity; i++ )
		if( !strcmp(attName, attr[i]->name) )
			return attr[i];	
	return NULL;
}

void RELATION_SCHEMA::printOn()
{
	int i;
	printf(" Name : %s \n",  name );
	//printf(" object_type : %d \n", object_type );
	printf(" cardinality : %d \n", cardinality );
	printf(" size        : %d \n", record_width );
	//displayFID( &fid );
	printf(" ATTRIBUTES :\n");
	for(i=0; i<arity; i++ )
	{
		printf(" Att Name    : %s\n", attr[i]->name );
		printf(" Cls Name    : %s\n", attr[i]->class_name );
		printf(" type        : %d\n", attr[i]->attr_type );
	} 
}


void printPlan( QUERY *aplan )
{
	if(!aplan) return;
	switch(aplan->my_operator )
	{
		case JOIN :
			printf(" operator : JOIN \n");
			printJoinArg( &aplan->argument.join);
			printPlan( (QUERY *) aplan->input[0] );
			printPlan( (QUERY *) aplan->input[1] );
			break;
		case SELECT :
			printf(" operator : SELECT \n");
			printJoinArg( aplan->argument.select.cond[0] );
			printf("\n");
			printPlan( (QUERY *) aplan->input[0] );
			break;
		case GET_SET :
			printf(" operator : GET_SET \n");
			printf(" attribute : %s \n", aplan->argument.get_set.attr );
			if( aplan->argument.get_set.use_self )
				printf(" USE SELF : TRUE \n");
			else
				printf(" USE SELF : FALSE \n");
			aplan->argument.get_set.input.printOn();
			printPlan( (QUERY *) aplan->input[0] );
			break;
		case MAT :
			printf(" operator : MAT \n");
			printf(" attribute : %s\n ", aplan->argument.mat.attr );
			aplan->argument.mat.input[0].printOn();
			printPlan( (QUERY *) aplan->input[0] );
			break;
		case UNNEST :
			printf(" operator : UNNEST \n");
			printPlan( (QUERY *) aplan->input[0] );
			break;
		default :
			printf(" Unknown Operator \n ");
	}
} 
	
		

main (int argc, char** argv)
{
    int error, i, jn, sel, mat, rel , it_num;
    PROPERTY_VECTOR pv, excl_pv ;
    PLAN  *plan ;
    COST   cost_limit ;
    QUERY *qry;
    FILE  *fp,*fp1;
    char  FPath[50];
    char  OutPath[100];


    sprintf(OutPath,"%s/tmp",getenv("MOOD_HOME"));

    
    it_num =1; 
    MEM = 1000;
    copy_cost(&cost_limit ,&infinite_cost); 

    sprintf(FPath,"%s/queryFile",OutPath);
    fp = fopen(FPath,"r");
    qry=get_qry(fp);
    fclose(fp); 

    printPlan( qry );
    
    init_property_vec (&pv) ;
    init_property_vec (&excl_pv) ;

    sprintf(FPath,"%s/planFile",OutPath);
    fp1= fopen(FPath,"w");
    error = volcano_optimizer ( qry,            
                                &pv,           
                                PV_EQUAL,     
                                &excl_pv,    
                                PV_EXCL_TRUE,
                                &cost_limit,
                                &plan) ;   

    if (error != OKAY) {
        printf (" error = %s\n",error_text(error)) ;
        fclose(fp1);
        unlink(FPath);
    }
    else
    {
        print_plan (plan) ;
        prnt_pln(plan,fp1);
        fclose(fp1);
    }

    exit (0) ;
}
