#include "mood_parser.h"
#include "keywords.h"
#include "mood_catalog.h"

putBlanks( int indent ) 
{
	int i;
	for( i=0; i<=indent; i++ ) printf(" ");
}

void printAttribute( char *attName, int MaxLength )
{
	int strLength;
	printf("%s", attName );
	strLength = strlen( attName );
	putBlanks( MaxLength - strLength  + 1 );
}

int findMaxLength( AttrList *aList )
{
	AttrList	*p;

	int maxLength=-1, currLength;

	for( p = aList; p; p = p->next )
	{
		currLength = strlen( p->entry.AttributeName );
		if( maxLength < currLength ) maxLength = currLength;
	}
	
	return maxLength;
}



printDescribe( AttrList *aList, int indent )
{
	int		i, inter_indent = indent, maxLength;
	AttrList	*p;
	

        if( !aList ) return TRUE;

	maxLength = findMaxLength( aList );	
	putBlanks( indent );

	switch( aList->entry.AttributeConstructor )
	{
		case 'T' :
			if( !strcmp( aList->entry.AttributeName, " "))
				inter_indent =  indent + 3;
			else {
				printAttribute( aList->entry.AttributeName, maxLength );
				//printf("%20s", aList->entry.AttributeName ); 
				inter_indent = indent + maxLength + 3;
			}
			printf("TUPLE (\n"); 
			printDescribe( aList->nextsub, inter_indent);
			putBlanks( indent );
			printf(" )\n");
			break;

		case 'S' :
			if( aList->nextsub->entry.AttributeConstructor = 'N') {
				//printf("%20s", aList->entry.AttributeName );
				printAttribute( aList->entry.AttributeName, maxLength );
				printf(" SET( %s ),\n" ,
					aList->nextsub->entry.AttrTypeName );
			} else {
				//printf("%20s", aList->entry.AttributeName );
				printAttribute( aList->entry.AttributeName, maxLength );
				printf(" SET( \n");
				inter_indent = indent + maxLength + 3;
				printDescribe( aList->nextsub, inter_indent );
				putBlanks( indent );
				printf(" ),");
			}
			break;
		
		case 'L' :
			if( aList->nextsub->entry.AttributeConstructor = 'N') {
				//printf("%20s", aList->entry.AttributeName );
				printAttribute( aList->entry.AttributeName, maxLength );
				printf(" LIST( %s ),\n" ,
					aList->nextsub->entry.AttrTypeName );
			} else {
				//printf("%20s", aList->entry.AttributeName );
				printAttribute( aList->entry.AttributeName, maxLength );
				printf(" LIST( \n");
				inter_indent = indent + maxLength + 3;
				printDescribe( aList->nextsub, inter_indent );
				putBlanks( indent );
				printf(" ),");
			}
			break;

		case 'R' :
			if( aList->nextsub->entry.AttributeConstructor = 'N') {
				//printf("%20s", aList->entry.AttributeName );
				printAttribute( aList->entry.AttributeName, maxLength );
				printf(" REF( %s ),\n" ,
					aList->nextsub->entry.AttrTypeName );
			} else {
				printf("%20s", aList->entry.AttributeName );
				printf(" REF( \n");
				inter_indent = indent + maxLength + 3;
				printDescribe( aList->nextsub, inter_indent );
				putBlanks( indent );
				printf(" ),");
			}
			break;

		case 'N' : 
			//printf("%20s ", aList->entry.AttributeName ); 
			printAttribute( aList->entry.AttributeName, maxLength );
			if( !strcmp( aList->entry.AttrTypeName, "STRING" ))
				printf( "%s[%d] ",aList->entry.AttrTypeName, aList->entry.AttributeLength);
			else 
				printf( "%s ",aList->entry.AttrTypeName);
			printf( ",\n");
	}

	printDescribe( aList->next, indent );

	return TRUE;
}

DescribeClause::parse()
{
	connectUserRequest();
	CHECK_TOKEN( DESCRIBE, 1, ""); // dummy check
	CHECK_TOKEN( IDENTIFIER, INVLD_CLSNAME_ERR, tokenValue() );
	setClassName( tokenValue() );
	return TRUE;
}

void printDLL( LinkedList *aList )
{
	ListItem	*p;
	if( !aList->firstItem ) return;
	for( p = aList->firstItem; p != aList->lastItem; p = p->next )
		printf("%s, ", (char *) p->item );
	printf("%s\n", (char *) p->item );
}


DescribeClause::execute()
{
	MoodsTypes	moodsType;
	Catalog		ct;
	AttrList	*al;
	CatBOOLEAN	status;
	ListItem	*p;

	superCls = new LinkedList();
	subCls = new LinkedList();
	att = NULL;


	if( !moodsType.getClassInfo(className,&ct,&att,superCls,subCls,&status ))
	{
		delete superCls;
		delete subCls;
		delete att;
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	if( status ){
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	if( interfaceMode == TEXTUAL_MODE ) { 

	switch( ct.ClassType )
	{
		case 't' :
			printf( " User Type %s\n", className );
			break;

		case 'T' :
			printf( " System Type %s\n", className );
			break;

		case 'c' :  
			printf( " User Class %s\n", className );
			if( superCls->firstItem ) {
				printf( " Inherits from ");
				printDLL( superCls );
			}
		
			if( subCls->firstItem ) {	
				printf( " Has subclasses ");
				printDLL( subCls );
			}

			break;

		case 'C' :  
			printf( " System Class %s\n", className );
			if( superCls->firstItem ) {
				printf( " Inherits from ");
				printDLL( superCls );
			}
			if( subCls->firstItem ) {	
				printf( " Has subclasses ");
				printDLL( subCls );
			}
			break;
	}


	for( al = att; al->next; al=al->next ); // find last tuple
	printDescribe( al, 0 );

	}

	return TRUE;	
}
	
