#include "mood_parser.h"
#include "mood_catalog.h"
#include "linklist.h"
#include "sign.h"


extern void lowerCase( char *, char *);


void readLine( char *line, FILE * source)
{
        int ch;
        ch = getc( source );
        while(ch != EOF )
        {
                *line = ch;
                line++;
                if( ch == '\n' ) break;
                ch = getc( source );
        }
        *line = '\0';
}

restoreHeader( char *className )
{
        char    command[180];
	char	backupName[180];
	char	fileName[180];

        sprintf( fileName, "%s/%s.h",  DatabaseHome, className );
	unlink( fileName );
	sprintf( backupName, "%s.old", fileName );
	rename( backupName, fileName );

	return TRUE;
}


deletePrototype( char *className, FunctionType ft )
{
        FILE    *source, *dest;
        char    fileName[120], backupName[120];
        char    command[180];
        char    line[180];

	strcat( ft.prototype, ";\n");
 
        sprintf( fileName, "%s/%s.h",  DatabaseHome, className );
	sprintf( backupName, "%s.old", fileName );
	rename( fileName, backupName );
        //sprintf( command, "cp %s %s.old", fileName, fileName );
 
        dest   = fopen( fileName,  "w" );
        source = fopen( backupName, "r");
 
        if( source == NULL || dest == NULL) return FALSE;
 
        while( !feof( source ) )
        {
                readLine( line, source );
                if( strcmp( line, ft.prototype ) ) fprintf( dest, "%s", line );
        }
        fclose( source );
        fclose( dest );
	sprintf( fileName, "%s/%s.c", DatabaseHome, ft.signature );
	unlink( fileName );
	sprintf( fileName, "%s/shobj/%s.so", DatabaseHome, ft.signature );
	unlink( fileName );
        return TRUE;
}

addPrototype( char *className, char *prototype )
{
        FILE    *source, *dest;
        char    fileName[120], backupName[120];
        char    command[180];
        char    line[180];
 
        sprintf( fileName, "%s/%s.h",  DatabaseHome, className );
	sprintf( backupName, "%s.old", fileName );
	rename( fileName, backupName );
        sprintf( command, "cp %s %s.old", fileName, fileName );
 
        dest   = fopen( fileName,  "w" );
        source = fopen( backupName, "r");
 
        if( source == NULL || dest == NULL) return FALSE;
 
        while( !feof( source ) )
        {
                readLine( line, source );
                if( !strcmp( line, "};\n" ) )
                        fprintf( dest, "%s;\n", prototype );
                fprintf( dest, "%s", line );
 
        }
        fclose( source );
        fclose( dest );
        return TRUE;
}

CreateFunction::parse()
{
	// parses CREATE FUNCTION FOR CLASS className
	//        WITH PROTOTYPE 'protoDec' USING 'fileName'
	connectUserRequest();
	CHECK_TOKEN( FUNCTION, FUNCTION_KWD_ERR, "");

	CHECK_TOKEN( FOR, FOR_KWD_ERR, "");
	CHECK_TOKEN( CLASS, CLASS_KYWD_ERR, "");
	CHECK_TOKEN( IDENTIFIER, INVLD_CLSNAME_ERR, "");
	strcpy( className, tokenValue());

	CHECK_TOKEN( WITH, WITH_KYWD_ERR, ""); 
	CHECK_TOKEN( PROTOTYPE, PROTOTYPE_KYWD_EXPECTED, "");
	CHECK_TOKEN( STRCONS, MISSING_PROTOTYPE, "" ); 
	strcpy( protoTypeStr, tokenValue() );

	CHECK_TOKEN( USING, USING_KYWD_EXPECTED ,"");
	CHECK_TOKEN( STRCONS, MISSING_FILE_NAME, "" ); 
	strcpy( fileName, tokenValue() );

	return TRUE;
}

readSignature( FunctionType *ft )
{
	char	fileName[180];
	FILE	*funcInfo;

	sprintf( fileName, "%s/tmp/fcninfo.dat", MoodHome );

	funcInfo = fopen( fileName, "rb");
 
	if( funcInfo == NULL ) return FALSE;

	fread( (char *) ft, sizeof( FunctionType), 1, funcInfo );

	fclose( funcInfo );

	unlink( fileName );

	return TRUE;
}

CreateFunction::execute()
{
	MoodsTypes	mt;
	FunctionType	ft;
	CatBOOLEAN	status;
	FindSignature	*fsign;
	char		command[4*MAXSTRLEN], *q;
	char		protoStr[4*MAXSTRLEN];
	FILE		*fp;
	

	// Check the existence of class
	if( !mt.findType( className, &status ) ) return FALSE;
	if( status ) {
		ur->setError( RUNTIME_ERROR, TYPE_NOT_EXIST, className );
		return FALSE;
	}

	// Check the existence of the filename
	fp = fopen( fileName, "r" );
	if( fp == NULL ) {
		ur->setError( RUNTIME_ERROR, INVALID_FILENAME, fileName );
		return FALSE;
	}
	else fclose( fp );

	if( fSpec == PUBLIC )
		sprintf( protoStr, "public : %s", protoTypeStr );
	else
		sprintf( protoStr, "public : %s", protoTypeStr );

	//modify the appropriate class header
	// Add prototype of function to "DatabaseHome/className.h"
	if( !addPrototype( className, protoStr ))
	{
		ur->setError( INTERNAL_ERROR, 0, "adding prototype");
		return FALSE;
	}


	// compile the C++ code with MOODCC
	if( !MoodBin )
	sprintf( command, 
        "CURR_MOOD_DB=%s; export CURR_MOOD_DB;MOODCC -c -pic -I%s %s", 
	DBrec.dbname, DatabaseHome, fileName);
	else
	sprintf( command, 
	"CURR_MOOD_DB=%s; export CURR_MOOD_DB;%s/MOODCC -c -pic -I%s %s", 
	DBrec.dbname, MoodBin, DatabaseHome, fileName);
	printf( "%s\n", command );
	system( command );

	if( !readSignature( &ft )){
		ur->setError( COMPILATION_ERROR, 0, "");
		restoreHeader( className );
		return FALSE;
	}

	//fsign = new FindSignature( ft.className, ft.funcName, ft.args );

	strcpy( ft.prototype, protoStr );

	if( !mt.addFunc( ft, &status )) {
		ur->userError = catalogError; 
		restoreHeader( className );
		return FALSE;
	}

        q = strrchr(fileName,'.');
        *q = 0;
	sprintf( command, "ld -o %s/shobj/%s.so %s.o -assert pure-text",DatabaseHome,ft.signature, fileName);
	
        printf("%s\n",command);
        system(command);
	// copy file as "signature".c
	sprintf( command, "cp %s.c %s/%s.c",fileName,DatabaseHome,ft.signature);
	printf( "%s\n", command );
        system(command);

	return TRUE;
}


DropFunction::parse()
{
	//DROP FUNCTION functionName 
	//[WITH PARAMETERS ParamTypes*|WITH PROTOTYPE DECLARATION 'protoDec']
	//FROM CLASS className;

	char lowerFuncName[ MAXSTRLEN ];

	CHECK_TOKEN( WITH, WITH_KYWD_ERR, "" );

	getToken();

	deletionType = tokenType();

	switch( tokenType() )
	{
		case SIGNATURE :
		case PROTOTYPE :
			CHECK_TOKEN( STRCONS, FUNCTION_NAME_EXP, "");
			break;

		case IDENTIFIER :

			lowerCase( tokenValue(), lowerFuncName );

			if( !strcmp( lowerFuncName, "name" )) {
				CHECK_TOKEN( IDENTIFIER, FUNCTION_NAME_EXP, tokenValue());
			}
			else {
				ur->setError( PARSE_ERROR, SIGN_PROTO_NAME_ERR, ""); 
				return FALSE;
			}

			break;

		default :
			ur->setError( PARSE_ERROR, SIGN_PROTO_NAME_ERR, "");
			return FALSE;
	}

	strcpy( delStr, tokenValue() ); 

	CHECK_TOKEN( FROM, FROM_KYWD_ERR, "");
	CHECK_TOKEN( CLASS, CLASS_KYWD_ERR, tokenValue());
	CHECK_TOKEN( IDENTIFIER, CLS_NAME_EXPECTED,"");
	strcpy( className, tokenValue() );

	return TRUE;
}

DropFunction::execute()
{
	MoodsTypes      moodsType;
	Catalog         ct;
	AttrList        *al;
	CatBOOLEAN      status, notFound;
	char 		funcSignature[2*MAXSTRLEN];
	char 		funcProto[2*MAXSTRLEN];
	FunctionType	ft;

	//check the existence of the class
	if( !moodsType.getClassInfo(className,&ct,&al, &notFound ))
	{
		delete al;
		ur->userError = catalogError;
		return FALSE;
	}

	delete al;

	if( notFound ) {
		ur->setError( PARSE_ERROR, INVLD_CLSNAME_ERR, className );
		return FALSE;
	}

	//construct signature of function
	switch( deletionType )
	{
		case IDENTIFIER :
			printf("name, is coming\n");
			break;
		case PROTOTYPE :
			sprintf( funcProto, "%s::%s", className, delStr );
			FindSignature	findSign( funcProto );
			strcpy( funcSignature, findSign.Signature );
			break;

		case SIGNATURE :
			strcpy( funcSignature, delStr );
			break;
	}

	//remove function prototype from appropriate header file
	if( !moodsType.deleteFunction( className, funcSignature, &status, &ft )){
		ur->userError = catalogError;
		return FALSE;
	}

	deletePrototype( className, ft );

	return TRUE;

}

ListFunction::parse()
{
	// LIST [ OWN | INHERITED | ALL] FUNCTION OF typeName/className;
	connectUserRequest();
	CHECK_TOKEN( MS_OF, OF_KYWD_ERR, "");
	CHECK_TOKEN( IDENTIFIER, CLS_NAME_EXPECTED, "" );
	strcpy( className, tokenValue() );
	return TRUE;
}

ListFunction::execute()
{
	MoodsTypes	mt;
	ListItem	*p;
	FunctionType	*ft;

	functionList = new LinkedList();

	if( !mt.listFunction( className, functionList ) ) {
		delete functionList;
		ur->userError = catalogError;
		return FALSE;
	}
	if( interfaceMode == TEXTUAL_MODE ) {
		printf("FUNCTIONS\n");	
		printf("---------\n");	
	
		for( p = functionList->firstItem; p; p=p->next ){
			ft = (FunctionType *) p->item;
			printf("%s\n", ft->signature );
		}
	}

	return TRUE;
}
