#include "index.h"
#include "keywords.h"
#include <mood_catalog.h>
#include <strings.h>

//extern ERTYPE ErrorCheck(int e,char *routine) ;

#define check( a ,b ) { if (*a) { printf("%s\n", b); return; }; } ;

void IndexClass::freeAttributes( AttrList *attributes )
{
    if( attributes == NULL ) return;
    freeAttributes( attributes->nextsub );
    freeAttributes( attributes->next );
    myFree( (char *) attributes );
}

void IndexClass::setAttrType( char *type, int *attrType )
{
    if( strcmp( type, "STRING" ) == 0 ) *attrType = SM_string; 

    if( strcmp( type, "INTEGER" ) == 0 ||
        strcmp( type, "BOOLEAN" ) == 0 ||
        strcmp( type, "CHAR" ) == 0 ) *attrType = SM_int;

    if( strcmp( type, "FLOAT" ) == 0 ) *attrType = SM_float; 
}

int IndexClass::getTypeLengthOffset( AttrList *attributes, char *attrName, int *attrType, int *attrLength, int * attrOffset )
{

    if( attributes == NULL ) return 0;

    if( strcmp( attributes->entry.AttributeName, attrName ) == 0 ){
        setAttrType( attributes->entry.AttrTypeName, attrType );
        //*attrLength = attributes->entry.AttributeLength;
        *attrOffset = attributes->entry.AttributeOffset;
        return 1;
    }
    
    if( !getTypeLengthOffset( attributes->nextsub, attrName, attrType, attrLength, attrOffset ) ) 
        if( !getTypeLengthOffset( attributes->next, attrName, attrType, attrLength, attrOffset ) ) return 0;
} 


void IndexClass::checkIndexName( FID handle, char *name, CatBOOLEAN *status )
{
   OID oid, oid1;
   CatBOOLEAN endMarker = FALSE, found = FALSE;
   OBJHDR objHdr;
   int err;
   USERDESC *userDesc;
   IndexEntry *indexEntry;

   *status = FALSE;
   err = sm_GetFirstOid( BufferGroup, &handle, &oid, &objHdr, &endMarker );
   ErrorCheck( err, "Cannot read the first OID, in create index" );

   while( ( endMarker == FALSE ) && ( found == FALSE ) ){
      err = sm_ReadObject( BufferGroup,&oid, 0, sizeof( IndexEntry ), &userDesc );
      ErrorCheck( err, "Cannot read the next index entry, in create index" );
      indexEntry = ( IndexEntry * )userDesc->basePtr;
      if( strcmp( indexEntry->indexName, name ) == 0 ) found = TRUE;
      sm_ReleaseObject( userDesc );
      sm_GetNextOid(BufferGroup, &oid, &oid1, &objHdr, &endMarker );
		memcpy( &oid, &oid1, sizeof( OID ) );
   }
   
   if( found ) *status = TRUE; 
}


OID IndexClass::insertIndexEntry( FID handle, char *indexName, char *className, char *attrName, int attrType, int attrLength, IID indexID )
{
    IndexEntry indexEntry;
    OID oid;
	 int i;  

    strcpy( indexEntry.indexName, indexName );
    strcpy( indexEntry.className, className );
    strcpy( indexEntry.attrName, attrName );
    indexEntry.attrType = attrType;
    //indexEntry.attrLength = attrLength;
    memcpy( &indexEntry.indexID, &indexID, sizeof( IID ) );

    sm_CreateObject(BufferGroup, &handle, NEAR_FIRST_PHYSICAL, NULL, NULL, sizeof( IndexEntry ), ( void * )&indexEntry, &oid );

    return oid; 
}


void IndexClass::createIndex( char *indexName, char *className, char *attrName, CatBOOLEAN unique, CatBOOLEAN indexType, CatBOOLEAN *status )
{
    IID indexID;
    OID element, oid, oid1;
    AttrList *attributes=NULL;
    int err, attrType, attrLength, attrOffset;
    FID handle;
    OBJHDR objHdr;
    USERDESC *userDesc;
    CatBOOLEAN endMarker=FALSE, success=FALSE;
    KEY key;


   MoodsTypes mt;
   Catalog cat;

   *status = FALSE;
   mt.getClassInfo( className, &cat, &attributes, status );
   check( status,"Error in retrieving class information." );

   *status = !getTypeLengthOffset( attributes, attrName, &attrType, &attrLength, &attrOffset );
   check( status, "Attribute name does not exist" );

   switch( attrType ){
      case SM_string: break;
      case SM_int   : break;
      case SM_float : break;
      default       : printf("Attribute is not of a basic type\n" );
                      *status = TRUE;
                      break;
   }

   /* get the file identifier of the class */
   mt.readType( "MOODS_INDEX", &cat, &success );
       
   checkIndexName( cat.Instance_File, indexName, status );
   check( status, "Index already exists." );

    /* create the index */
printf("************************************");
printf(" UNIQUE index creationinin check et ");
printf("************************************");
    if( unique ){
        if( indexType == BTREE )
           err = sm_CreateIndex( VolumeId, BufferGroup, SM_BTREENDX, attrType, attrLength, sizeof( OID ), TRUE, &indexID );
        else
           err = sm_CreateIndex( VolumeId, BufferGroup, SM_HASHNDX, attrType, attrLength, sizeof( OID ), TRUE, &indexID );
    } else {
        if( indexType == BTREE )
           err = sm_CreateIndex( VolumeId, BufferGroup, SM_BTREENDX, attrType, attrLength, sizeof( OID ), FALSE, &indexID );
        else
           err = sm_CreateIndex( VolumeId, BufferGroup, SM_HASHNDX, attrType, attrLength, sizeof( OID ), FALSE, &indexID );
    }

    ErrorCheck( err, "Cannot create index" );

    /* get the file identifier of the class */
   mt.readType( className, &cat, &success );
   memcpy( &handle, &cat.Instance_File, sizeof( FID ) );
 
   /* initialize the key structure */
   key.length = attrLength;
   key.valuePtr = ( void * )myAlloc( attrLength );

   /* insert the key values into the index */
   err = sm_GetFirstOid( BufferGroup, &handle, &oid, &objHdr, &endMarker );
   ErrorCheck( err, "Cannot read the first OID, in create index" );

   while( endMarker != TRUE ){
      err = sm_ReadObject( BufferGroup,&oid, 0, READ_ALL, &userDesc );
      ErrorCheck( err, "Cannot read the object , in create index" );

      memcpy(key.valuePtr, (void *) &userDesc->basePtr[attrOffset], attrLength);
      //memcpy((void *) &element, (void *) &userDesc->oid, sizeof(OID));
      memcpy((char *) &element, (char *) &oid, sizeof(OID));

      err = sm_ReleaseObject( userDesc );
      ErrorCheck( err, "Cannot read the object" );


      err = sm_InsertEntry(&indexID, BufferGroup, &key, &element);
      ErrorCheck( err, "Cannot insert the key into the index" );

      err = sm_GetNextOid(BufferGroup, &oid, &oid1, &objHdr, &endMarker );
      ErrorCheck( err, "Cannot read the next OID" );
      memcpy( &oid, &oid1, sizeof( OID ) );
   }

   myFree( key.valuePtr );


   /* insert the address (i.e. OID ) of the catalog entry into the list maintained in the catalog entry corresponding to the class */

   freeAttributes( attributes );

    mt.readType( "MOODS_INDEX", &cat, &success );
    oid = insertIndexEntry( cat.Instance_File, indexName, className, attrName, attrType, attrLength, indexID );

    *status = FALSE;
    mt.createIndex( className, oid, status ); 
    check( status, "Error in inserting the oid of the index into the list" );
}


void IndexClass::dropIndex( char *indexName, CatBOOLEAN *status )
{
   OID oid, oid1;
   CatBOOLEAN endMarker = FALSE, found = FALSE, success;
   OBJHDR objHdr;
   int err;
   USERDESC *userDesc;
   IndexEntry *indexEntry;
   MoodsTypes mt;
   Catalog cat;

   /* get the file identifier of the class */
   mt.readType( "MOODS_INDEX", &cat, &success );

   err = sm_GetFirstOid( BufferGroup, &cat.Instance_File, &oid, &objHdr, &endMarker );
   ErrorCheck( err, "Cannot read the first OID, in create index" );

   while( ( endMarker == FALSE ) && ( found == FALSE ) ){
      err = sm_ReadObject( BufferGroup,&oid, 0, sizeof( IndexEntry ), &userDesc );
      ErrorCheck( err, "Cannot read the next index entry, in create index" );
      indexEntry = ( IndexEntry * )userDesc->basePtr;
      if( strcmp( indexEntry->indexName, indexName ) == 0 ) found = TRUE;
      err = sm_ReleaseObject( userDesc );
		ErrorCheck( err, "Cannot release the object, inside IndexClass::dropIndex" );
		if( !found ){
			 err = sm_GetNextOid(BufferGroup, &oid, &oid1, &objHdr, &endMarker );
          ErrorCheck( err, "Cannot read the next oid,inside IndexClass::dropIndex" );
			 memcpy( &oid, &oid1, sizeof( OID ) );
      }
   }

   if( !found ){
      printf( "Index does not exist\n" );
      *status = TRUE;
      return;
   }

   err = sm_ReadObject( BufferGroup,&oid, 0, sizeof( IndexEntry ), &userDesc );
   ErrorCheck( err, "Cannot read the next index entry, in create index" );
   
   indexEntry = ( IndexEntry * )userDesc->basePtr;
   err = sm_DestroyIndex( ( struct IID *)&(indexEntry->indexID), BufferGroup );
   ErrorCheck( err, "Cannot destroy the index, inside create index" );
   *status = FALSE;
   mt.dropIndex( indexEntry->className, oid, status );
   //check( status, "Cannot remove oid of the index from its class"); 
   err = sm_ReleaseObject( userDesc );
   ErrorCheck( err, "Cannot release the object, inside IndexClass::dropIndex" );
   err = sm_DestroyObject( BufferGroup, &oid );
   ErrorCheck( err, "Cannot destroy the object, inside create index" );
}


IID IndexClass::getIID( char *className, char * attrName, CatBOOLEAN *status )
{
   OID oid, oid1;
   CatBOOLEAN endMarker = FALSE, found = FALSE, success;
   OBJHDR objHdr;
   int err;
   USERDESC *userDesc;
   IndexEntry *indexEntry;
   IID indexId;
   MoodsTypes mt;
   Catalog cat;

   /* get the file identifier of the class */
   mt.readType( "MOODS_INDEX", &cat, &success );


   *status = FALSE;
   err = sm_GetFirstOid( BufferGroup, &cat.Instance_File, &oid, &objHdr, &endMarker );
   ErrorCheck( err, "Cannot read the first OID, in create index" );

   while( ( endMarker == FALSE ) && ( found == FALSE ) ){
      err = sm_ReadObject( BufferGroup,&oid, 0, sizeof( IndexEntry ), &userDesc );
      ErrorCheck( err, "Cannot read the object, in create index" );
      indexEntry = ( IndexEntry * )userDesc->basePtr;
      if( strcmp( indexEntry->attrName, attrName ) == 0 &&
          strcmp( indexEntry->className, className ) == 0 ){
          found = TRUE;
          memcpy( &indexId, &indexEntry->indexID, sizeof( IID ) );
      }
      sm_ReleaseObject( userDesc );
      if( !found ){
			 sm_GetNextOid(BufferGroup, &oid, &oid1, &objHdr, &endMarker );
			 memcpy( &oid, &oid1, sizeof( OID ) );
      }
   }

   if( !found ){
      printf("Index not found.\n");
      *status = TRUE;
   }
   
   return indexId;

}

int IndexClass::createIndexParser()
{
MoodsTypes typeclass;
CatBOOLEAN status;

	connectUserRequestCreate();

	//IS_THIS_KEYWORD(CREATE);

//	CHECK_TOKEN(CLASS, CLASS_KYWD_ERR, "" ); 

/* ASSUMPTION: create index is parsed without the create keyword
	       which is checked beforehand

	       {CREATE} [UNIQUE] [HASH] INDEX indexname ON attributename
	       FOR classname TO indexname
*/

	getToken();
	if( tokenType() == MS_UNIQUE ){
	   indexEntry.uniqueFlag=1;
	   getToken();
	}

	if( tokenType() == HASH ){ 
	   indexEntry.indexType = MS_HASH;
	   getToken(); 
	}

	if( tokenType() != INDEX );
	else{
	   printf("Missing keyword INDEX\n");
	   return 0;
	}

	getToken();
	if( tokenType() != ON ){
	   printf("Missing keyword ON\n");
	   return 0;	   
	}

	/* ATTRIBUTE NAME ERRORU VERECEK SEKILDE DEGISTIR */
	CHECK_TOKEN(IDENTIFIER,CLS_NAME_EXPECTED, ""); 
	strcpy( indexEntry.attrName, tokenValue() );

	getToken();
	if( tokenType() != FOR ){
	   printf("Missing keyword FOR\n");
	   return 0;	   
	}

	CHECK_TOKEN(IDENTIFIER,CLS_NAME_EXPECTED, ""); 
	strcpy( indexEntry.className, tokenValue() );
    
	getToken();
	if( tokenType() != TO ){
	   printf("Missing keyword TO\n");
	   return 0;	   
	}

	/* INDEX NAME ERRORU VERECEK SEKILDE DEGISTIR */
	CHECK_TOKEN(IDENTIFIER,CLS_NAME_EXPECTED, ""); 
	strcpy( indexEntry.indexName, tokenValue() );

	getToken();
	if( tokenType() != SEMICOLON ){
	   printf("; expected\n");
	   return 0;
	}

	this->createIndex( indexEntry.indexName, indexEntry.className, 
		     indexEntry.attrName, indexEntry.uniqueFlag, 
		     indexEntry.indexType, &status );

	if( status == 1 ) return 0;

	return 1;
} 

int IndexClass::dropIndexParser()
{
/* Syntax: DROP INDEX indexname; */
CatBOOLEAN status;

	char indexName[MAXSTRLEN];

        connectUserRequestDrop();

	getToken();
	if( tokenType() != INDEX ){
	   printf("Missing keyword INDEX\n");
	   return 0;
	}	   

	/* INDEX NAME ERRORU VERECEK SEKILDE DEGISTIR */
	CHECK_TOKEN(IDENTIFIER,CLS_NAME_EXPECTED, ""); 
	strcpy( indexName, tokenValue() );

	dropIndex( indexName, &status );

	if( status == 1 ) return 0;

	return 1;
}
