#include "mood_parser.h"

void NewClause::printOn()
{
	ListItem *currItem;
	LinkedList *ll;

	printf("\n ------- NEW STATEMENT -------- \n");
	printf("Class name  : %s \n",className);
	currItem = newExpList->firstItem;
	while (currItem) {
		switch(currItem->itemType) {
			case NUL_PTR:	printf("EXP   : NULL value \n");
					break;
			case EXP_PTR:	printf("EXP   : ");
					currItem->printOn();
					break;
			case QRY_PTR:	printf("QUERY : ");
					currItem->printOn();
					break;
			case NEW_PTR:	printf("NEW   : ");
					currItem->printOn();
					break;
			case PNW_PTR:	printf("OBJECT : ");
					currItem->printOn();
					break;
			case SET_PTR:	printf("SET   : ");
					currItem->printOn();
					break;
			case LST_PTR:	printf("LIST  : ");
					currItem->printOn();
					break;
		} // switch
		currItem = currItem->next;
	} // while
	printf(" ----- END NEW STATEMENT ------ \n");
} // printOn


NewClause::parse()
{
	return newClause();
}


NewClause::newClause()
{
	Expression *exp;
	int	   Comma = FALSE;

	CHECK_TOKEN(NEW, NEW_KYWD_ERR, "");	// Check NEW keyword

	CHECK_TOKEN(IDENTIFIER, CLS_NAME_EXPECTED,"");		// Check class name

	strcpy(className,(char *)tokenValue());	// Take it into a variable

        CHECK_TOKEN(LPAR, MISSING_LP, "");	// error "missing ("

	do {
		getToken();

		switch(tokenType()) {
			case MS_SELECT: { // Parse the "select" statement...
				Query *query;

				backtrack();
				query = new Query(ur);
				newExpList->insert(QRY_PTR, VOID query);
				if (!query->parse()) {
				//	delete query;
					return FALSE;
				}
				break;
			}
			case NEW: { // Parse the "new" statement...
				NewClause *nestedNew;

				backtrack();
				nestedNew = new NewClause(ur);
				newExpList->insert(NEW_PTR, VOID nestedNew);
				if (!nestedNew->parse()) {
					//delete nestedNew;
					return FALSE;
				}
				break;
			}
			case LCB: { // Parse a set...
				LinkedList *SetExp;

				backtrack();
				SetExp = new LinkedList();
				newExpList->insert(SET_PTR, VOID SetExp);
				if (!parseSetClause(SetExp)) {
					//delete SetExp;
					return FALSE;
				}
				break;
			}
			case LB: { // Parse a list...
				LinkedList *ListExp;

				backtrack();
				ListExp = new LinkedList();
				newExpList->insert(LST_PTR, VOID ListExp);
				if (!parseListClause(ListExp)) {
					//delete ListExp;
					return FALSE;
				}
				break;
			}
			case NULLVALUE: { // Null statement...
				newExpList->insert(NUL_PTR, NULL);
				break;
			}
			case STAR: { // Parse the "new" statement (type)...
				NewClause *nestedNew;

				nestedNew = new NewClause(ur);
				newExpList->insert(PNW_PTR, VOID nestedNew);
				if (!nestedNew->parse()) {
					//delete nestedNew;
					return FALSE;
				}
				break;
			}
			case RPAR: { // Right paranthesis (constructor with no argument)

				if (Comma) {
					setError(PARSE_ERROR,UNKNOWN_SYNTAX,"");
					return FALSE;
				}

				backtrack();

				break;
			}
			default: { // Parse an expression...
				backtrack();
				exp = new Expression(ur);
				if (!exp->parse()) {
					//delete exp;
					return FALSE;
				}

				// Insert into the list
				newExpList->insert(EXP_PTR, VOID exp);
			} // default
		} // switch

		getToken();
		Comma = TRUE;

	} while (tokenType() == COMMA);

	backtrack();	
	CHECK_TOKEN(RPAR, MISSING_RP, "");	// error "missing )"

	return TRUE;

} // newClause


NewClause::parseSetClause(LinkedList *SetExp)
{
	Expression *exp;

        CHECK_TOKEN(LCB, LCB_EXPECTED, "");	// error "missing {"

	getToken();

	switch(tokenType()) {
		case MS_SELECT: { // Parse the "select" statement...
			Query *query;

			backtrack();
			query = new Query(ur);
			SetExp->insert(QRY_PTR, VOID query);
			if (!query->parse()) {
			//	delete query;
				return FALSE;
			}
			break;
		}
		case NEW: { // Parse the sequence of "new" statements...
			NewClause *nestedNew;

			backtrack();

			do {
				nestedNew = new NewClause(ur);
				SetExp->insert(NEW_PTR, VOID nestedNew);
				if (!nestedNew->parse()) {
					//delete nestedNew;
					return FALSE;
				} // if

				getToken();

			} while (tokenType() == COMMA);

			backtrack();

			break;
		}
		default: {
			// error "SELECT or NEW keyword expected" 
			ur->setError( PARSE_ERROR, SELNEW_KWD_EXPECTED, "" );
			return FALSE;
		} // default
	} // switch
	
	CHECK_TOKEN(RCB, RCB_EXPECTED, "");	// error "missing }"

	return TRUE;

} // parseSetClause


NewClause::parseListClause(LinkedList *ListExp)
{
	Expression *exp;
	NewClause *nestedNew;

        CHECK_TOKEN(LB, LB_EXPECTED, "");	// error "missing ["

	do {
		CHECK_TOKEN(NEW, NEW_KYWD_ERR, "");	// Check NEW keyword

		backtrack();
		nestedNew = new NewClause(ur);
		ListExp->insert(NEW_PTR, VOID nestedNew);
		if (!nestedNew->parse()) {
			//delete nestedNew;
			return FALSE;
		} // if

		getToken();

	} while (tokenType() == COMMA);

	backtrack();
	CHECK_TOKEN(RB, MISSING_RB, "");	// error "missing ]"

	return TRUE;

} // parseListClause



