#include <Cursor.h>
#include "mood_parser.h"
#include <mood_catalog.h>
#include <memory.h>
#include <dld.h>

extern evaluateExpr(JOIN_ARG *, CNST *);
extern int IncRefCount(OID);
extern int PrintRefCounts();

int NewClause::execute(OID *Oid) 
{
	ListItem 	*currItem, *currAItem;
	char 		*objBuff, *ConsObj, errorMsg[MAXSTRLEN];
	MoodsTypes 	catMngr;
	Catalog 	classInfo;
	AttrList	*attributes;
	AttrList	*currAttr;
	CatBOOLEAN	status;
	int		bufPntr,GrNdx,ParCount;
	// variables for constructor calling
	Function	*ConsFunc;
	Params		*Par;
	FunctionType	funcType;
	CatBOOLEAN	constructor=FALSE;
        void            *FreeList[15];
        int             ArgNum, i, Err;

	checkConstructor(&funcType,&ParCount,&constructor);

/*	if (!constructor) printf("constructor not found!\n");
		     else printf("constructor found!\n"); */

	Err = sm_OpenBufferGroup(4,BF_MRU,&GrNdx,NOFLAGS);
	if (Err != esmNOERROR) {
		ur->setError( EXODUS_ERROR, 1, "" );
		return FALSE;
	}

	// Get class information from the catalog... 
	catMngr.getClassInfo(className,&classInfo,&attributes,&status); 
	if (status != 0) {
		ur->userError = catalogError;
                ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	// Convert attribute tree to a linked list...
	convertToList(attributes);

#ifdef DEBUG
	currItem = attrList->firstItem;
	while (currItem) {
		currAttr  = (AttrList *) (currItem->item);

		printf(" \nFor the Attribute : \n");
		printf(" Name   : %s \n", currAttr->entry.AttributeName );
		printf(" Type   : %s \n", currAttr->entry.AttrTypeName );
		printf(" Offset : %d \n", currAttr->entry.AttributeOffset );
		printf(" Length : %d \n", currAttr->entry.AttributeLength );
		printf(" Cons.  : %c \n", currAttr->entry.AttributeConstructor );
		printf(" Member : %x \n", currAttr->entry.ClassMember );

		currItem = currItem->next;
	} // while
#endif

	// Allocate enough memory for the new object...
	objBuff = (char *)myAlloc(classInfo.Size);
	bufPntr = 0;
        ArgNum = 0;

	// first item will be the TypeId, place it to the buffer
	memcpy(objBuff+bufPntr,(char *)&classInfo.TypeId,sizeof(int));
	bufPntr += sizeof(int); 

	// Create a parameter list in case of constructor
	if (constructor) {
		Par = new Params(ParCount+1);
		ConsObj = (char *) myAlloc(classInfo.Size);
		Par->add_parameters(ConsObj);
	}

	// Process the newExpList until end of list... 

	currItem  = newExpList->firstItem;
	currAItem = attrList->firstItem->next;

	while ((currItem && currAItem) ||
	       (currItem && constructor)) {

		if (currAItem)
			currAttr  = (AttrList *) (currAItem->item);

		// Move the new information to the buffer...

		switch (currItem->itemType) {
		case NUL_PTR: { // "NULL value" (but may match if SET, REF, LIST are expected)

			if (!constructor) {
				OID Oid;

				if ((currAttr->entry.AttributeConstructor == 'R') ||
				    (currAttr->entry.AttributeConstructor == 'S') ||
				    (currAttr->entry.AttributeConstructor == 'L')) {
					INVALIDATE_OID(Oid);
					memcpy(objBuff+bufPntr,&Oid,sizeof(OID));
					bufPntr += sizeof(OID);
				}
				else {
					ur->setError( RUNTIME_ERROR, NULL_VL_NOT_SUPP, "" );
					return FALSE;
				}
			}
			else {
				// Needs further processing
				Par->add_parameters(NULL);
			}

			break;
		}
 		case EXP_PTR: { // "expression"
			Expression	*exp;
			JOIN_ARG	*jExp;
			CNST		*retVal;

			retVal = (CNST *) malloc(sizeof(CNST));
                        FreeList[ArgNum++] = (void *) retVal;

			exp = ((Expression *) currItem->item);

			jExp = exp->convertToIlker(exp->lastItem);

			evaluateExpr(jExp,retVal);

			// CHECK --> exp type == catalog type ...
			// & place return value to the buffer...
			switch(retVal->kind) {
			case INTEGER :
				if (!constructor) {
					if (strcmp(currAttr->entry.AttrTypeName,"INTEGER")) {
						sprintf( errorMsg, "%d (%s was expected)",
							ArgNum, currAttr->entry.AttrTypeName);
						ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
						return FALSE;
					}
					memcpy(objBuff+bufPntr,(char *)&retVal->value.int_value,
								sizeof(int));
					bufPntr += sizeof(int);
				}
				else { 
					Par->add_parameters((char *)&retVal->value.int_value);
				}

				break;
			case REAL    : 
				if (!constructor) {
					if (strcmp(currAttr->entry.AttrTypeName,"FLOAT")) {
						sprintf( errorMsg, "%d (%s was expected)",
							ArgNum, currAttr->entry.AttrTypeName);
						ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
						return FALSE;
					}
					memcpy(objBuff+bufPntr,(char *)&retVal->value.real_value,
								sizeof(float));
					bufPntr += sizeof(float);
				}
				else {
					Par->add_parameters((char *)&retVal->value.real_value);
				}

				break;
			case TEXT    : 
				if (!constructor) {
					if (strcmp(currAttr->entry.AttrTypeName,"STRING")) {
						sprintf( errorMsg, "%d (%s was expected)",
							ArgNum, currAttr->entry.AttrTypeName);
						ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
						return FALSE;
					}
					memcpy(objBuff+bufPntr,(char *)&retVal->value.text_value,
								currAttr->entry.AttributeLength);
					bufPntr += currAttr->entry.AttributeLength;
				}
				else {
					Par->add_parameters((char *)&retVal->value.text_value);
				}

				break;
			} // switch

			break; 
		}
		case QRY_PTR: { // "query (ref)"
			Query	*query;
			OID	*refOid;
			char	*QryClName;
			int	objCntr;
			MOODSCursor	ResCur;
			char		Flag;
			int		Invalidated;

			refOid = (OID *)malloc(sizeof(OID));
                        FreeList[ArgNum++] = (void *) refOid;

			if (!constructor && (currAttr->entry.AttributeConstructor != 'R')){
				sprintf( errorMsg, "%d (REF was not expected)", ArgNum);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

			query = (Query *) currItem->item;

                        QryClName=query->checkAltintas();
			if (!QryClName) {
				ur->setError( RUNTIME_ERROR, INVLD_QUERY, " in REF");
				return FALSE;
			}

			// CHECK return ObjType == catalog type 
			if (!constructor && strcmp((currAttr->nextsub)->entry.AttrTypeName,QryClName)){
				sprintf( errorMsg, "%d ( REF(%s) was expected)", ArgNum,
						(currAttr->nextsub)->entry.AttrTypeName);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

			if (!(query->generateAllTree())) return FALSE;
			if (!(query->generatePlan())) return FALSE;

			OpenCursor(query->solutionFile,&ResCur);

			if (!GetFirstObject(&ResCur,&Flag)) {
				ur->setError( RUNTIME_ERROR, CANNOT_GETOBJ, "" );
				return FALSE;
			}

			if (!Flag) {
				memcpy((char *)refOid,(char *)ResCur.CurUserdesc->basePtr,sizeof(OID));

				if (!GetNextObject(&ResCur,&Flag)) {
					ur->setError( RUNTIME_ERROR, CANNOT_GETOBJ, "" );
					return FALSE;
				}

				if (!Flag) {
					// at least two objects...
					ur->setError( RUNTIME_ERROR, TOO_MANY_OBJS, "" );
					return FALSE;
				}
				Invalidated = FALSE;
			} // if
			else {
				INVALIDATE_OID(*refOid);
				Invalidated = TRUE;
			} // else	

			CloseCursor(&ResCur);

			if (!Invalidated) {
				if (!IncRefCount(*refOid)) {
					ur->userError = catalogError;
                                        ur->operationType = HANDLE_ERROR;
					return FALSE;
				}
			}

			if (!constructor) {
				// place refOid to the buffer...
				memcpy(objBuff+bufPntr,(char *)refOid,sizeof(OID));
				bufPntr += sizeof(OID);
			}
			else {
				Par->add_parameters((char *)refOid);
			}

			break; 
		}
		case NEW_PTR: { // "ref"
			NewClause *nestedNew;
			OID	  *refOid;

			refOid = (OID *)malloc(sizeof(OID));
                        FreeList[ArgNum++] = (void *) refOid;

			if (!constructor && (currAttr->entry.AttributeConstructor != 'R')){
				sprintf( errorMsg, "%d (REF was not expected)", ArgNum);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

			nestedNew = (NewClause *) currItem->item;

			if (!constructor && strcmp((currAttr->nextsub)->entry.AttrTypeName,
						   nestedNew->className)) {
				sprintf( errorMsg, "%d ( REF(%s) was expected)", ArgNum,
						(currAttr->nextsub)->entry.AttrTypeName);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}


			if (!nestedNew->execute(refOid)) return FALSE;

			if (!IncRefCount(*refOid)) {
				ur->userError = catalogError;
                                ur->operationType = HANDLE_ERROR;
				return FALSE;
			}

			if (!constructor) {
				// place refOid to the buffer...
				memcpy(objBuff+bufPntr,(char *)refOid,sizeof(OID));
				bufPntr += sizeof(OID);
			}
			else {
				Par->add_parameters((char *)refOid);
			}

			break; 
		}
		case PNW_PTR: { // actual object (*new)
			NewClause *nestedNew;
			OID	  POid;
			char	  *NestedObj;	
			int	  ObjSize;
			USERDESC  *UserDesc;
			DeleteClause del(ur);

			if (!constructor && (currAttr->entry.AttributeConstructor != 'N')){
				sprintf( errorMsg, "%d (REF was not expected)", ArgNum);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

			nestedNew = (NewClause *) currItem->item;

			if (!constructor && strcmp(currAttr->entry.AttrTypeName,
							nestedNew->className)) {
				sprintf( errorMsg, "%d ( class %s was expected)", ArgNum,
						currAttr->entry.AttrTypeName);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

			if (!nestedNew->execute(&POid)) return FALSE;

			Err = sm_ReadObject(GrNdx,&POid,0,READ_ALL,&UserDesc);
			if (Err != esmNOERROR) {
				ur->setError( EXODUS_ERROR, 1, "" );
				return FALSE;
			}

			ObjSize = UserDesc->objectSize;

			NestedObj = (char *)malloc(ObjSize);
			memcpy(NestedObj,UserDesc->basePtr,ObjSize);

			Err = sm_ReleaseObject(UserDesc);
			if (Err != esmNOERROR) {
				ur->setError( EXODUS_ERROR, 1, "" );
				return FALSE;
			}

			if (!del.DeleteObject(&POid,nestedNew->className,TRUE))
				return FALSE;

			if (!constructor) {
				// place the object to the buffer...
				memcpy(objBuff+bufPntr,NestedObj,ObjSize);
				bufPntr += ObjSize;
			}
			else {
				Par->add_parameters((char *)NestedObj);
			}

			break;
		}
		case SET_PTR: { // "set"
			Set	   *set;
			LinkedList *setExps;
			ListItem   *setItem;
			OID	   *setOid;
			int	   typeId, pSize;

			setOid = (OID *)malloc(sizeof(OID));
                        FreeList[ArgNum++] = (void *) setOid;

			if (!constructor && (currAttr->entry.AttributeConstructor != 'S')) {
				sprintf( errorMsg, "%d (SET was not expected)", ArgNum);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

                        set = new Set(&classInfo.Instance_File,
				      NEAR_FIRST_PHYSICAL,
				      NULL,&status);
			if (status != 0) return FALSE;

                        *setOid = set->getSetOid(&status);
			if (status != 0) return FALSE;

			setExps = (LinkedList *) currItem->item;
			setItem = setExps->firstItem;

			if (setItem->itemType == NEW_PTR) {
				NewClause *newStmt;
				OID 	  newOid;

				while (setItem) {
					newStmt = (NewClause *) setItem->item;
					if (!constructor && 
						strcmp((currAttr->nextsub)->entry.AttrTypeName,newStmt->className)) {
						sprintf( errorMsg, "%d ( SET(%s) was expected)", 
							ArgNum, (currAttr->nextsub)->entry.AttrTypeName);
						ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
						return FALSE;
					}

					if (!newStmt->execute(&newOid)) return FALSE;

					// insert it to the set.
					set->insertToSet(&newOid,&status);
					if (status != 0) return FALSE;

					if (!IncRefCount(newOid)) {
						ur->userError = catalogError;
                                                ur->operationType = HANDLE_ERROR;
						return FALSE;
					}

					setItem = setItem->next;
				} // while
			} // if
			else { // it is a query (set).
				Query	*query;
				char	*QryClName;
				MOODSCursor	ResCur;
				char		Flag;
				OID		refOid;

				query = (Query *) setItem->item;

                                QryClName=query->checkAltintas();
				if (!QryClName) {
					ur->setError( RUNTIME_ERROR, INVLD_QUERY, " in SET");
					return FALSE;
				}

				// CHECK return ObjType == catalog type 
				if (!constructor && 
					strcmp((currAttr->nextsub)->entry.AttrTypeName,QryClName)) {
					sprintf( errorMsg, "%d ( SET(%s) was expected)", 
						ArgNum, (currAttr->nextsub)->entry.AttrTypeName);
					ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
					return FALSE;
				}

				if (!(query->generateAllTree())) return FALSE;
				if (!(query->generatePlan())) return FALSE;

				OpenCursor(query->solutionFile,&ResCur);

				if (!GetFirstObject(&ResCur,&Flag)) {
					ur->setError( RUNTIME_ERROR, CANNOT_GETOBJ, "" );
					return FALSE;
				}

				while (!Flag) {
					memcpy((char *)&refOid,
					       (char *)ResCur.CurUserdesc->basePtr,
					       sizeof(OID));

					// insert it to the set.
					set->insertToSet(&refOid,&status);
					if (status != 0) return FALSE;

					if (!IncRefCount(refOid)) {
						ur->userError = catalogError;
                                                ur->operationType = HANDLE_ERROR;
						return FALSE;
					}


					if (!GetNextObject(&ResCur,&Flag)) {
						ur->setError( RUNTIME_ERROR, CANNOT_GETOBJ, "" );
						return FALSE;
					}
				} // while
 
				CloseCursor(&ResCur);
			} // else

			if (!constructor) {
				// place the set OID to the buffer...
				memcpy(objBuff+bufPntr,(char *)setOid,sizeof(OID));
				bufPntr += sizeof(OID);
			}
			else {
				Par->add_parameters((char *)setOid);
			}

			break; 
		}
		case LST_PTR: { // "list"
			List		*list;
			LinkedList	*listExps;
			ListItem	*listItem;
			NewClause	*newStmt;
			OID		*listOid, newOid;
			int		typeId, pSize;

			listOid = (OID *)malloc(sizeof(OID));
                        FreeList[ArgNum++] = (void *) listOid;

			if (!constructor && (currAttr->entry.AttributeConstructor != 'L')) {
				sprintf( errorMsg, "%d (LIST was not expected)", ArgNum);
				ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
				return FALSE;
			}

			list = new List(&classInfo.Instance_File,
					NEAR_FIRST_PHYSICAL,
					NULL,20,&status);
			if (status != 0) return FALSE;

                        *listOid = list->getListOid(&status);
			if (status != 0) return FALSE;

			listExps = (LinkedList *) currItem->item;
			listItem = listExps->firstItem;

			while (listItem) {
				newStmt = (NewClause *) listItem->item;
				if (!constructor && 
					strcmp((currAttr->nextsub)->entry.AttrTypeName,newStmt->className)) {
					sprintf( errorMsg, "%d ( LIST(%s) was expected)", 
						ArgNum, (currAttr->nextsub)->entry.AttrTypeName);
					ur->setError( RUNTIME_ERROR, R_TYPE_MISMATCH, errorMsg );
					return FALSE;
				}

				if (!newStmt->execute(&newOid)) return FALSE;

				// insert it to the list.
				list->insertToList(&newOid,&status);
				if (status != 0) return FALSE;

				if (!IncRefCount(newOid)) {
					ur->userError = catalogError;
                                        ur->operationType = HANDLE_ERROR;
					return FALSE;
				}

				listItem = listItem->next;
			} // while

			if (!constructor) {
				// place the list OID to the buffer...
				memcpy(objBuff+bufPntr,(char *)listOid,sizeof(OID));
				bufPntr += sizeof(OID);
			}
			else {
				Par->add_parameters((char *)listOid);
			}

			break; 
		}
		} // switch

		currItem  = currItem->next;

		if (currAItem)
			currAItem = currAItem->next; 

	} // while

	// Create the new object through the storage manager...
	// get OID from exodus ( *Oid = ...).

	if (constructor) { // call appropriate constructor
		//printf("using constructor...\n");

		ConsFunc = new Function(funcType.signature);
		if (ConsFunc->error) {
			ur->setError( RUNTIME_ERROR, DLD_ERROR, "" );
			return FALSE;
		}

		(void *) ConsFunc->execute(Par);
		if (ConsFunc->error) {
			ur->setError( RUNTIME_ERROR, DLD_ERROR, "" );
			return FALSE;
		}

		// copy excluding TypeId field
		memcpy(objBuff+4,ConsObj+4,classInfo.Size-4); 
	}
	else {
		if (currAItem && !currItem) {
			ur->setError( RUNTIME_ERROR, MISSING_ARG, "" );
			return FALSE;
		}

		if (currItem && !currAItem) {
			ur->setError( RUNTIME_ERROR, EXTRA_ARG, "" );
			return FALSE;
		}
	}

	Err = sm_CreateObject(GrNdx,&classInfo.Instance_File,NEAR_LAST,
				NULL,NULL,classInfo.Size,(void *)objBuff,Oid);
	if (Err != esmNOERROR) {
		ur->setError( EXODUS_ERROR, 1, "" );
		return FALSE;
	}

	if (!IncRefCount(*Oid)) {
		ur->userError = catalogError;
                ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	//printf("Printing Ref Counts:\n");
	PrintRefCounts();
	
	Err = sm_CloseBufferGroup(GrNdx);
	if (Err != esmNOERROR) {
		ur->setError( EXODUS_ERROR, 1, "" );
		return FALSE;
	}

	// Free the allocated memory...
	myFree((char *) objBuff);
	myFree((char *) ConsObj);

	// deallocate param-list...
        for(i=0;i<ArgNum;++i)
           free(FreeList[i]);

	return TRUE; 

} // execute

int NewClause::checkConstructor(FunctionType *funcType, int *ndx, CatBOOLEAN *success)
{
	ListItem	*currItem;
	char		**arguments;
	int		i;
	int		appOffset;
	MoodsTypes	moodsType;

	arguments = (char **) myAlloc(newExpList->size*sizeof(char *));

	*ndx=0;
	currItem  = newExpList->firstItem;

	while (currItem) {

		switch (currItem->itemType) {
		case NUL_PTR: { // "NULL value"

			// Needs further processing

			arguments[*ndx] = (char *) myAlloc(5);
			strcpy(arguments[*ndx], "NULL");

			break;
		}
 		case EXP_PTR: { // "expression"
			Expression	*exp;
			char		expType[30];

			exp = ((Expression *) currItem->item);

			exp->checkSemantic( NULL, exp->lastItem, expType);

			arguments[*ndx] = (char *) myAlloc(strlen(expType)+1);
			strcpy(arguments[*ndx], expType);

			break;
		}
		case QRY_PTR: { // "query (ref)"
			Query	*query;
			char	*ClName;

			query = (Query *) currItem->item;

                        ClName=query->checkAltintas();
			if (!ClName) {
				ur->setError( RUNTIME_ERROR, INVLD_QUERY, " in REF");
				return FALSE;
			}

			arguments[*ndx] = (char *) 
				myAlloc(strlen(ClName)+5);
			strcpy(arguments[*ndx],"REF_");
			strcat(arguments[*ndx],ClName);

			break; 
		}
		case NEW_PTR: { // "ref"
			NewClause *nestedNew;

			nestedNew = (NewClause *) currItem->item;

			arguments[*ndx] = (char *) 
				myAlloc(strlen(nestedNew->className)+5);
			strcpy(arguments[*ndx],"REF_");
			strcat(arguments[*ndx],nestedNew->className);

			break; 
		}
		case PNW_PTR: { // "object"
			NewClause *nestedNew;

			nestedNew = (NewClause *) currItem->item;

			arguments[*ndx] = (char *) 
				myAlloc(strlen(nestedNew->className)+1);
			strcpy(arguments[*ndx],nestedNew->className);

			break; 

			break;
		}
		case SET_PTR: { // "set"
			LinkedList *setExps;
			ListItem   *setItem;
			//char	  *ClName;
			char	  ClName[MAXSTRLEN];

			setExps = (LinkedList *) currItem->item;
			setItem = setExps->firstItem;

			if (setItem->itemType == NEW_PTR) {
				NewClause *newStmt;

				newStmt = (NewClause *) setItem->item;
				strcpy(ClName,newStmt->className);

				setItem = setItem->next;

				while (setItem) {
					newStmt = (NewClause *) setItem->item;
					if (strcmp(ClName,newStmt->className)) {
						ur->setError( RUNTIME_ERROR, CLS_NAME_DIFF," in SET");
						return FALSE;
					}

					setItem = setItem->next;
				} // while

			} // if
			else { // it is a query (set).
				Query	*query;
				char	*astr;

				query = (Query *) setItem->item;

                                astr=query->checkAltintas();
				if (!astr) {
					ur->setError( RUNTIME_ERROR, INVLD_QUERY, " in SET");
					return FALSE;
				}
                                else strcpy( ClName, astr );
			} // else

			arguments[*ndx] = (char *) myAlloc(strlen(ClName)+5);
			strcpy(arguments[*ndx],"SET_");
			strcat(arguments[*ndx],ClName);

			break; 
		}
		case LST_PTR: { // "list"
			LinkedList	*listExps;
			ListItem	*listItem;
			NewClause	*newStmt;
			char		ClName[MAXSTRLEN];

			listExps = (LinkedList *) currItem->item;
			listItem = listExps->firstItem;

			newStmt = (NewClause *) listItem->item;
			strcpy(ClName,newStmt->className);

			listItem = listItem->next;

			while (listItem) {
				newStmt = (NewClause *) listItem->item;
				if (strcmp(ClName,newStmt->className)) {
					ur->setError( RUNTIME_ERROR, CLS_NAME_DIFF," in LIST");
					return FALSE;
				}

				listItem = listItem->next;
			} // while

			arguments[*ndx] = (char *) myAlloc(strlen(ClName)+6);
			strcpy(arguments[*ndx],"LIST_");
			strcat(arguments[*ndx],ClName);

			break; 
		}
		} // switch

		currItem  = currItem->next;
		(*ndx)++;

	} // while

	// Check the existence of the function
	if (!moodsType.searchSign(className, className, (char **)arguments,
		*ndx, success, funcType, &appOffset, 1 )) {
			ur->userError = catalogError;
                        ur->operationType = HANDLE_ERROR;
			return FALSE;
	}

	return TRUE;

} // checkConstructor


void NewClause::convertToList(AttrList *inAttr)
{

	if (inAttr->entry.AttributeConstructor == 'T')
		convertToList(inAttr->nextsub);

	if (inAttr->entry.AttributeConstructor != 'T') 
		attrList->insert(ATR_PTR, VOID inAttr);

	if (inAttr->next)
	   convertToList(inAttr->next);

} // convertToList
