#include <memory.h>
#include <sm_client.h>
#include <mood_catalog.h>
#include <math.h>

#define SLICE_LENGTH	10000

ReadSlice( char *buffer, int start, int length, OID Oid, int GrNdx )
{
	int status;
	USERDESC *userDesc;

	/* read the requested object */
	status = sm_ReadObject(GrNdx,&Oid, start, length, &userDesc);
	if (status != esmNOERROR) {
                printf("ESM Error : %s\n",sm_Error(sm_errno));
                return FALSE;
        }


	/* copy the object to the buffer */
	memcpy(buffer+start, userDesc->basePtr, length );

	status = sm_ReleaseObject( userDesc );
        if (status != esmNOERROR) {
                printf("ESM Error : %s\n",sm_Error(sm_errno));
                return FALSE;
        }
}

int PutObject(OID *Oid, char *buf, int ObjSize) {

	int GrNdx, status;

	if( buf == NULL ) return 1;

	status = sm_OpenBufferGroup(20,BF_MRU,&GrNdx,NOFLAGS);
	if (status != esmNOERROR) return FALSE;

	/* write the buf to "Unsized" objects' file... */
	status = sm_CreateObject(GrNdx, &UnsizedFID,NEAR_LAST,   
				NULL,NULL,ObjSize,(void *)buf,Oid);
	if (status != esmNOERROR) {
                printf("ESM Error : %s\n",sm_Error(sm_errno));
                return FALSE;
        }

	status = sm_CloseBufferGroup(GrNdx);
	if (status != esmNOERROR) return FALSE;

	return 1;

} /* PutObject */


int GetObject(OID Oid, char **buf, int *ObjSize) {

	int GrNdx, status, noOfSlices, i, remainderLength;
	OBJHDR   objHeader;

	if( ObjSize == NULL ) return 1;

	status = sm_OpenBufferGroup(20,BF_MRU,&GrNdx,NOFLAGS);
	if (status != esmNOERROR) return FALSE;

	/* Read Header of Object */
	status = sm_ReadObjectHeader(GrNdx,&Oid, &objHeader );
	if( status != esmNOERROR ) return FALSE;

	/* allocate memory for object buffer */
	*buf = (char *) malloc(objHeader.size);
	if( *buf == NULL ) return FALSE;

	*ObjSize = objHeader.size;

	noOfSlices = ( int ) ceil( 1.0 * objHeader.size / SLICE_LENGTH );

	remainderLength = objHeader.size - (noOfSlices-1) * SLICE_LENGTH;

	// read intermediate slices
	for( i = 0; i < noOfSlices-1; i++ )
		ReadSlice( *buf, i*SLICE_LENGTH, SLICE_LENGTH, Oid, GrNdx );  

	// read last slice
	ReadSlice( *buf, i*SLICE_LENGTH, remainderLength, Oid, GrNdx );  


	status = sm_CloseBufferGroup(GrNdx);
	if (status != esmNOERROR) return FALSE;

	return 1;
} /* GetObject */

