/*
*	It does the lexical analysis of a text. It is used in the
*	"userreq.h" in the function member gettoken
*/

#include <stdio.h>
#include <string.h>
#include "keytbl.h"
#include "keywords.h"
#include "userreq.h"
#include "mood_parser.h"
#include "bplusddl.h"
#include "ddl.h"
#include "decldml.h"


extern char *parseErrors[];
extern char *runtimeErrors[];


char * upperCase( char *str )
{
	char *s = str;
	while( *s ) {
		if( (*s >= 'a')  && (*s <= 'z') ) *s = *s - 'a' + 'A';
		else 
		s++;
	}
	return str;
}


void printToken(Token token)
{
	switch(token.type)
	{
		case LPAR : printf(" LPAR "); break;
		case RPAR : printf(" RPAR "); break;
		case LCB : printf(" LCB "); break;
		case RCB : printf(" RCB "); break;
		case LB : printf(" LB "); break;
		case RB : printf(" RB "); break;
		case DOT : printf(" DOT "); break;
		case COMMA : printf(" COMMA "); break;
		case SEMICOLON : printf(" SEMICOLON "); break;
		case MINUS : printf(" MINUS "); break;
		case PLUS : printf(" PLUS "); break;
		case SLASH : printf(" SLASH "); break;
		case PERCENT : printf(" PERCENT "); break;
		case STAR : printf(" STAR "); break;
		case EQ : printf(" EQ "); break;
		case GT : printf(" GT "); break;
		case GE : printf(" GE "); break;
		case LT : printf(" LT "); break;
		case LE : printf(" LE "); break;
		case NE : printf(" NE "); break;
		case UNDEFINEDTOKEN : printf(" UNDEFINEDTOKEN "); break;
		case STRCONS : printf(" STRCONS-"); break;
		case IDENTIFIER : printf(" IDENTIFIER-");
				  printf("%s", token.value);
				  break;
		case UNSIGNEDREAL : printf(" UNSIGNEDREAL-");
				  printf("%s", token.value);
				  break;
		case UNSIGNEDINT : printf(" UNSIGNEDINT-");
				  printf("%s", token.value);
				  break;
		default :
			printf(" %d ", token.type);
	}

	printf("Token Type  : %d\n",token.type);
	printf("Token Value : %s\n",token.value);
	//printf("Current : %d \n",curpos);  
}

// searches the keyword table for a given string
int searchKeyTable(char *p)
{
        int i=0;
        for(i = 0; strcmp(keytbl[i],"ENDOFKEYS") ; i++) {
                if(!strcasecmp((char *) keytbl[i],p)) 
			break;
	}

	if( !strcmp(keytbl[i],"ENDOFKEYS") )
        	return NOT_A_KEYWORD;

	if( i >= MS_FCN && i <= MS_TYPES )
	{
		switch( i ) {
			case MS_FCN : return FUNCTION;
			case MS_LS  : return LIST;
			case MS_DB  : return DATABASE;
			case MS_DATABASES : return DATABASE;
			case MS_FUNCTIONS : return FUNCTION;
			case DESC : return DESCRIBE;
			case MS_TYPES : return TYPE;
			default :  return i;
		}
	}
	else return i;
}

void UserRequest::printTime()
{
	printf(" Parsing Timem     : %10.4f secs.\n", parseTime/1000000 );
	printf(" Optimization Time : %10.4f secs.\n", optimizeTime/1000000 );
	printf(" Execution Time    : %10.4f secs.\n", actualTime/1000000 );
}

UserRequest::executeMSQL( char *inText)
{
	// reset();
	strcpy( text, inText );
	curpos = 0;
        prevpos = 0;
        userError.setError( NO_ERROR, 0, "");
        operation = NULL;
	startXAct();
	if( !parseText() ) return FALSE;
	if( !handleOperation()) return FALSE;
	//handleError();
	commitXAct();
	return TRUE;
}

UserRequest::execute_MSQL( char *inText)
{
	// reset();
	strcpy( text, inText );
	curpos = 0;
        prevpos = 0;
        userError.setError( NO_ERROR, 0, "");
        operation = NULL;
	if( !parseText() ) return FALSE;
	startXAct();
	return handleOperation();
}



// returns a substring of the text between prevpos and curpos
char *UserRequest::getValue()
{
	static char p[TOKENSIZE];
	unsigned i;
	if(curpos<prevpos) return(NULL);
	for( i = 0; i < curpos-prevpos ; i++ )
		p[i] = text[i+prevpos];
	p[i] = '\0';
	return(p);
}

char *UserRequest::getClassName( char *bindVar )
{
	return ( (Query *) operation)->getClassName(bindVar);
}


// skip tab characters during lexical analysis
void UserRequest::skipTabs() 
{
	while(text[curpos] == ' ' || text[curpos] == '\t' ||
	    text[curpos] == '\n')  curpos++;
}


// find a token in a text from current position curpos
UserRequest::getToken()
{
	int type;
	char *value;
	skipTabs();
	prevpos = curpos;
	token.type = -1;

	//printf("Current : %d\n", curpos);

	if(text[curpos] == '\''){
		char *p;
		curpos++;
		while( text[curpos] != '\'' && text[curpos]) curpos++;
		curpos++;
		token.type = STRCONS;
		p = getValue();
		p++; // omit beginning '
		strcpy(token.value,p );
		token.value[strlen(p)-1] = '\0';
#ifdef DEBUG
		printToken(token);
#endif
		return token.type;
	}

	if( (text[curpos] >= 'a') && (text[curpos] <= 'z') ||
	    (text[curpos] >= 'A') && (text[curpos] <= 'Z') ) {
		while( text[curpos] &&
		    (( text[curpos] >= 'a' && text[curpos] <= 'z') ||
		    ( text[curpos] >= 'A' && text[curpos] <= 'Z')  ||
		    ( text[curpos] >= '0' && text[curpos] <= '9')
		    || (text[curpos] =='_'))) curpos++;

		value = getValue();

		if( ( token.type=searchKeyTable(value)) == NOT_A_KEYWORD) {
			token.type = IDENTIFIER;
			strcpy(token.value,value);
		}
		else strcpy( token.value, upperCase( value ) );

#ifdef DEBUG
		printToken(token);
#endif
		return token.type;
	}

	if( text[curpos] >= '0' && text[curpos] <= '9') {
		while( text[curpos] && text[curpos] >= '0' &&
		    text[curpos] <= '9') curpos++;

		if(text[curpos] == '.') {
			curpos++;
			if( text[curpos] >= '0' && text[curpos] <= '9') {
				while( text[curpos] && text[curpos] >= '0' &&
				    text[curpos] <= '9') curpos++;
				if(text[curpos] == 'E') {
					curpos++;
					if( text[curpos] == '-' || text[curpos] == '+')
						curpos++;
					if( text[curpos] >= '0' && text[curpos] <= '9')
						while( text[curpos] && text[curpos] >= '0' &&
						    text[curpos] <= '9') curpos++;
				}
			}
			token.type = UNSIGNEDREAL;
		} else
			if(text[curpos] == 'E') {
				curpos++;
				if( text[curpos] == '-' || text[curpos] == '+')
					curpos++;
				if( text[curpos] >= '0' && text[curpos] <= '9')
					while( text[curpos] && text[curpos] >= '0' &&
					    text[curpos] <= '9') curpos++;
				token.type = UNSIGNEDREAL;
			}
			else token.type = UNSIGNEDINT ;
		strcpy(token.value,getValue());
#ifdef DEBUG
		printToken(token);
#endif
		return token.type;
	}

	switch (text[curpos]) {

	case '(' : 
		type = LPAR  ; 
		curpos++ ; 
		break;
	case ')' : 
		type = RPAR  ; 
		curpos++ ; 
		break;
	case '{' : 
		type = LCB   ; 
		curpos++ ; 
		break;
	case '}' : 
		type = RCB   ; 
		curpos++ ; 
		break;
	case '[' : 
		type = LB    ; 
		curpos++ ; 
		break;
	case ']' : 
		type = RB    ; 
		curpos++ ; 
		break;
	case '.' : 
		type = DOT   ; 
		curpos++ ; 
		break;
	case ',' : 
		type = COMMA ; 
		curpos++ ; 
		break;
	case ';' : 
		type = SEMICOLON ; 
		curpos++ ; 
		break;

	case '-' : 
		type = MINUS   ; 
		curpos++ ; 
		break;
	case '+' : 
		type = PLUS    ; 
		curpos++ ; 
		break;
	case '/' : 
		type = SLASH   ; 
		curpos++ ; 
		break;
	case '*' : 
		type = STAR    ; 
		curpos++ ; 
		break;
	case '%' : 
		type = PERCENT ; 
		curpos++ ; 
		break;

	case '=' : 
		type = EQ; 
		curpos++ ; 
		break;

	case '>' : 
		type = GT;
		curpos++;
		/* Check >= */
		if(text[curpos] == '=') {
			type = GE;
			curpos++;
		}
		break;

	case '<' : 
		type = LT;
		curpos++;
		if(text[curpos] == '=') {
			type = LE;
			curpos++;
		}
		if(text[curpos] == '>') {
			type = NE ;
			curpos++;
		}
		break;
	default  : 
		type = UNDEFINEDTOKEN;
		//setError(PARSE_ERROR, 20,"" );
	}
	token.type = type;
	if(type != UNDEFINEDTOKEN) strcpy(token.value,getValue());
#ifdef DEBUG
	printToken(token);
#endif

	return token.type;
}



// read the text of the user request
void UserRequest::readText()
{
        int   i=0;
        char  ch;
        curpos  = 0;
        prevpos = 0;
	printf("MoodSQL>");
        while((ch=getchar()) != ';')
        {
                if(ch=='\n') printf(">");
                text[i++] = ch;
        }
        text[i++] = ';';
        text[i] = '\0';
	ch = getchar();
}

void UserRequest::printText()
{
#ifdef DEBUG
	printf("******\n %s ********\n", text);
#endif
}

int UserRequest::handleOperation()
{
	// Check if there is an active database
	if( noDBInUse ) {
		switch( operationType )
		{
			case USE_DATABASE : 
			case DROP_DATABASE : 
			case CREATE_DATABASE :
			case LIST_DATABASES :
			case START_FILE :
			case EXIT_FROM_MOODS :	
			case HANDLE_ERROR :
				// do nothing
				break;
			default :
				setError( RUNTIME_ERROR, NO_DB_IN_USE, "");
				return MS_ERROR;
		}
	}

        switch(operationType)
        {
                case SELECT_STMT :
			//printf(" SUCCESSFULLY PARSED SQL STATEMENTS \n");
			//((Query *) operation)->printOn();
			if( !((Query *) operation)->generateAllTree())
				return MS_ERROR;
		//	((Query *) operation)->printOn();
	        	if( !((Query *) operation)->generatePlan() )
				return MS_ERROR;
			return MS_NOERROR;

		case DECLARE_STMNT :
			if(!((DeclStmnt *)operation)->createClasses())
				return MS_ERROR;
			printf("\nClass(es)/Type(s) created successfully...\n");
			return MS_NOERROR;

		case DESCRIBE_STMT :
			if( !((DescribeClause *) operation)->execute())
				return MS_ERROR;
			return MS_NOERROR;

                case START_FILE :
                        if( !((StartFile *) operation)->execute())
                                return MS_ERROR;
                        return MS_NOERROR;


                case DELETE_STMT :
			//(( DeleteClause *) operation)->printOn();
			if (!((( DeleteClause *) operation)->execute()))
				return MS_ERROR;
			printf("\nObject(s) deleted successfully...\n");
			return MS_NOERROR;

                case UPDATE_STMT :
			//(( UpdateClause *) operation)->printOn();
			if (!((( UpdateClause *) operation)->execute()))
                                return MS_ERROR;
			printf("\nObject(s) updated successfully...\n");
			return MS_NOERROR;

		case DISP_HIER :
			if( !(( DisplayHierarchy *) operation)->execute())
				return MS_ERROR;
			return MS_NOERROR;
			

		case INSERT_STMT :
			break;

		case OS_COMMAND :
		{
			char	command[120];
			strcpy( command, text+curpos);
			system( command );
			break;
		}

                case NEW_STMT    : {
			OID RetOid;

			//(( NewClause *) operation)->printOn();
			if (!((( NewClause *) operation)->execute(&RetOid)))
				return MS_ERROR;
			printf("\nObject(s) created successfully...\n");
			return MS_NOERROR;
		}

                case CREATE_INDEX :
			// if( !(( IndexCreate *) operation)->execute())
				// return MS_ERROR;
			return MS_NOERROR;

		case DROP_INDEX :
/*
			if( !(( IndexDrop *) operation)->execute())
				return MS_ERROR;
*/
			return MS_NOERROR;

		case CREATE_CLASS :
			return MS_NOERROR;

		case DROP_CLASS :
			if(!(( ClassDrop *)operation)->execute())
				return MS_ERROR;
			printf("\nClass(es) dropped successfully...\n");
			return MS_NOERROR;

		case LIST_TYPES :
			if(!(( ListType *)operation)->execute())
				return MS_ERROR;
			return MS_NOERROR;

		case CREATE_FUNCTION :
			if( !((CreateFunction *) operation)->execute())
				return MS_ERROR;
			printf( "\nFunction created successfully...\n");
			return MS_NOERROR;

		case DROP_FUNCTION :
			if( !((DropFunction *) operation)->execute())
				return MS_ERROR;
			printf( "\nFunction dropped successfully...\n");
			return MS_NOERROR;

		case LIST_FUNCTION :
			if( !((ListFunction *) operation)->execute())
				return MS_ERROR;
			return MS_NOERROR;

		case CREATE_DATABASE :
			if( !(( CreateDatabase *) operation)->execute())
				return MS_ERROR;
			setError( NO_ERROR, 0, "Database created successfully");
			return MS_NOERROR;

		case USE_DATABASE :
			if( !(( UseDatabase *) operation)->execute())
				return MS_ERROR;
			setError( NO_ERROR, 0, "Database is in use");
			return MS_NOERROR;

		case DROP_DATABASE :
			if( !(( DropDatabase *) operation)->execute())
				return MS_ERROR;
			setError( NO_ERROR, 0, "Database dropped successfully");
			return MS_NOERROR;

		case LIST_DATABASES :
			if( !(( ListDatabase *) operation)->execute())
				return MS_ERROR;
			return MS_NOERROR;


		case COMMIT_XACT :
			if(!commitXAct()) return MS_ERROR;
			return MS_NOERROR;
	
		case ABORT_XACT :
			if( !abortXAct() ) return MS_ERROR;
			return MS_NOERROR;

		case TAKE_SAVEPOINT :
			return MS_NOERROR;

		case ROLLBACK_TO_SAVEPOINT :
			return MS_NOERROR;

		case EXIT_FROM_MOODS :
			commitXAct();
			closeESM();
			exit( 1 );

                case HANDLE_ERROR :
			//printf("ERROR");
                        //handleError();
			return MS_ERROR;
        }
	return MS_NOERROR;
}

int UserRequest::reset()
{
	curpos = 0;
	prevpos = 0;
	strcpy( text, "");
	userError.setError( NO_ERROR, 0, "");
	textLength = 0;
	expectedTime = 0.0;
	actualTime = 0.0;
	parseTime = 0.0;
	optimizeTime = 0.0;

	if( !operation ) {
		operationType	= NO_OPERATION;
		operationTypeOnError = NO_OPERATION;
		return TRUE;
	}
	
        switch( operationTypeOnError )
        {
                case SELECT_STMT :
			delete ((Query *) operation); break;

                case DELETE_STMT :
			delete (( DeleteClause *) operation); break;

                case UPDATE_STMT :
			delete (( UpdateClause *) operation); break;

		case INSERT_STMT :
			break;

                case NEW_STMT    : 
			delete (( NewClause *) operation); break;

                case CREATE_INDEX :
			break;

		case DROP_INDEX :
			break;

		case CREATE_BJINDEX :
			break;

		case DROP_BJINDEX :
			break;

		case DESCRIBE_STMT :
			delete ((DescribeClause *) operation); break;

		case OS_COMMAND :
			break;

		case CREATE_TYPE :
		case CREATE_CLASS :
		case DECLARE_STMNT :
			delete (( DeclStmnt *) operation); break;

		case DROP_CLASS :
			delete (( ClassDrop *) operation); break;

		case CREATE_DATABASE :
			delete (( CreateDatabase *) operation); break;

		case USE_DATABASE :
			delete (( UseDatabase *) operation); break;

		case DROP_DATABASE :
			delete (( DropDatabase *) operation); break;

		case LIST_DATABASES :
			delete (( ListDatabase *) operation); break;

		case LIST_TYPES :
			delete (( ListType *) operation); break;

		case LIST_FUNCTION :
			delete (( ListFunction *) operation); break;

		case CREATE_FUNCTION :
			delete ((CreateFunction *) operation); break;

		case DROP_FUNCTION :
			delete ((DropFunction *) operation); break;

		case COMMIT_XACT :
			break; 
	
		case ABORT_XACT :
			break;

		case TAKE_SAVEPOINT :
			break;

		case ROLLBACK_TO_SAVEPOINT :
			break;

		case EXIT_FROM_MOODS :
			break;

                case HANDLE_ERROR :
			break;

		default :
			printf(" Unhandled Operation in Reset : %d \n",
				operationTypeOnError );
        }
	operationType	     = NO_OPERATION;
	operationTypeOnError = NO_OPERATION;
	operation            = NULL;
	return TRUE;
}

char *UserRequest::getErrorText()
{
	char *p;

	p = (char *) myAlloc( 3*MAXSTRLEN );

	switch( userError.errorType )
	{
		case PARSE_ERROR   :
			sprintf( p,"Parse Error :\n  %s \n  %s\n",
				parseErrors[userError.errorNo], userError.errorMsg);
			break;
		case RUNTIME_ERROR :
			sprintf( p, "Run time Error :\n  %s\n  %s\n",
				runtimeErrors[ userError.errorNo ],  userError.errorMsg);
			break;
		case EXODUS_ERROR :
			sprintf(p, "%s\n%s\n%s","ESM Error :",sm_Error(sm_errno),userError.errorMsg);
			break;
		case INTERNAL_ERROR :
			strcpy(p, "Internal Error :\n");
			strcat(p, userError.errorMsg);
			break;
		case EXCEPTION_ERROR :
			strcpy(p, "Exception Error :\n"); 
			strcat(p, userError.errorMsg);
			break;
		case NO_ERROR :
			strcpy(p,"");
	}
	return p;
}

char *UserRequest::getMessageText()
{
	char *p = new char(100);
	strcpy( p, userError.errorMsg );
	return p;
}

int UserRequest::execute()
{
	userError.errorType = NO_ERROR;
	parseText();
	if ( operationType == HANDLE_ERROR )
		return MS_ERROR;	
	return handleOperation();
	
}

extern dataDefinition(UserRequest *);

UserRequest::parseText()
{
	moodTimer.Start();

        getToken();
        switch(token.type){

		case EXIT :
			operationType = EXIT_FROM_MOODS;
			break;
 
                case RUN :
			operationType = RUN_PREVIOUS;
                        break;

		case START :
		{
			StartFile	*sf;
			sf = new StartFile( this );
			if( !sf->parse() ) operationType = HANDLE_ERROR;
                        break;
		}

		case MS_DISPLAY :
		{
			DisplayHierarchy	*dh;
			dh = new DisplayHierarchy( this );
			if( !dh->parse() ) operationType = HANDLE_ERROR;
			break;
		}
			

		/*case SAVE :
		case GET  :
		case EDIT :
			printf(" Not implemented yet\n");
			break;*/

 
                case DROP :
			if( !dropStatements(this)) operationType = HANDLE_ERROR;
                        break;

		case ABORT :
			operationType = ABORT_XACT;
			operationTypeOnError = ABORT_XACT;
			if( !abortXAct() ) {
				userError = catalogError;
				operationType = HANDLE_ERROR;
				return FALSE;
			}
			setError( NO_ERROR, 0, "Transaction Aborted");
			if( !startXAct()) {
				userError = catalogError;
				operationType = HANDLE_ERROR;
				return FALSE;
			}
			break;

		case COMMIT :
			operationType = COMMIT_XACT;
			operationTypeOnError = COMMIT_XACT;
			if( !commitXAct() ) {
				userError = catalogError;
				operationType = HANDLE_ERROR;
				return FALSE;
			}
			setError( NO_ERROR, 0, "Transaction Committed");
			if( !startXAct()) {
				userError = catalogError;
				operationType = HANDLE_ERROR;
				return FALSE;
			}
			break;

		case HOST :
			operationType = OS_COMMAND;
			operationTypeOnError = OS_COMMAND;
			return  TRUE;
 
                case DESCRIBE :
		{
			DescribeClause	*descStmt;

			backtrack();

			descStmt = new DescribeClause( this );

			if(!descStmt->parse() ) operationType = HANDLE_ERROR;

                        break;
		}
 
                case DELETE :
		{
			DeleteClause	*deleteStmt;

			backtrack();

			deleteStmt = new DeleteClause(this);

                        if(!deleteStmt->parse()) operationType = HANDLE_ERROR;

			break;
		}

		case UPDATE :
		{
			UpdateClause	*updateStmt;

			backtrack();

			updateStmt = new UpdateClause(this);

			if( !updateStmt->parse() ) operationType = HANDLE_ERROR;
			//updateStmt->printOn();

			break;
		}

		case BEGIN : 
		{
			DeclStmnt *ds=new DeclStmnt(this);
			
			if( !ds->parse() ) operationType = HANDLE_ERROR;

			break;
			
		}

                case CREATE :
			if( !createStatements(this)) operationType = HANDLE_ERROR;
                        break;

		// use database clause
		case USE :
		{
			UseDatabase	*dbUse;
			getToken();
			if( tokenType() != DATABASE )
			{
				setError( PARSE_ERROR, UNKNOWN_SYNTAX, "");
				return MS_ERROR;
			}
			dbUse = new UseDatabase( this );
			dbUse->connectUserRequest();
			if( !dbUse->parse() ) operationType = HANDLE_ERROR;
			break;
		}

		case LIST :
		{
			getToken();
			switch( tokenType() ) {
			case  DATABASE  : {
				ListDatabase	*dbListing;
				dbListing = new ListDatabase( this );
				dbListing->connectUserRequest();
				break;
			}
			case TYPE :  {
				ListType	*typeListing;
				typeListing = new ListType( this );
				typeListing->connectUserRequest();
				break;
			} 
                       	case FUNCTION : {
				ListFunction	*fnListing;
				fnListing = new ListFunction( this );
				if( !fnListing->parse() )
					operationType = HANDLE_ERROR;
				break;
			}
			default :
				operationType = HANDLE_ERROR;
				setError( PARSE_ERROR, UNKNOWN_SYNTAX, "");
				return MS_ERROR;
			}
			break;
		}

		case MS_SELECT :
		{

			Query *query;

			backtrack();

			query = new Query(this);

			query->connectUserRequest();

			if( !query->parse() ) operationType = HANDLE_ERROR;

			break;

		}

		case NEW :
		{

			NewClause *newStmt;	

			backtrack();

			newStmt = new NewClause(this);

			newStmt->connectUserRequest();

			if( !newStmt->parse() ) operationType = HANDLE_ERROR;

			break;

		}

                default :
                        setError(PARSE_ERROR, 12, "");
                        operationType = HANDLE_ERROR;

        }
	moodTimer.Stop();
	parseTime = moodTimer.RealTime();

	return TRUE; 
}

RELATION_SCHEMA *UserRequest::getBulkSchema()
{
	if( operationType == SELECT_STMT ) 
        	return ((Query *) operation)->solutionFile;
	else return NULL;
}


